/*
  This file is part of the Ayyi Project. http://ayyi.org
  copyright (C) 2004-2012 Tim Orford <tim@orford.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 3
  as published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#define __am_private__
#include <model/common.h>
#include <ayyi/ayyi_list.h>
#include <ayyi/ayyi_shm.h>
#include <ayyi/ayyi_client.h>
#include <ayyi/ayyi_song.h>
#include <ayyi/ayyi_mixer.h>
#include <ayyi/ayyi_action.h>
#include <ayyi/ayyi_dbus.h>
#include "ayyi/ayyi_dbus_proxy.h"

#include "model/am_plugin.h"
#include "model/am_track_peer.h"
#include "model/am_song.h"
#include "model/am_client.h"


AyyiClient*
am__init(AyyiService** services)
{
	// if services is null, the default Ardourd service will be used.

	static gboolean done = false;
	g_return_val_if_fail(!done, NULL);

	AyyiClient* a = ayyi_client_init();

	ayyi_client_add_services(services);

	ayyi_log_init();
	am_plugin__init();
	am_tracks_init();
	am_song__new();

	am_song__add_handlers();

	done = true;

	return a;
}


/*
 *  Convenience function that will establish both messaging, and shm.
 *  The shm callback is only called once _all_ segments are attached.
 *  -it is the same as calling am__connect_messaging and am__get_shm.
 */
void
am__connect_services(AyyiService** services, AyyiHandler2 on_connect, gpointer _user_data)
{
	g_return_if_fail(services && services[0]);

	typedef struct
	{
		AyyiHandler2  on_all_shm;
		gpointer      user_data;
		int           timeout;
		AyyiService** services;
		int           s;
		void          (*on_messaging)(GError*, gpointer);
		GError*       error;  // currently only the first error is forwarded.
	} C;

	void clear_timeout(C* c)
	{
		am_source_remove0(c->timeout);
	}

	void cleanup(C* c)
	{
		clear_timeout(c);
		g_free(c);
	}

	gboolean am_connect_timeout(gpointer _c)
	{
		C* c = _c;
		call(c->on_all_shm, c->error, c->user_data);
		//TODO free c - check is always safe
		return TIMER_STOP;
	}

	void am__next_service(C* c)
	{
		GError** error = NULL;
		dbg(2, "s=%i", c->s + 1);
		if(c->services[++c->s])
			am__connect_messaging(c->services[c->s], c->on_messaging, c, error);
		else{
			dbg(2, "no more services");
			cleanup(c);
		}
	}

	void am_connect_services__on_shm(AyyiShmCallbackData* d)
	{
		C* c = d->user_data;

		if(!d->shm_seg){
			// its too late to set the GError here. Error is indicated by NULL segment.
			c->on_all_shm(c->error, c->user_data);
			cleanup(c);
			return;
		}

		bool
		service_shm_setup_is_complete(AyyiService* service)
		{
			dbg(1, "shm.song=%p ((AyyiSongService*)ayyi.service)->amixer=%p", ((AyyiSongService*)ayyi.service)->song, ((AyyiSongService*)ayyi.service)->amixer);

			if(service == known_services[AYYI_SERVICE_MIXER]){
				GList* l = service->segs;
				for(;l;l=l->next){
					ShmSeg* seg = l->data;
					if(!seg->attached || seg->invalid){ dbg(2, "not complete."); return false; }
				}
				return g_list_length(service->segs) >= G_N_ELEMENTS(((AyyiMixerService*)service)->segments);
			}
			return ((((AyyiSongService*)service)->song) && (((AyyiSongService*)service)->amixer));
		}

		if(service_shm_setup_is_complete(d->service)){
			clear_timeout(c);
			//am__register_signals(ayyi.service);
			int i; for(i=0;c->services[i];i++){
				if(c->services[i] == known_services[AYYI_SERVICE_AYYID1]){ //FIXME create a service_load fn?
					am_song__load();
					break;
				}
			}
			c->on_all_shm(c->error, c->user_data);
		}

		am__next_service(c);
	}

	void am__on_messaging(GError* error, gpointer _c)
	{
		C* c = _c;
		if(error){
			gwarn("error");
			c->error = g_error_copy(error);
			call(c->on_all_shm, c->error, c->user_data);
		}else{
			AyyiService* service = c->services[c->s];
			am__get_shm(service, am_connect_services__on_shm, c);
		}
	}

	C* c = AYYI_NEW(C,
		.on_all_shm = on_connect,
		.user_data = _user_data,
		.services = services,
		.s = -1,
		.on_messaging = am__on_messaging
	);
	c->timeout = g_timeout_add(4000, am_connect_timeout, c);

	am__next_service(c);
}


/*
 *  Convenience function that will establish both messaging, and shm for the Ardourd song and mixer services.
 *  The shm callback is only called once _all_ segments are attached.
 *  -is the same as calling am__connect_messaging and am__get_shm.
 *
 *  If you wish to specify which services to connect to, use am__connect_services instead
 *
 *  To check for errors, check the GError argument in the callback.
 */
void
am__connect_all(AyyiHandler2 on_connect, gpointer _user_data)
{
	AyyiService* services[] = {
		(AyyiService*)&ardourd_song,
		(AyyiService*)&ardourd_mixer,
		NULL,
	};

	typedef struct
	{
		AyyiHandler2 user_callback;
		gpointer     user_data;
		int          timeout;
	} C;
	C* c = g_new0(C, 1);
	c->user_callback = on_connect;
	c->user_data = _user_data;

	void am__connect_connected(GError* error, gpointer _c)
	{
		C* c = _c;
		call(c->user_callback, error, c->user_data);
		g_free(c);
	}

	am__connect_services((AyyiService**)&services, am__connect_connected, c);
}


gboolean
am__connect_messaging(AyyiService* service, AyyiHandler2 callback, gpointer user_data, GError** error)
{
	// Currently this should not to be used for connecting to plugins because of the signal registration.

	static bool signal_reg_done = false;

	HandlerData* c = g_new0(HandlerData, 1);
	c->callback = (AyyiHandler)callback;
	c->user_data = user_data;

	void am__connect_messaging__on_connected(GError* error, gpointer user_data)
	{
		PF;
		HandlerData* c = user_data;
		call((AyyiHandler2)c->callback, error, c->user_data);
		g_free(c);
	}

	if(ayyi_client_connect(service, am__connect_messaging__on_connected, c, error)){
		if(!signal_reg_done) ayyi_client_set_signal_reg(NEW | DELETED | CHANGED | TRANSPORT | LOCATORS | PROPERTY | PROGRESS);
		signal_reg_done = true;
		return true;
	}
	return false;
}


void
am__get_shm(AyyiService* service, void (*on_shm)(AyyiShmCallbackData*), gpointer user_data)
{
	ayyi_client__dbus_get_shm(service, on_shm, user_data);
}


gboolean
am__get_shm_(AyyiService** services, void (*on_shm)(AyyiShmCallbackData*), gpointer user_data)
{
	g_return_val_if_fail(services, false);

	int i; for(i=0;services[i];i++){
		if(!ayyi_client__dbus_get_shm(services[i], on_shm, user_data)) return false;
	}
	return true;
}


