/*
  This file is part of AyyiGtk. http://ayyi.org
  copyright (C) 2004-2017 Tim Orford <tim@orford.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 3.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#include <model/am_midi_part.h>
#include "am_part_manager.h"

static void am_note_selection_free(GList**);


MidiPart*
am_midi_part__new()
{
	return AYYI_NEW(MidiPart,
		.part = {
			.ident.type = AYYI_OBJECT_MIDI_PART,
			.length = (AyyiSongPos){8,},
			.ref_count = 1,
		}
	);
}


MidiNote*
am_midi_part__get_note_by_idx(AMPart* part, AyyiIdx shm_idx)
{
	AyyiMidiRegion* midi_region = ayyi_song__midi_region_at(part->ayyi->shm_idx);
	AyyiContainer* container = &midi_region->events;
	MidiNote* note = (MidiNote*)ayyi_song_container_get_item(container, shm_idx);

	return note;
}


MidiNote*
am_midi_part__get_next_event(AMPart* part, MidiNote* prev)
{
													// TODO change to be like am_midi_part__get_prev_midi_event below
	g_return_val_if_fail(part, NULL);

	AyyiMidiRegion* region = ayyi_song__midi_region_at(part->ayyi->shm_idx);
	AyyiContainer* container = &region->events;
	if(!container || container->last < 0) return NULL;

	if(prev && (am_midi_part__find_note(region, prev) >= AYYI_BLOCK_SIZE)) gwarn ("FIXME");

	int not_first = 0;

	int b; for(b=0;b<=container->last;b++){
		block* blk = ayyi_song__translate_address(container->block[b]);
		void** table = blk->slot;
		if(!table){ gerr ("shm error. table=null block=%i", b); return NULL; }

		if(!prev){
			//return first
			dbg (2, "first! region=%p", ayyi_song__translate_address(table[0]));
			return ayyi_song__translate_address(table[0]);
		}

		int i = (not_first++) ? 0 : am_midi_part__find_note(region, prev) + 1;
		for(;i<=blk->last;i++){
			dbg (2, "b=%i i=%i", b, i);
			MidiNote* next = ayyi_song__translate_address(table[i]);
			if(!next) continue;
			if(!is_song_shm_local(next)){ gerr ("bad shm pointer. table[%u]=%p", i, table[i]); return NULL; }
			return next;
		}
	}
	return NULL;
}


MidiNote*
am_midi_part__get_prev_midi_event(AMPart* part, MidiNote* prev)
{
	AyyiMidiRegion* region = ayyi_song__midi_region_at(part->ayyi->shm_idx);
	AyyiContainer* container = &region->events;
	return (MidiNote*)ayyi_midi_note_prev((AyyiItem*)prev);
}


void
am_midi_part__print_events(AMPart* part)
{
	AyyiMidiRegion* region = (AyyiMidiRegion*)part->ayyi;

	int n_events = ayyi_song_container_count_items(&region->events);
	if(n_events){
		printf("      events (%i):\n", n_events);
		int n = 0;
		MidiNote* note = NULL;
		while((note = (MidiNote*)ayyi_song_container_next_item(&region->events, note))){
			printf("        %i: %i %i %i %i\n", n, note->note, note->velocity, note->start, note->length);
			n++;
		}
	}
	else printf("      no events\n");
}


int
am_midi_part__find_note(AyyiMidiRegion* region, MidiNote* note)
{
	//temp? probably should add shm_idx to the note struct like the others.
	g_return_val_if_fail(region, false);
	g_return_val_if_fail(note, false);

	if(!&region->events){ gwarn("region has no events"); return -1; }

	return ayyi_container_find(&region->events, note, ayyi_song__translate_address);
}


gboolean
am_midi_part__sync(MidiPart* part)
{
	g_return_val_if_fail(part, false);
	gboolean err = false;
	AMPart* gpart = (AMPart*)part;
	g_return_val_if_fail(gpart->ayyi, false);

	AyyiMidiRegion* shared = ayyi_song__midi_region_at(gpart->ayyi->shm_idx);
	if(!shared->id){ gerr ("illegal id 0");/* err = true;*/ }

	_am_part_sync_name(gpart, (AyyiRegionBase*)shared);
	_am_part_sync_pos(gpart);

	am_part__sync_length(gpart);

	return !err;
}


/*
 *   Initialise data for a part which is new to the gui but already exists in core.
 */
MidiPart*
am_midi_part__init_from_index(MidiPart* mpart, AyyiIdx ayyi_index)
{
	AMPart* part = (AMPart*)mpart;

	part->ayyi       = (AyyiRegionBase*)ayyi_song_container_get_item(&((AyyiSongService*)ayyi.service)->song->midi_regions, ayyi_index);
	part->ident      = (AyyiIdent){AYYI_OBJECT_MIDI_PART, ayyi_index};
	part->track      = _am_part_get_ayyi_track(part);
	part->pool_item  = NULL;

	if(!part->track) return NULL;
	if(!am_midi_part__sync(mpart)) gerr ("init failed - should abort...\n");

	if(part->length.beat > AM_MAX_PART_LENGTH_BEATS){
		gwarn ("part length too long (%i).\n", part->length.beat);
		part->length = (AyyiSongPos){100, 0, 0};
	}

	am_parts__emit("add", part, NULL);

	return mpart;
}


/*
 *   Ownership of the selection is taken.
 */
void
am_midi_part__set_note_selection(MidiPart* part, GList* selection, gpointer user_data)
{
	am_note_selection_free(&part->note_selection);
	part->note_selection = selection;

	g_signal_emit_by_name (am_parts, "item-changed", part, AM_CHANGE_SELECTION, user_data);
}


#define CHECK_NOTE(note) { \
		g_return_if_fail(note->note <= 128); \
		g_return_if_fail(note->length > 0); \
	}

void
am_midi_part__set_notes(AMPart* part, const GList* notes, AyyiHandler callback, gpointer user_data)
{
	//@param notes, list of TransitionalNote*.

	typedef struct {
		AyyiHandler callback;
		gpointer    user_data;
	} C;

	void
	am_midi_set_notes_done(AyyiAction* action)
	{
		PF;
		C* c = action->app_data;
		call(c->callback, (AyyiIdent){action->obj.type, action->ret.idx}, &action->ret.error, c->user_data);
	}

	AyyiMidiRegion* region = (AyyiMidiRegion*)part->ayyi;

	AyyiAction* a = ayyi_action_new("set notes");
	a->callback = am_midi_set_notes_done;
	a->app_data = AYYI_NEW(C,
		.callback = callback,
		.user_data = user_data
	);

	GPtrArray* array = g_ptr_array_new();

	GList* l = (GList*)notes;
	int i = 0;
	for(;l;l=l->next){
		TransitionalNote* tn = l->data;
		MidiNote* note = &tn->transient;
		dbg(1, "idx=%i note=%i start=%u v=%i", note->idx, note->note, note->start, note->velocity);
		CHECK_NOTE(note);
		g_return_if_fail(ayyi_container_index_ok(&region->events, note->idx));

		GValue value = {0,};
		g_value_init (&value, DBUS_STRUCT_UINT_BYTE_BYTE_UINT_UINT);
		g_value_take_boxed (&value, dbus_g_type_specialized_construct (DBUS_STRUCT_UINT_BYTE_BYTE_UINT_UINT));

		dbus_g_type_struct_set (&value, 0, note->idx, 1, note->note, 2, note->velocity, 3, note->start, 4, note->length, G_MAXUINT);
		g_ptr_array_add (array, g_value_get_boxed(&value));
		dbg(2, "array_len=%i key=%i", array->len, note->note);

		i++;
	}
	if(array->len) dbus_set_notelist(a, part->ayyi->shm_idx, array);

	g_ptr_array_free(array, TRUE);
}


static void
am_note_selection_free(GList** selection)
{
	PF;
	if(*selection){
		g_list_free0(*selection);
	}
}


