/*
  This file is part of the Ayyi Project. http://ayyi.org
  copyright (C) 2004-2017 Tim Orford <tim@orford.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 3
  as published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#define __am_common_c__
#include "config.h"
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <getopt.h>
#include <time.h>
#include <unistd.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <glib.h>
#include "model/ayyi_model.h"
#include "test/am_common.h"

int        n_failed = 0;
int        n_passed = 0;
bool       abort_on_fail = true;
gboolean   passed = false;
int        test_finished = false;  //current test has finished. Go onto the next test.
gpointer*  current_tests;
int        current_test = -1;
char       current_test_name[64];
GMainLoop* main_loop = NULL;

static int __n_tests = 0;


void
am_test_init(gpointer tests[], int n_tests)
{
	__n_tests = n_tests;
	dbg(2, "n_tests=%i", __n_tests);

	memset(&app, 0, sizeof(struct _app));

	g_log_set_handler (NULL, G_LOG_LEVEL_WARNING | G_LOG_LEVEL_CRITICAL | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION, log_handler, NULL);
	g_log_set_handler ("Gtk", G_LOG_LEVEL_CRITICAL | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION, log_handler, NULL);
	g_log_set_handler ("Gdk", G_LOG_LEVEL_CRITICAL | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION, log_handler, NULL);
	g_log_set_handler ("GLib", G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION, log_handler, NULL);
	g_log_set_handler ("GLib-GObject", G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION, log_handler, NULL);
	g_log_set_handler ("Gdl", G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION, log_handler, NULL);
	g_log_set_handler ("Ayyi", G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION, log_handler, NULL);

#ifndef HAVE_GLIB_2_36
	g_thread_init(NULL);
	g_type_init();
#endif

	am__init(NULL);
}


void
app_on_all_shm(GError* e, gpointer user_data)
{
	if(!((AyyiSongService*)ayyi.service)->song){
		exit(EXIT_FAILURE);
	}

	if(e){
		printf("failed to connect\n");
		if(main_loop){
			g_main_loop_unref(main_loop);
			main_loop = NULL;
		}
		exit(EXIT_FAILURE);
	}

	if(((AyyiSongService*)ayyi.service)->song->version != AYYI_SHM_VERSION){
		gerr ("wrong shm version: %i - should be %i", ((AyyiSongService*)ayyi.service)->song->version, AYYI_SHM_VERSION);
		return;
	}

	log_print(LOG_OK, "connected to server");

	next_test();
}


void
next_test()
{
	bool
	run_test(gpointer test)
	{
		((Test)test)();
		return TIMER_STOP;
	}

	bool _exit()
	{
		exit(EXIT_SUCCESS);
		return TIMER_STOP;
	}

	printf("\n");
	current_test++;
	if(current_test < __n_tests){
		if(app.timeout) g_source_remove (app.timeout);
		test_finished = false;
		gboolean (*test)() = current_tests[current_test];
		dbg(2, "test %i of %i.", current_test + 1, __n_tests);
		g_timeout_add(300, run_test, test);

		bool on_test_timeout(gpointer _user_data)
		{
			FAIL_TEST_TIMER("TEST TIMEOUT\n");
			return TIMER_STOP;
		}
		app.timeout = g_timeout_add(20000, on_test_timeout, NULL);
	}
	else{ printf("finished all. passed=%s%i%s failed=%s%i%s\n", green, app.n_passed, white, (n_failed ? red : white), n_failed, white); g_timeout_add(1000, _exit, NULL); }
}


void
test_finished_()
{
	dbg(2, "... passed=%i", passed);
	if(passed) app.n_passed++;// else n_failed++; n_failed is now incremented earlier to avoid duplicates
	log_print(passed ? LOG_OK : LOG_FAIL, "%s", current_test_name);
	if(!passed && abort_on_fail) current_test = 1000;
	next_test();
}


gboolean
get_random_boolean()
{
	int r = rand();
	int s = RAND_MAX / 2;
	int t = r / s;
	return t;
}


int
get_random_int(int max)
{
	int r = rand();
	int s = RAND_MAX / max;
	int t = r / s;
	return t;
}


void
errprintf4(char* format, ...)
{
	char str[256];

	va_list argp;
	va_start(argp, format);
	vsprintf(str, format, argp);
	va_end(argp);

	printf("%s%s%s\n", red, str, white);
}


/*
 *  Periodically call the test fn. wait is finished when test returns true.
 */
void
am_wait_for(gboolean (*test)(gpointer), AyyiHandler callback, gpointer user_data)
{
	g_return_if_fail(callback);
	g_return_if_fail(test);

	typedef struct
	{
		AyyiHandler callback;
		gpointer    user_data;
		gboolean    (*test)(gpointer);
		int         wait_count;
	} C;

	bool finish(C* c)
	{
		g_free(c);
		return G_SOURCE_REMOVE;
	}

	bool am_wait_for_iter(gpointer _c)
	{
		C* c = _c;
		g_return_val_if_fail(c, finish(c));

		if(!c->test(c->user_data)){
			dbg(((c->wait_count > 0) && (c->wait_count % 4)) ? 2 : 3, "not yet ready (%i)", c->wait_count);
			c->wait_count = c->wait_count + 1;
			return c->wait_count < 100 ? TIMER_CONTINUE : finish(c);
		}
		dbg(2, "wait done.");
		c->callback(AYYI_NULL_IDENT, NULL, c->user_data);
		return finish(c);
	}

	g_timeout_add(1000, am_wait_for_iter, AYYI_NEW(C,
		.callback = callback,
		.user_data = user_data,
		.test = test
	));
}


void reset_timeout(int ms)
{
	if(app.timeout) g_source_remove (app.timeout);

	bool on_test_timeout(gpointer _user_data)
	{
		FAIL_TEST_TIMER("TEST TIMEOUT\n");
		return TIMER_STOP;
	}
	app.timeout = g_timeout_add(ms, on_test_timeout, NULL);
}


