/**
* +----------------------------------------------------------------------+
* | This file is part of the Ayyi project. http://www.ayyi.org           |
* | copyright (C) 2004-2017 Tim Orford <tim@orford.org>                  |
* +----------------------------------------------------------------------+
* | This program is free software; you can redistribute it and/or modify |
* | it under the terms of the GNU General Public License version 3       |
* | as published by the Free Software Foundation.                        |
* +----------------------------------------------------------------------+
*
*/
#include "global.h"
#include <glib.h>
#include <glib-object.h>
#include "support.h"
#include "settings.h"
#include "shell.h"
#include "song.h"
#include "transport.h"
#include "panels/plugin.h"
#include "panels/arrange.h"

SMConfig* config;

#define AYYI_TYPE_APP            (ayyi_app_get_type ())
#define AYYI_APP(obj)            (G_TYPE_CHECK_INSTANCE_CAST ((obj), AYYI_TYPE_APP, AyyiApp))
#define AYYI_APP_CLASS(klass)    (G_TYPE_CHECK_CLASS_CAST ((klass), AYYI_TYPE_APP, AyyiAppClass))
#define AYYI_IS_APP(obj)         (G_TYPE_CHECK_INSTANCE_TYPE ((obj), AYYI_TYPE_APP))
#define AYYI_IS_APP_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), AYYI_TYPE_APP))
#define AYYI_APP_GET_CLASS(obj)  (G_TYPE_INSTANCE_GET_CLASS ((obj), AYYI_TYPE_APP, AyyiAppClass))

typedef struct _AyyiApp AyyiApp;
typedef struct _AyyiAppClass AyyiAppClass;
typedef struct _AyyiAppPrivate AyyiAppPrivate;

static gpointer ayyi_app_parent_class = NULL;

GType ayyi_app_get_type (void) G_GNUC_CONST;

enum {
	AYYI_APP_DUMMY_PROPERTY
};

static GObject* ayyi_app_constructor (GType, guint n_construct_properties, GObjectConstructParam*);
static void     ayyi_app_finalize    (GObject*);


AyyiApp*
ayyi_app_construct (GType object_type)
{
	AyyiApp* self = (AyyiApp*) g_object_new (object_type, NULL);
	return self;
}


AyyiApp*
app_new (void)
{
	app = ayyi_app_construct (AYYI_TYPE_APP);

	app->dnd = (struct AppDnd){
		.file_drag_types_count = 2,
		.file_drag_types = {
			{ "text/uri-list", 0, AYYI_TARGET_URI_LIST },
			{ "text/plain",    0, TARGET_TEXT_PLAIN }
		}
	};

	config = g_new0(SMConfig, 1);

	return app;
}


static GObject*
ayyi_app_constructor (GType type, guint n_construct_properties, GObjectConstructParam* construct_properties)
{
	GObjectClass* parent_class = G_OBJECT_CLASS (ayyi_app_parent_class);
	GObject* obj = parent_class->constructor (type, n_construct_properties, construct_properties);
	return obj;
}


static void
ayyi_app_class_init (AyyiAppClass* klass)
{
	ayyi_app_parent_class = g_type_class_peek_parent (klass);
	G_OBJECT_CLASS (klass)->constructor = ayyi_app_constructor;
	G_OBJECT_CLASS (klass)->finalize = ayyi_app_finalize;
	g_signal_new ("app_stop_meters", AYYI_TYPE_APP, G_SIGNAL_RUN_LAST, 0, NULL, NULL, g_cclosure_marshal_VOID__VOID, G_TYPE_NONE, 0);
	g_signal_new ("app_start_meters", AYYI_TYPE_APP, G_SIGNAL_RUN_LAST, 0, NULL, NULL, g_cclosure_marshal_VOID__VOID, G_TYPE_NONE, 0);
	g_signal_new ("styles_available", AYYI_TYPE_APP, G_SIGNAL_RUN_LAST, 0, NULL, NULL, g_cclosure_marshal_VOID__VOID, G_TYPE_NONE, 0);
	g_signal_new ("shortcuts_changed", AYYI_TYPE_APP, G_SIGNAL_RUN_LAST, 0, NULL, NULL, g_cclosure_marshal_VOID__VOID, G_TYPE_NONE, 0);
	g_signal_new ("focus_changed", AYYI_TYPE_APP, G_SIGNAL_RUN_LAST, 0, NULL, NULL, g_cclosure_marshal_VOID__VOID, G_TYPE_NONE, 0);
}


static void
ayyi_app_instance_init (AyyiApp* self)
{
	self->state = 0;
	self->recording = FALSE;
}


static void
ayyi_app_finalize (GObject* obj)
{
	G_OBJECT_CLASS (ayyi_app_parent_class)->finalize (obj);
}


GType
ayyi_app_get_type (void)
{
	static volatile gsize ayyi_app_type_id__volatile = 0;
	if (g_once_init_enter (&ayyi_app_type_id__volatile)) {
		static const GTypeInfo g_define_type_info = { sizeof (AyyiAppClass), (GBaseInitFunc) NULL, (GBaseFinalizeFunc) NULL, (GClassInitFunc) ayyi_app_class_init, (GClassFinalizeFunc) NULL, NULL, sizeof (AyyiApp), 0, (GInstanceInitFunc) ayyi_app_instance_init, NULL };
		GType ayyi_app_type_id;
		ayyi_app_type_id = g_type_register_static (G_TYPE_OBJECT, "AyyiApp", &g_define_type_info, 0);
		g_once_init_leave (&ayyi_app_type_id__volatile, ayyi_app_type_id);
	}
	return ayyi_app_type_id__volatile;
}


void
app_set_state(AppState state)
{
	static int aborting = 0;

	app->state = state;

	switch(state){
		case APP_STATE_STARTED:
			if(song && song->periodic){ g_source_remove(song->periodic); song->periodic = 0; }
			break;
		case APP_STATE_ENGINE_OK:
			am_plugin__sync();
			break;
		case APP_STATE_SONG_LOAD_FAILED:
			{
				Arrange* arrange;
				if((arrange = ARRANGE_FIRST)) arr_cursor_reset(arrange, gtk_widget_get_toplevel(arrange->canvas->widget)->window);
			}
			break;
		case APP_STATE_ABORT:
			transport_disable();
			if(!aborting){
				aborting = TRUE;
				app_quit();
			}
			break;
	}
}


void
app_quit()
{
	// though not currently used, we would like the main loop to be still running here.

	app_set_state(APP_STATE_SHUTDOWN);

	GList* l = shell_list;
	for(;l;l=l->next){
		set_cursor(((AyyiShell*)l->data)->widget->window, CURSOR_WATCH); //set Busy cursor.
	}

	log_print(0, "shutting down...");
	while(gtk_events_pending()) gtk_main_iteration(); //update the display.

	config_save(); // FIXME only do this if config and windows actually were loaded...

	gtk_accel_map_save(config->accels_file);

	song_unload();
	if(!song_free()) gerr ("song free failed.");

	toolbar_class_free();
	plugins_free();

	printf("exiting...\n");
	exit(0);
}



