/*
  This file is part of the Ayyi Project. http://www.ayyi.org
  copyright (C) 2004-2017 Tim Orford <tim@orford.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 3
  as published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#include "global.h"
#define RSVG_DISABLE_DEPRECATION_WARNINGS
#include <librsvg/rsvg.h>

#include "arrange.h"
#include "support.h"
#include "icon.h"

#include "../pixmaps/cursor-openhand.xbm"
#include "../pixmaps/cursor-openhand_mask.xbm"
#include "../pixmaps/cursor-closedhand.xbm"
#include "../pixmaps/cursor-closedhand_mask.xbm"
#include "pixmaps/scissors.xbm"
//#include "pixmaps/cursor-zoom-mask.xbm"
//#include "pixmaps/cursor-zoom.xbm"
#include "pixmaps/zoom_in.xbm"
#include "pixmaps/zoom_in_mask.xbm"

extern SMConfig* config;

static GHashTable* icons = NULL;

static GdkPixbuf* svg_pixbuf_new(const char* filename, int size);


void
icons_init()
{
	icons = g_hash_table_new(g_str_hash, g_str_equal);
}


GdkPixbuf*
get_icon(const char* name, int size)
{
	//name - the svg file name without the svg extension.

	//the icon cache only holds icons with size of 16px.

	//caller must unref the pixbuf when finished with it.

	dbg(2, "name=%s", name);
	GdkPixbuf* pixbuf = (size == 16) ? g_hash_table_lookup(icons, name) : NULL;
	if(pixbuf){
		g_object_ref(pixbuf);
	}else{
		if(size == 16) dbg(3, "icon cache miss. '%s'", name);
		gchar* filename = g_strdup_printf("%s.svg", name);
		if((pixbuf = svg_pixbuf_new(filename, size)) && (size == 16)){
			g_hash_table_insert(icons, (char*)name, pixbuf);
			g_object_ref(pixbuf);
		}
		g_free(filename);
	}
	return pixbuf;
}


GtkWidget*
svg_image_new(const char* svgpath)
{
	//creates a new gtkimage from an svg file.

	gchar* path = g_strdup_printf("%s/%s.svg", config->svgdir, svgpath);

	if(!g_file_test(path, G_FILE_TEST_EXISTS)) gwarn("file doesnt exist '%s'", path);

	GdkPixbuf* pixbuf = get_icon(svgpath, 16);
	if(!pixbuf) gwarn("svg load failed '%s'.", path);

	//put rendered svn icon onto a widget:
	GtkWidget* image = gtk_image_new_from_pixbuf(pixbuf);

	g_free(path);
	return image;
}


static GdkPixbuf*
svg_pixbuf_new(const char* filename, int size)
{
	char* path = g_strdup_printf("%s/%s", config->svgdir, filename);

	GdkPixbuf* pixbuf = rsvg_pixbuf_from_file_at_size(path, size, size, NULL);
	if(!pixbuf) gwarn("svg load failed: %s", path);
	g_free(path);
	return pixbuf;
}


#if 0
static void
create_bitmap_and_mask_from_xpm (GdkBitmap** bitmap, GdkBitmap** mask, gchar** xpm)
{
	int height, width, colors;
	char pixmap_buffer [(32 * 32)/8];
	char mask_buffer [(32 * 32)/8];
	int x, y, pix;
	int transparent_color, black_color;

	sscanf (xpm [0], "%d %d %d %d", &height, &width, &colors, &pix);

	g_assert (height == 32);
	g_assert (width == 32);
	g_assert (colors == 3);

	transparent_color = ' ';
	black_color = '.';

	for (y = 0; y < 32; y++) {
		for (x = 0; x < 32;) {
			char value = 0, maskv = 0;

			for (pix = 0; pix < 8; pix++, x++) {
				if (xpm [4+y][x] != transparent_color) {
					maskv |= 1 << pix;

					if (xpm [4+y][x] != black_color) {
						value |= 1 << pix;
					}
				}
			}

			pixmap_buffer [(y * 4 + x/8) - 1] = value;
			mask_buffer [(y * 4 + x/8) - 1] = maskv;
		}
	}

	*bitmap = gdk_bitmap_create_from_data (NULL, pixmap_buffer, 32, 32);
	*mask = gdk_bitmap_create_from_data (NULL, mask_buffer, 32, 32);
}
#endif


void
create_cursors (void)
{
	GdkBitmap* bitmap;
	GdkBitmap* mask;
	GdkColor white = {0, 0xffff, 0xffff, 0xffff};
	GdkColor black = {0, 0x0000, 0x0000, 0x0000};

	app->cursors[CURSOR_MOVE] = gdk_cursor_new (GDK_FLEUR);
	app->cursors[CURSOR_PENCIL] = gdk_cursor_new (GDK_PENCIL);
	app->cursors[CURSOR_CROSSHAIR] = gdk_cursor_new (GDK_CROSSHAIR);
	//TODO try this one too:
	//sweep_cursors[CURSOR_CROSSHAIR] = gdk_cursor_new (GDK_XTERM);
	//app->cursors[CURSOR_MAG] = gdk_cursor_new (GDK_CIRCLE);
	app->cursors[CURSOR_HAND1] = gdk_cursor_new (GDK_HAND1);
	app->cursors[CURSOR_LEFT_PTR] = gdk_cursor_new (GDK_LEFT_PTR);
	app->cursors[CURSOR_WATCH] = gdk_cursor_new (GDK_WATCH);
	app->cursors[CURSOR_H_DOUBLE_ARROW] = gdk_cursor_new (GDK_SB_H_DOUBLE_ARROW);
	app->cursors[CURSOR_V_DOUBLE_ARROW] = gdk_cursor_new (GDK_SB_V_DOUBLE_ARROW);

	//create_bitmap_and_mask_from_xpm (&bitmap, &mask, horiz_xpm);
	//sweep_cursors[SWEEP_CURSOR_HORIZ] = gdk_cursor_new_from_pixmap (bitmap, mask, &white, &black, 8, 8);

	//create_bitmap_and_mask_from_xpm (&bitmap, &mask, horiz_plus_xpm);
	//sweep_cursors[SWEEP_CURSOR_HORIZ_PLUS] = gdk_cursor_new_from_pixmap (bitmap, mask, &white, &black, 8, 8);

	//create_bitmap_and_mask_from_xpm (&bitmap, &mask, horiz_minus_xpm);
	//sweep_cursors[SWEEP_CURSOR_HORIZ_MINUS] = gdk_cursor_new_from_pixmap (bitmap, mask, &white, &black, 8, 8);


	bitmap = gdk_bitmap_create_from_data (NULL, (const gchar *) scissors_bits, scissors_width, scissors_height);
	app->cursors[CURSOR_SCISSORS] = gdk_cursor_new_from_pixmap (bitmap, bitmap, &black, &white, cursor_openhand_x_hot, cursor_openhand_y_hot);

	void new_cursor(int i, const gchar* bits, const gchar* mask_bits, int width, int height, int x_hot, int y_hot)
	{
		GdkBitmap* bitmap = gdk_bitmap_create_from_data (NULL, (const gchar *)bits, width, height);
		GdkBitmap* mask = gdk_bitmap_create_from_data (NULL, (const gchar *)mask_bits, width, height);
		app->cursors[i] = gdk_cursor_new_from_pixmap (bitmap, mask, &black, &white, x_hot, y_hot);
	}

	//gimp zoom - too big?
	//new_cursor(CURSOR_ZOOM_IN, (const gchar*)cursor_zoom_bits, (const gchar*)cursor_zoom_mask_bits, cursor_zoom_width, cursor_zoom_height, cursor_zoom_x_hot, cursor_zoom_y_hot);

	new_cursor(CURSOR_ZOOM_IN, (const gchar*)zoom_in_bits, (const gchar*)zoom_in_mask_bits, zoom_in_width, zoom_in_height, zoom_in_x_hot, zoom_in_y_hot);

	new_cursor(CURSOR_HAND_OPEN, (const gchar*)cursor_openhand_bits, (const gchar*)cursor_openhand_mask_bits, cursor_openhand_mask_width, cursor_openhand_height, cursor_openhand_x_hot, cursor_openhand_y_hot);

	bitmap = gdk_bitmap_create_from_data (NULL, (const gchar *) cursor_closedhand_bits, cursor_closedhand_width, cursor_closedhand_height);
	mask = gdk_bitmap_create_from_data (NULL, (const gchar *) cursor_closedhand_mask_bits, cursor_closedhand_mask_width, cursor_closedhand_mask_height);
	app->cursors[CURSOR_HAND_CLOSE] = gdk_cursor_new_from_pixmap (bitmap, mask, &black, &white, cursor_closedhand_x_hot, cursor_closedhand_y_hot);
}

