/**
* +----------------------------------------------------------------------+
* | This file is part of the Ayyi project. http://www.ayyi.org           |
* | copyright (C) 2010-2017 Tim Orford <tim@orford.org>                  |
* +----------------------------------------------------------------------+
* | This program is free software; you can redistribute it and/or modify |
* | it under the terms of the GNU General Public License version 3       |
* | as published by the Free Software Foundation.                        |
* +----------------------------------------------------------------------+
*
*/

/*
 *   Tracks in GlCanvas
 *   ------------------
 *
 *   currently we have:
 *     - a hashtable of GlTrack
 *        purpose:
 *          - store GlCanvas related track information
 *     - a list of TrackControlTrackActor (currently private)
 *        purpose:
 *          - scene graph object (required)
 *        - is there any reason actor details should be private? no obvious reason
 *        - there is a possible naming/conceptual issue: TrackControlTrackActor vs TrackActor - should the TrackControlTrackActor be used for the whole track, not just the trackcontrol part?
 *
 *    requirements:
 *      - must be iterable
 *      - must not cause creation-time problems
 *
 *    >> conclusion: now that we have a list of TrackControlTrackActor, it is likely that the hastable of GlTrack is redundant. move the fbo cache to the actor.
 */
#define __gl_canvas_priv__
#include "global.h"
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <GL/gl.h>
#include <gtk/gtk.h>
#include "agl/utils.h"
#include "agl/ext.h"
#include "agl/fbo.h"
#include "agl/pango_render.h"
#include "waveform/actor.h"
#include "model/ayyi_model.h"
#include "song.h"
#include "io.h"
#include "support.h"
#include "arrange.h"
#include "arrange/gl_canvas.h"
#include "arrange/shader.h"
#include "arrange/gl_track_control.h"
#include "arrange/part_renamer.h"
#include "arrange/gl_button.h"

#define USE_VBO
#ifdef USE_VBO
GLuint indices[] = {0, 1, 2, 3};
GLuint index_buffer_id = 0;
#endif

#define BG_OPACITY 0.75

typedef void (*ButtonAction)   (AGlActor*, gpointer);
typedef bool (*ButtonGetState) (AGlActor*, gpointer);

extern AGlActor*     gl_button (int, ButtonAction, ButtonGetState, gpointer);
extern AGlShaderText meter_text;

typedef enum {
	OBJ_RESIZE = 0,
	OBJ_DRAG,
	OBJ_MAX
} Objs;

typedef struct
{
	AGlActor*    pressed;
	struct {
		int      y;
	}            orig;
	struct {
		ArrTrk*  atr;
	}            new;
} Drag;
static Drag drag;

static ViewObjectClass drag_class = {0,};
static ViewObjectClass resize_class = {
	.cursor = CURSOR_V_DOUBLE_ARROW,
};

static void track_control_set_state(AGlActor*);
static void gl_track_control_track_set_text(TrackControlTrackActor*);

static AGl* agl = NULL;


static void
track_control_set_size(AGlActor* actor)
{
}


static void
track_control_set_state(AGlActor* actor)
{
}


static void
track_control_set_state_gl1(AGlActor* actor)
{
}


static bool
track_control_on_event(AGlActor* actor, GdkEvent* event, AGliPt xy)
{
	switch (event->type){
		case GDK_2BUTTON_PRESS:
			dbg(1, "double click");
			am_song__add_track(0, NULL, 0, NULL, NULL);
			return HANDLED;
		default:
			break;
	}
	return NOT_HANDLED;
}


typedef struct {
    AGlActor    actor;
    uint32_t    default_bg;
    uint32_t    selected_bg;
} TrackControlActor;


AGlActor*
track_control_actor_new(GtkWidget* panel)
{
	Arrange* arrange = (Arrange*)panel;

	/*
	 *  Fbo caching:
	 *
	 *  Each track is cached separately to an fbo.
	 *
	 *  The track controls are not cached with the rest of the track as there are many cases
	 *  where the controls need to be redrawn but the (potentially complex) track itself is unchanged.
	 *
	 */
	agl = agl_get_instance();

	float db_xlate(float db) { return MAX(0.0, (db + 70.0) / 70.0); }

	bool track_control_paint(AGlActor* actor)
	{
#if 0
		Arrange* arrange = actor->root->user_data;

		GdkColor default_bg;
		GdkColor selected_bg;
		widget_get_bg_colour((GtkWidget*)arrange, &default_bg, GTK_STATE_NORMAL);
		widget_get_bg_colour((GtkWidget*)arrange, &selected_bg, GTK_STATE_SELECTED);
		((TrackControlActor*)actor)->default_bg = color_gdk_to_rgba(&default_bg);
		((TrackControlActor*)actor)->selected_bg = color_gdk_to_rgba(&selected_bg);
#endif
		char font_string[64];

		get_font_string(font_string, -3);
		agl_set_font_string(font_string);

		return true;
	}

	void resize_xy_to_val(ViewObject* obj, AGliPt xy, AGliPt offset)
	{
		g_return_if_fail(obj);

		am_object_val(obj->model).i = xy.y;

		TrackControlTrackActor* ta = obj->user_data;
		Arrange* arrange = ta->track->arrange;
		ta->track->height = MAX(MIN_TRACK_HEIGHT, am_object_val(obj->model).i / CGL->zoom.v.val.f);

		arr_on_view_changed(arrange, AM_CHANGE_ZOOM_V);
	}

#if 0
	bool track_control_track_on_event(Actor* actor, Arrange* arrange, GdkEvent* event, Pti xy)
	{
		return actor__on_viewobject_event(actor, arrange, event, xy);
	}
#endif

	void on_tracks_change(GObject* _song, TrackNum t_num1, TrackNum t_num2, AMChangeType change_type, Arrange* arrange)
	{
		if((change_type & AM_CHANGE_NAME)){
			gtk_widget_queue_draw(CGL->image);
		}

		// TODO this duplicates some code in on_view_change below
		if((change_type & AM_CHANGE_HEIGHT)){
			TrackNum t; for(t=t_num1;t<=t_num2;t++){
				ArrTrk* atr = track_list__track_by_index(arrange->priv->tracks, t);
				if(atr){
					GlTrack* gltrack = g_hash_table_lookup(CGL->tracks, atr->track);
					TrackControlTrackActor* actor = gltrack->actor;
					((AGlActor*)actor)->region.y2 = atr->height;

					int height = atr->height * arrange->vzoom;
					((VOActor*)actor)->objs[OBJ_RESIZE].region = (iRegion){0, height - 2, 1024, height};
					((VOActor*)actor)->objs[OBJ_DRAG].region.y2 = height - 3;
				}
			}
		}

		if((change_type & (AM_CHANGE_MUTE | AM_CHANGE_SOLO | AM_CHANGE_ARMED | AM_CHANGE_NAME | AM_CHANGE_OUTPUT))){
			TrackNum t; for(t=t_num1;t<=t_num2;t++){
				ArrTrk* atr = track_list__track_by_index(arrange->priv->tracks, t);
				if(atr){
					GlTrack* gltrack = g_hash_table_lookup(CGL->tracks, atr->track);
					TrackControlTrackActor* actor = gltrack->actor;
					GList* l = ((AGlActor*)actor)->children;
					for(;l;l=l->next){
						agl_actor__invalidate((AGlActor*)l->data);
						if(change_type & AM_CHANGE_NAME || change_type & AM_CHANGE_OUTPUT){
							gl_track_control_track_set_text(actor);
						}
					}
				}
			}
		}
	}

	void gl_tracks_on_change_by_list(GObject* _song, GList* tracks, AMChangeType change_type, AyyiPanel* originator, Arrange* arrange)
	{
		PF;
		TrackList* tl = arrange->priv->tracks;

		void rebuild_track_list(Arrange* arrange, AGlActor* actor)
		{
			// if tracks have been re-ordered the actor child list has to be rebuilt to preserve order by display

			TrackControlTrackActor* find_ta(AGlActor* actor, ArrTrk* atr)
			{
				GList* l1 = actor->children;
				for(;l1;l1=l1->next){
					TrackControlTrackActor* a = l1->data;
					if(a->track == atr) return a;
				}
				return NULL;
			}

			GList* l2 = NULL;
			TrackDispNum d; for (d=0; d<AM_MAX_TRK && tl->display[d] && tl->display[d]->track->type; d++){
				ArrTrk* atr = tl->display[d];
				TrackControlTrackActor* ta = find_ta(actor, atr);
				l2 = g_list_append(l2, ta);
			}
			g_list_free(actor->children);
			actor->children = l2;
		}

		GList* l = tracks;
		for(;l;l=l->next){
			AMTrack* track = l->data;

			if((change_type & AM_CHANGE_TRACK)){ // track reordering
				ArrTrk* atr = track_list__lookup_track(tl, track);
				GlTrack* gltrack = g_hash_table_lookup(CGL->tracks, atr->track);
				TrackControlTrackActor* actor = gltrack->actor;

				((AGlActor*)actor)->region.y1 = atr->y * CGL->zoom.v.val.f;
				((AGlActor*)actor)->region.y2 = (atr->y + atr->height) * CGL->zoom.v.val.f;

				rebuild_track_list(arrange, ((AGlActor*)actor)->parent);
			}
		}
		gtk_widget_queue_draw(CGL->image);
	}

	AGlActor* track_control_track_actor_new(Arrange* arrange, ArrTrk* track)
	{
		void _set_height(AMObject* obj, int property, AMVal value, AyyiHandler2 _, gpointer user_data)
		{
			AMTrack* track = ((TrackControlTrackActor*)obj->user_data)->track->track;
			arr_track_set_height(NULL, track, value.i);
		}

		TrackControlTrackActor* ta = g_new0(TrackControlTrackActor, 1); // ta is used in initialisation so cannot combine
		*ta = (TrackControlTrackActor){
			.actor = {
				.parent = {
					.name = g_strdup_printf("%s", track->track->name),
				},
				.n_objs = OBJ_MAX
			},
			.height = {
				.set = _set_height,
				.user_data = ta,
			}
		};
		AGlActor* a = (AGlActor*)ta;

		int height = track->height * arrange->vzoom;

		void resize_change_done(ViewObject* obj, ArrTrk* trk, AyyiSongPos* pos, AyyiHandler2 h, gpointer u)
		{
			AMVal val;
			songpos_ayyi2gui(&val.pos, pos);
			obj->model->set(obj->model, 0, val, h, u);
		}

		resize_class.change_done = resize_change_done; // dont use. will be removed. see _set_height instead.
		resize_class.xy_to_val = resize_xy_to_val;

		view_object_init(&resize_class, &((VOActor*)ta)->objs[OBJ_RESIZE], (ViewObject){
			.model = &ta->height,
			.region = (iRegion){0, height - 2, 1024, height},
			.user_data = ta,
		});

		{
			void _drag_motion(ViewObject* obj, AGliPt xy, AGliPt offset)
			{
																				// TODO why is this called when mouse apparently has not moved ?
				TrackControlTrackActor* ta = obj->user_data;
				AGlActor* actor = (AGlActor*)ta;
				Arrange* arrange = ta->track->arrange;

				if(!drag.pressed){
					drag.pressed = actor;
					drag.orig.y = xy.y;
				}

				if(ABS(xy.y - drag.orig.y) > 4){
					set_cursor(arrange->canvas->widget->window, CURSOR_HAND_OPEN);
					TrackDispNum track_d;
					arrange->canvas->pick(arrange, xy.x, /*actor->region.y1 +*/ xy.y, &track_d, NULL);
					if(track_d > -1){
						ArrTrk* atr = track_list__track_by_display_index(arrange->priv->tracks, track_d);
						if(atr != ta->track && atr != drag.new.atr){
							dbg(0, "track changed %i: (%s)", arr_t_to_d(arrange->priv->tracks, atr->track->track_num), atr->track->name);
							drag.new.atr = atr;
						}
					}
				}
			}

			void _set_drag(AMObject* obj, int property, AMVal value, AyyiHandler2 _, gpointer user_data)
			{
				PF;
				g_return_if_fail(drag.pressed);

				TrackControlTrackActor* ta = obj->user_data;
				Arrange* arrange = ta->track->arrange;
				TrackList* tl = arrange->priv->tracks;
				ArrTrk* atr = drag.new.atr;
				if(atr && (atr != ta->track)){
					dbg(0, ">> new track: %i (%s)", atr->track->track_num, atr->track->name);

					arr_track_move_to_pos(arrange, track_list__get_display_num(tl, ta->track->track), track_list__get_display_num(tl, atr->track));
				}
				drag.pressed = NULL;
				drag.new.atr = NULL;
				set_cursor(arrange->canvas->widget->window, CURSOR_NORMAL);
			}

			drag_class.xy_to_val = _drag_motion;

			am_object_init(&ta->drag, 1);
			ta->drag.set = _set_drag;
			ta->drag.user_data = ta;

			view_object_init(&drag_class, &((VOActor*)ta)->objs[OBJ_DRAG], (ViewObject){
				.model = &ta->drag,
				.region = (iRegion){0, 0, 1024, height - 3},
				.user_data = ta,
			});
		}

#ifdef AGL_ACTOR_RENDER_CACHE
		#define INVALIDATE agl_actor__invalidate(actor)
#else
		#define INVALIDATE ;
#endif
		void track_arm_toggle  (AGlActor* actor, gpointer at) { am_track__arm(((ArrTrk*)at)->track, !am_track__is_armed(((ArrTrk*)at)->track), NULL, NULL); INVALIDATE; }
		void track_solo_toggle (AGlActor* actor, gpointer at) { AMTrack* tr = ((ArrTrk*)at)->track; am_track__solo(tr, !am_track__is_solod(tr), NULL, NULL);INVALIDATE;  }
		void track_mute_toggle (AGlActor* actor, gpointer at) { am_track__mute(((ArrTrk*)at)->track, !am_track__is_muted(((ArrTrk*)at)->track), NULL, NULL);INVALIDATE;  }

		bool get_armd_state (AGlActor* a, gpointer at) { return am_track__is_armed(((ArrTrk*)at)->track); }
		bool get_mute_state (AGlActor* a, gpointer at) { return am_track__is_muted(((ArrTrk*)at)->track); }
		bool get_solo_state (AGlActor* a, gpointer at) { return am_track__is_solod(((ArrTrk*)at)->track); }

		AGlActor* button = agl_actor__add_child(a, gl_button(TEXTURE_ICON_REC, track_arm_toggle, get_armd_state, track));
		button->region = (AGliRegion){0, 0, 16, height};

		button = agl_actor__add_child(a, gl_button(TEXTURE_ICON_SOLO, track_solo_toggle, get_solo_state, track));
		button->region = (AGliRegion){16, 0, 16 + 16, height};

		button = agl_actor__add_child(a, gl_button(TEXTURE_ICON_MUTE, track_mute_toggle, get_mute_state, track));
		button->region = (AGliRegion){32, 0, 32 + 16, height};

		bool track_control_track_paint(AGlActor* actor)
		{
			Arrange* arrange = actor->root->user_data;
			TrackControlActor* parent = (TrackControlActor*)actor->parent;
			TrackControlTrackActor* ta = (TrackControlTrackActor*)actor;
			ArrTrk* at = ta->track;
			AMTrack* track = at->track;

			double total_width = agl_actor__width(actor);
			double width = MAX(0, (total_width - 3 * 16) / 2);
			double widths[] = {width * 5 / 4, width * 3 / 4};

			double x1 = 16.0 * 3.0;
			double x2 = x1 + widths[0];

			bool track_is_selected = arr_track_is_selected(arrange, track);

			int tcolour = am_track__get_colour(track);
			uint32_t colour = track_is_selected
				? parent->selected_bg
				: tcolour > -1 ? song->palette[tcolour] : parent->default_bg;

			// the button background colour is determined by its parent, so it is set here.
			GList* l = ((AGlActor*)ta)->children;
			for(;l;l=l->next){
				AGlActor* b = l->data;
				((ButtonActor*)b)->bg_colour = (colour & 0xffffff00) + ((int)(0xff * BG_OPACITY));
			}

			if(!(actor->region.x2 > x1)) return true;

			double height = at->height * CGL->zoom.v.val.f;

			// track control background
			x1 = 16.0 * 3.0;
			x2 = 16.0 * 3.0 + widths[0] + widths[1];
			if(agl->use_shaders){
#if 0
				agl->shaders.plain->uniform.colour = (ta->animatable.val.f > 0.01)
					? colour_mix_rgba(colour, colour_lighter_rgba(colour, 16), ta->animatable.val.f)
					: colour;
				agl->shaders.plain->uniform.colour = (agl->shaders.plain->uniform.colour & 0xffffff00) + 0xff * BG_OPACITY;
#else
				// modulating alpha instead of colour has the advantage that it works with white
				// but the effect is stronger on some colours more than others
				int alpha = 0xff * (BG_OPACITY + (1 - BG_OPACITY) * ta->animatable.val.f / 2);
				agl->shaders.plain->uniform.colour = (colour & 0xffffff00) + alpha;
#endif
				agl_use_program((AGlShader*)agl->shaders.plain);

																		// TODO does this mess up subequent objects?
				glBlendFuncSeparate(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA, GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
#ifdef USE_VBO
				glEnableClientState(GL_VERTEX_ARRAY);
				glBindBufferARB(GL_ARRAY_BUFFER_ARB, ta->vbo_id);
				glVertexPointer(2, GL_FLOAT, 0, 0); // load the vbo data into the gpu starting at 0

				glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER, index_buffer_id);
				glDrawElements(GL_QUADS, 4, GL_UNSIGNED_INT, 0);

				glDisableClientState(GL_VERTEX_ARRAY);
				glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
#else
				glRectf(x1, 0.0, x2, height);
#endif
			}else{
				agl_enable(AGL_ENABLE_BLEND /* !AGL_ENABLE_TEXTURE_2D */);

				AGlColourFloat cf;
				colour_rgba_to_float(&cf, track_is_selected ? parent->selected_bg : parent->default_bg);
				glColor4f(cf.r, cf.g, cf.b, BG_OPACITY);

				glRectf(x1, 0.0, x2, height);
			}

			// track name

			x2 = x1 + widths[0];

			agl_enable_stencil(x1, 0.0, widths[0], height);

			// we cannot vertically centre using layout height because
			// the text is not centred in the layout.
			int bl = PANGO_PIXELS(pango_layout_get_baseline(ta->layout));
			int yb = (height + 16) / 2 - 1;
			PangoRectangle logical_rect;
			pango_layout_get_pixel_extents(ta->layout, NULL, &logical_rect);
			// changed to using text colour instead of fg colour but it does not look as good - contrast too high
			// - TODO is this true generally?
			#define FIX 1
#if 0
			agl_print_layout2(x1 + 8, yb - bl - logical_rect.y - FIX, 0.0, get_style_text_color_rgba(GTK_STATE_NORMAL), ta->layout);
#else
			uint32_t textcolour = get_style_text_color_rgba(GTK_STATE_NORMAL);
			pango_gl_render_layout(ta->layout, x1 + 8, yb - bl - logical_rect.y - FIX, 0.0, (Colour32*)&textcolour, 0);
#endif
			agl_disable_stencil();

			// output name
			{
				x1 = x2;
				x2 += widths[1];

				agl_enable_stencil(x1, 0.0, widths[1], height);

				pango_layout_get_pixel_extents(ta->output_layout, NULL, &logical_rect);
				bl = PANGO_PIXELS(pango_layout_get_baseline(ta->output_layout));

				//agl_print_layout2(x1 + 8, yb - bl - logical_rect.y, 0.0, get_style_text_color_rgba(GTK_STATE_NORMAL), ta->output_layout);
				//textcolour = get_style_text_color_rgba(GTK_STATE_NORMAL);
				pango_gl_render_layout(ta->output_layout, x1 + 8, yb - bl - logical_rect.y, 0.0, (Colour32*)&textcolour, 0);
				agl_disable_stencil();
			}

			if(TRK_IS_MIDI(track) && track_is_selected){
				ArrTrackMidi* atm = (ArrTrackMidi*)at;
				if(agl->use_shaders){
					agl->shaders.plain->uniform.colour = 0xffffffff;
					agl_use_program((AGlShader*)agl->shaders.plain);
				}
				agl_rect_((AGlRect){x2 - 48, 0, 48, height});

				int kh = atm->note_height * 12 / 7;
				int bottom_note = arr_track_get_note_num_from_y(atm, height - 1);
				int offset = note_is_black(bottom_note) ? kh / 2 : 0;

				agl->shaders.plain->uniform.colour = 0x000000ff;
				agl->shaders.plain->shader.set_uniforms_();
				int height = at->height * arrange->vzoom;
				int n_notes = arr_track_midi_get_n_notes(atm);
				int i; for(i=0;i<n_notes;i++){
					glRecti(x2 - 48, height -offset - kh * i, x2, height -offset - kh * i + 1);
				}

				int black[] = {0, 1, 3, 4, 5};
				//              C  C# D  Eb E  F  F# G  Ab A  Bb B
				int start[] = { 0, 0, 0, 1, 1, 2, 2, 2, 3, 4, 4, 4}; // for each note num, the note offset to the start of the black keys sequence
// TODO check these for black notes
				int st   [] = {-1, 0, 0, 1, 1, 2, 3, 3, 4, 4, 5, 5}; // increasing value moves black notes up
				int s = start[bottom_note%12];
				for(i=0;i<((n_notes * 5) / 7);i++){
					#define OCTAVE() 7 * ((i + s) / 5)
					float y = height - offset + kh * st[bottom_note%12] - kh * OCTAVE() + /*s * kh*/ - kh * black[(i + s) % 5];
					float w = ((float)kh) / 3;
					glRectf(x2 - 48, y - w, x2 - 18, y + w);
				}

				//         C  C# D  Eb E  F  F# G  Ab A  Bb B
				int p[] = {0, 0, 1, 1, 2, 3, 3, 4, 4, 5, 5, 6};
				if(atm->hover_note){
					agl->shaders.plain->uniform.colour = 0x00000033;
					agl->shaders.plain->shader.set_uniforms_();
					if(!note_is_black(atm->hover_note)){
						int y = height - kh * p[atm->hover_note - bottom_note];
						glRecti(x2 - 48, y, x2, y - kh);
					}
				}

#if 0
				agl_print(x2 -16, arr_track_note_pos_y(at, &(MidiNote){.note=60}) + 1, 0.0, 0x000000ff, "C4");
#else
				int n_white_notes_difference(int a, int b)
				{
					//         C  C# D  Eb E  F  F# G  Ab A  Bb B
					int l[] = {0, 0, 1, 1, 2, 3, 3, 4, 4, 5, 5, 6};
					return 7 * (a/12 - b/12) + l[a%12] - l[b%12];
				}
				int yy = height -kh + offset - n_white_notes_difference(60, bottom_note) * kh;
				agl_print(x2 - 14, yy + 2, 0.0, 0x000000ff, "C4");
#endif
			}
			return true;
		}

		void gl_track_control_track_free(AGlActor* actor)
		{
			TrackControlTrackActor* ta = (TrackControlTrackActor*)actor;

			g_object_unref(ta->layout);
			g_object_unref(ta->output_layout);
			glDeleteBuffersARB(1, &ta->vbo_id);
#ifdef AGL_DEBUG_ACTOR
			g_free(actor->name);
#endif
			g_free(actor);
		}

		void gl_track_control_track_init(AGlActor* actor)
		{
			TrackControlTrackActor* ta = (TrackControlTrackActor*)actor;

#ifdef USE_VBO
			glGenBuffersARB(1, &ta->vbo_id);

			if(!index_buffer_id){
				glGenBuffersARB(1, &index_buffer_id);
				glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER, index_buffer_id);
				glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER, 4 * sizeof(GLuint), indices, GL_STATIC_DRAW);
			}
#endif

#ifdef AGL_ACTOR_RENDER_CACHE
											// TODO perhaps this should be done automatically the scene graph?
			actor->fbo = agl_fbo_new(agl_actor__width(actor), agl_actor__height(actor), 0, AGL_FBO_HAS_STENCIL);
			actor->cache.enabled = true;
#endif

			PangoGlRendererClass* PGRC = g_type_class_peek(PANGO_TYPE_GL_RENDERER);
			PangoFontDescription* font_desc = pango_font_description_new();
			pango_font_description_set_family(font_desc, "Sans");

			{
				ta->layout = pango_layout_new (((PangoGlRendererClass*)g_type_class_peek(PANGO_TYPE_GL_RENDERER))->context);

				pango_font_description_set_size(font_desc, 10 * PANGO_SCALE);
				pango_font_description_set_weight(font_desc, PANGO_WEIGHT_SEMIBOLD);

				/* TODO
				// for some reason there seems to be an issue with pixmap fonts
				if(!font_is_scalable(PGRC, pango_font_description_get_family(font_desc))){
					pango_font_description_set_family(font_desc, family);
				}
				*/
				pango_layout_set_font_description(ta->layout, font_desc);
			}
			{
				ta->output_layout = pango_layout_new(PGRC->context);

				pango_font_description_set_size(font_desc, 7 * PANGO_SCALE);
				pango_font_description_set_weight(font_desc, PANGO_WEIGHT_NORMAL);
				pango_layout_set_font_description(ta->output_layout, font_desc);

			}

			pango_font_description_free (font_desc);
			gl_track_control_track_set_text(ta);

			agl_actor__set_size(actor);
		}

		void track_control_track_set_size(AGlActor* actor)
		{
			// note that the y position is not set here as it depends on the preceding tracks.
			// -in particular, at->y is not valid during track resize operations.

			TrackControlTrackActor* ta = (TrackControlTrackActor*)actor;
			ArrTrk* at = ta->track;
			Arrange* arrange = at->arrange;

			int top = at->y * CGL->zoom.v.val.f;
			int h = at->height * CGL->zoom.v.val.f;

			actor->region = (AGliRegion){0, top, actor->parent->region.x2, top + h};

			{ // main button
				double width = MAX(0, (agl_actor__width(actor) - 3 * 16) / 2);
				double widths[] = {width * 5 / 4, width * 3 / 4};

				float x1 = 16.0 * 3.0;
				float x2 = 16.0 * 3.0 + widths[0] + widths[1];
				float vertices[] = {
					x1, 0.0,
					x2, 0.0,
					x2, h,
					x1, h,
				};
				glBindBufferARB(GL_ARRAY_BUFFER_ARB, ta->vbo_id);
				glBufferDataARB(GL_ARRAY_BUFFER_ARB, G_N_ELEMENTS(vertices) * sizeof(float), vertices, GL_STATIC_DRAW_ARB);
			}

			((VOActor*)ta)->objs[OBJ_RESIZE].region = (iRegion){0, h - 2, 1024, h};
			((VOActor*)ta)->objs[OBJ_DRAG].region.y2 = h - 3;

			GList* k = actor->children;
			for(;k;k=k->next){
				AGlActor* button = k->data;
				button->region.y2 = (ta->track->height * CGL->zoom.v.val.f);
			}
		}

		bool track_on_event(AGlActor* actor, GdkEvent* event, AGliPt xy)
		{
			TrackControlTrackActor* ta = (TrackControlTrackActor*)actor;
			ArrTrk* at = ta->track;
			Arrange* arrange = at->arrange;

			switch (event->type){
				case GDK_ENTER_NOTIFY:
					ta->hover_opacity = 1.0;
					agl_actor__start_transition(actor, g_list_append(NULL, &ta->animatable), NULL, NULL);
					break;
				case GDK_LEAVE_NOTIFY:
					ta->hover_opacity = 0.0;
					agl_actor__start_transition(actor, g_list_append(NULL, &ta->animatable), NULL, NULL);
					break;
				case GDK_SCROLL:
					if(TRK_IS_MIDI(at->track) && arr_track_is_selected(arrange, at->track)){
						GdkEventScroll* ev = (GdkEventScroll*)event;

						GdkModifierType state; gdk_event_get_state(event, &state);
						if(state & GDK_SHIFT_MASK){
							((ArrTrackMidi*)at)->note_height = CLAMP(((ArrTrackMidi*)at)->note_height +
								(ev->direction == GDK_SCROLL_UP ? 1 : -1), 2, 10);
						}else{
							((ArrTrackMidi*)at)->centre_note = CLAMP(((ArrTrackMidi*)at)->centre_note +
								(ev->direction == GDK_SCROLL_UP ? 1 : -1), 0, 127);
						}

						agl_actor__invalidate(actor);

						GlTrack* gltrack = g_hash_table_lookup(CGL->tracks, at->track);
						if(gltrack){
							gltrack->actor->parts.cache.valid = false;
							gtk_widget_queue_draw(CGL->image);
						}
					}
					break;
				default:
					break;
			}
			if(actor__on_viewobject_event(actor, event, xy) == HANDLED) return HANDLED;

			switch (event->type){
				case GDK_BUTTON_PRESS:
					switch(event->button.button){
						case 3:
							dbg(1, "right click");
							if(xy.x < 100){
								arr_track_menu_show(arrange, at->track->track_num, (GdkEventButton*)event);
							}else{
								if(gui_song.ch_menu){
									AMChannel* ch = am_track__lookup_channel(at->track);
									output_menu_update((AyyiPanel*)arrange, ch);
									output_menu_popup();
								}
							}
							break;
						case 1:
							if(at && !arr_track_is_selected(arrange, at->track)){
								arr_track_select(arrange, at->track);
							}
							break;
						default:
							break;
					}
					return HANDLED;
				case GDK_2BUTTON_PRESS:
					dbg(1, "double click");

					void rename_done(char* new_text, gpointer track)
					{
						PF;
						am_track__set_name((AMTrack*)track, new_text, NULL, NULL);
					}

					AGliPt o = agl_actor__find_offset_2(actor);
					Pti pos = {16 * 3 + 2, o.y + agl_actor__height(actor) / 2 - 10};
					track_rename_popup(&arrange->panel, arrange->canvas->widget, pos, FALSE, at->track, rename_done, at->track);

					return HANDLED;
				default:
					break;
			}
			return NOT_HANDLED;
		}

		ta->track = track;
		am_object_init(&ta->height, 1);

		a->paint = track_control_track_paint;
		a->on_event = track_on_event;
		a->init = gl_track_control_track_init;
		a->free = gl_track_control_track_free;
		a->set_size = track_control_track_set_size;

		ta->animatable = (WfAnimatable){
			.model_val.f = &ta->hover_opacity,
			.start_val.f = 0.0,
			.val.f       = 0.0,
			.type        = WF_FLOAT
		};

		GlTrack* gltrack = g_hash_table_lookup(CGL->tracks, track->track);
		if(gltrack){
			gltrack->actor = ta;
		}
		else gwarn("unable to set gltrack->actor");

		return a;
	}

	void _on_track_add(GObject* _song, AMTrack* tr, Arrange* arrange)
	{
		ArrTrk* at = track_list__lookup_track(arrange->priv->tracks, tr);
		g_return_if_fail(at);

		agl_actor__add_child(CGL->actors[ACTOR_TYPE_TC], track_control_track_actor_new(arrange, at));

		gtk_widget_queue_draw(CGL->image);
	}

	void _on_track_delete(GObject* _tracks, AMTrack* track, Arrange* arrange)
	{
		PF;

		AGlActor* find_track_actor(Arrange* arrange, ArrTrk* at)
		{
			AGlActor* tc = CGL->actors[ACTOR_TYPE_TC];
			GList* l = tc->children;
			for(;l;l=l->next){
				TrackControlTrackActor* ta = (TrackControlTrackActor*)l->data;
				if(ta->track == at) return (AGlActor*)ta;
			}
			gwarn("not found");
			return NULL;
		}

		ArrTrk* atr = track_list__lookup_track(arrange->priv->tracks, track);
		g_return_if_fail(atr);
		AGlActor* child = find_track_actor(arrange, atr);
		g_return_if_fail(child);
		agl_actor__remove_child(CGL->actors[ACTOR_TYPE_TC], child);
		dbg(1, "actor removed ok");

		track_control_actor_on_view_change(arrange, AM_CHANGE_POS_Y);
	}

	void gl_track_control_on_selection_change(GObject* am_song, AyyiPanel* sender, Arrange* arrange)
	{
		AGlActor* tc = CGL->actors[ACTOR_TYPE_TC];

#ifdef AGL_ACTOR_RENDER_CACHE
		GList* l = tc->children;
		for(;l;l=l->next){
			((AGlActor*)l->data)->cache.valid = false;
		}
#endif
		gtk_widget_queue_draw(tc->root->gl.gdk.widget);
	}

	void gl_track_control_free(AGlActor* actor)
	{
		Arrange* arrange = actor->root->user_data;

		int n, i = 0;
		#define song_handler_disconnect(FN) if((n = g_signal_handlers_disconnect_matched (song, G_SIGNAL_MATCH_FUNC | G_SIGNAL_MATCH_DATA, 0, 0, NULL, FN, arrange)) != 1){ gwarn("song handler disconnection (%i) n=%i", i, n); } i++;
		song_handler_disconnect(on_tracks_change);
		song_handler_disconnect(gl_tracks_on_change_by_list);
		#undef song_handler_disconnect

		#define tracks_handler_disconnect(FN) if(g_signal_handlers_disconnect_matched (am_tracks, G_SIGNAL_MATCH_FUNC | G_SIGNAL_MATCH_DATA, 0, 0, NULL, FN, arrange) != 1) gwarn("track handler disconnection");
		tracks_handler_disconnect(_on_track_add);
		tracks_handler_disconnect(_on_track_delete);
		tracks_handler_disconnect(gl_track_control_on_selection_change);
		#undef tracks_handler_disconnect

		g_free(actor);
	}

	void gl_track_control_init(AGlActor* actor)
	{
		((TrackControlActor*)actor)->default_bg = get_style_bg_color_rgba(GTK_STATE_NORMAL);
		((TrackControlActor*)actor)->selected_bg = get_style_bg_color_rgba(GTK_STATE_SELECTED);
	}

	TrackControlActor* a = AGL_NEW(TrackControlActor,
		.actor = {
			.name      = "TrackControl",
			.init      = gl_track_control_init,
			.set_size  = track_control_set_size,
			.set_state = agl->use_shaders ? track_control_set_state : track_control_set_state_gl1,
			.paint     = track_control_paint,
			.on_event  = track_control_on_event,
			.free      = gl_track_control_free,

			.region    = (AGliRegion){0, arrange->ruler_height, arrange->tc_width.val.i, 2048}
		}
	);

	if(((AyyiPanel*)arrange)->model_is_loaded){
		AMTrack* track = NULL;
		ArrTrk* at = NULL;
		while((at = track_list__next_visible(arrange->priv->tracks, at ? at->track : NULL)) && (track = at->track)){
			agl_actor__add_child((AGlActor*)a, track_control_track_actor_new(arrange, at));
		}
		gtk_widget_queue_draw(CGL->image);
	}

	am_song__connect("tracks-change", G_CALLBACK(on_tracks_change), arrange);
	am_song__connect("tracks-change-by-list", G_CALLBACK(gl_tracks_on_change_by_list), arrange);
	g_signal_connect(song->tracks, "add", G_CALLBACK(_on_track_add), arrange);
	g_signal_connect(song->tracks, "delete", G_CALLBACK(_on_track_delete), arrange);
	g_signal_connect(song->tracks, "selection-change", G_CALLBACK(gl_track_control_on_selection_change), arrange);

	return (AGlActor*)a;
}


void
track_control_actor_on_view_change(Arrange* arrange, AMChangeType change)
{
	if(change & AM_CHANGE_ZOOM_V || change & AM_CHANGE_POS_Y){
		AGlActor* tc = CGL->actors[ACTOR_TYPE_TC];
		if(tc){
			agl_actor__set_size(tc);

			int y = 0;
			GList* l = tc->children;
			for(;l;l=l->next){
				TrackControlTrackActor* ta = (TrackControlTrackActor*)l->data;

				int h = ta->track->height * CGL->zoom.v.val.f;
				// note ta->track->y is not used here in order to handle transient sizes during track resize.
				((AGlActor*)ta)->region.y1 = y; // note no scroll offset. scrolling done by container.
				((AGlActor*)ta)->region.y2 = ((AGlActor*)ta)->region.y1 + h;

				y += h;
			}
		}
	}
}


static void
gl_track_control_track_set_text(TrackControlTrackActor* ta)
{
	char str[256];

	pango_layout_set_text(ta->layout, ta->track->track->name, -1);
	pango_layout_set_text(ta->output_layout, am_track__get_channel_label(ta->track->track, str), -1);
}

