/**
* +----------------------------------------------------------------------+
* | This file is part of the Ayyi project. http://www.ayyi.org           |
* | copyright (C) 2004-2017 Tim Orford <tim@orford.org>                  |
* +----------------------------------------------------------------------+
* | This program is free software; you can redistribute it and/or modify |
* | it under the terms of the GNU General Public License version 3       |
* | as published by the Free Software Foundation.                        |
* +----------------------------------------------------------------------+
*
*/

typedef struct { float x, y, w, h; } Rect;


/*
 *  The Tracks actor displays the contents of each track. There is one child actor for each track.
 *  The track controls are not part of this actor
 *  Render caching of tracks is done locally and not by the scene graph because it pre-dates this feature.
 */
static AGlActor*
arr_gl_tracks(GtkWidget* panel)
{
#ifdef USE_FBO
	void do_track_cache(TrackControlTrackActor* tcta, AMTrack* tr)
	{
		inline void calc_cache_geometry(Arrange* arrange, AMTrack* tr, int w, iRange track, int* start)
		{
			// the cache is bigger than the viewport and is positioned such that the current viewport is at the caches centre
			// (also taking into account the start and end of the content).

			WfViewPort viewport; arr_gl_get_viewport(arrange, &viewport);

			if(w < AGL_MAX_FBO_WIDTH){
				// whole track fits in the cache.
				*start = track.start;
			}else{
				int viewport_width = viewport.right - viewport.left;
				g_return_if_fail(viewport_width <= AGL_MAX_FBO_WIDTH);
				int excess = MAX(0, w - viewport_width); // <----
				int request_start = MAX(0, (int)viewport.left - track.start);
				int request_end = MAX(0, track.end - viewport.right);
				excess = MIN(excess, request_start + request_end);
				//int debug_initial_excess = excess;
				int excess_end = MIN(request_end, excess / 2);
				excess -= excess_end;
				request_end -= excess_end;
				int excess_start = MIN(request_start, excess);
				excess -= excess_start;
				request_start -= excess_start;
				excess_end += excess;
				*start = MAX(track.start, -CGL->actors[ACTOR_TYPE_MAIN]->scrollable.x1 - excess_start);
				//dbg(0, "viewport_width=%i w=%i excess=%i: %i, %i x=%.1f", viewport_width, w, debug_initial_excess, excess_start, excess_end, CGL->visible.x);
			}
		}

		AGlActor* _actor = (AGlActor*)tcta;
		Arrange* arrange = _actor->root->user_data;
		GlCanvas* c = arrange->canvas->gl;
		GlTrack* gltrack = g_hash_table_lookup(c->tracks, tr);

		FilterIterator* i = am_part_iterator_new ((Filter)am_part__has_track, tr);
		AMPart* part = (AMPart*)filter_iterator_next(i);
		if(part){
			AyyiSongPos track_end_pos, track_start_pos;
			am_track__get_start_end(tr, &track_start_pos, &track_end_pos);
			iRange track = {.start = arr_gl_pos2px_(arrange, &track_start_pos), .end = arr_gl_pos2px_(arrange, &track_end_pos)};

			int w = MIN(AGL_MAX_FBO_WIDTH, track.end - track.start);

			if(tcta->parts.cache.fbo){
				int h = GL_TRACK_HEIGHT2(gltrack) + 2 * PART_TRANSITION_BORDER;
				if(agl_power_of_two(tcta->parts.cache.fbo->width) == agl_power_of_two(w) && agl_power_of_two(tcta->parts.cache.fbo->height) == h){
					agl_fbo_set_size(tcta->parts.cache.fbo, w, GL_TRACK_HEIGHT2(gltrack) + 2 * PART_TRANSITION_BORDER);
				}else{
					// some strange issues with resizing an existing fbo, so recreating...
					agl_fbo_free(tcta->parts.cache.fbo);
					tcta->parts.cache.fbo = agl_fbo_new(w, h, 0, AGL_FBO_HAS_STENCIL);
				}
			}else{
				tcta->parts.cache.fbo = agl_fbo_new(w, GL_TRACK_HEIGHT2(gltrack) + 2 * PART_TRANSITION_BORDER, 0, AGL_FBO_HAS_STENCIL);
			}

			calc_cache_geometry(arrange, tr, w, track, &tcta->parts.cache_start_px);

			AyyiSongPos range_start, range_end;
			arr_gl_px2songpos(arrange, &range_start, tcta->parts.cache_start_px);
			arr_gl_px2songpos(arrange, &range_end, tcta->parts.cache_start_px + w);
			AMTrackSegment range = {tr, &range_start, &range_end};

			FilterIterator* j = am_part_iterator_new ((Filter)am_part__in_track_segment, &range);
			AMPart* part = (AMPart*)filter_iterator_next(j);
			if(part){
				GlPart* glpart = g_hash_table_lookup(c->parts, part);
#if 0
				char bbst1[32];
				char bbst2[32];
				ayyi_pos2bbst(&range_start, bbst1);
				ayyi_pos2bbst(&range_end, bbst2);
				dbg(0, "range: %s -> %s", bbst1, bbst2);
#endif
#ifdef DEBUG
				char bbst[32];
				ayyi_pos2bbst(&track_end_pos, bbst);
				dbg(2, "cache_start=%i cache_end=%i track_end=%s w=%ipx", tcta->parts.cache_start_px, tcta->parts.cache_start_px + w, bbst, w);
#endif

				agl_draw_to_fbo(tcta->parts.cache.fbo) {
					glClearColor(0.0, 0.0, 0.0, 0.0); // background colour must be same as foreground for correct antialiasing
					glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

					do arr_gl_draw_part(arrange, glpart); while((part = (AMPart*)filter_iterator_next(j)) && (glpart = g_hash_table_lookup(c->parts, part)));

				} agl_end_draw_to_fbo;
				tcta->cache_is_empty = false;
			}else{
				dbg(1, "cache range contains no parts");
				tcta->cache_is_empty = true; // dont delete the fbo but indicate that it should not be displayed
			}

			filter_iterator_unref(j);
		}else{
			// part has no tracks
			if(tcta->parts.cache.fbo) agl_fbo_free0(tcta->parts.cache.fbo);
		}
		filter_iterator_unref(i);
		tcta->parts.cache.valid = true;
	}

	void arr_gl_draw_parts_(AGlActor* tracks, AMTrack* tr)
	{
		// render the FBO to screen

		Arrange* arrange = tracks->root->user_data;
		GlCanvas* c = arrange->canvas->gl;
		GlTrack* gltrack = g_hash_table_lookup(c->tracks, tr);
		g_return_if_fail(gltrack);
		TrackControlTrackActor* tcta = gltrack->actor;

		if(!tcta->parts.cache.valid) do_track_cache(tcta, tr);

		AGlFBO* fbo = tcta->parts.cache.fbo;
		if(!fbo || tcta->cache_is_empty) return; // empty tracks have no fbo

		Rect r = {
			.x = tcta->parts.cache_start_px,
			.y = ((AGlActor*)tcta)->region.y1 - PART_TRANSITION_BORDER,
			.w = fbo->width,
			.h = fbo->height,
		};

// TODO this test needs to be changed
		if((r.y > c->actors[ACTOR_TYPE_MAIN]->scrollable.y2) || (((AGlActor*)tcta)->region.y1 + r.h < c->actors[ACTOR_TYPE_MAIN]->scrollable.y1)) return; // outside viewport

		r.y *= (1 - c->iso_factor);

		if(agl->use_shaders){
			agl->shaders.texture->uniform.fg_colour = 0xffffffff;
			agl_use_program((AGlShader*)agl->shaders.texture);
		}else{
			agl_enable(AGL_ENABLE_TEXTURE_2D);
			glColor4f(1.0, 1.0, 1.0, 1.0); // seems to make a difference for alpha
		}
		agl_textured_rect(fbo->texture, r.x, r.y, r.w, r.h, &(AGlQuad){0.0, r.h / agl_power_of_two(r.h), r.w / agl_power_of_two(r.w), 0.0});
	}
#endif

	bool arr_gl_tracks_paint(AGlActor* actor)
	{
#ifdef USE_FBO
		AMTrack* tr = NULL;
		AMIter i;
		model_collection_iter_init (am_tracks, &i);
		while((tr = model_collection_iter_next(am_tracks, &i))){
			arr_gl_draw_parts_(actor, tr);
		}
#endif
		return true;
	}

#ifdef USE_FBO
	void arr_gl_tracks_invalidate(AGlActor* actor)
	{
		Arrange* arrange = actor->root->user_data;

		PF2;
#if 0
		void __track_invalidate(gpointer key, gpointer value, gpointer _arrange)
		{
			((GlTrack*)value)->cache.valid = false;
		}
		g_hash_table_foreach(arrange->canvas->gl->tracks, __track_invalidate, arrange);
#else
		GList* l = CGL->actors[ACTOR_TYPE_TC]->children;
		for(;l;l=l->next){
			TrackControlTrackActor* tcta = l->data;
			tcta->parts.cache.valid = false;
		}
#endif
	}
#endif

	void arr_gl_tracks_init(AGlActor* actor)
	{
		if(agl->use_shaders && !part_shader.shader.program){
			agl_create_program(&part_shader.shader);
			part_shader.uniform.border_colour = 0x000000aa;
		}
	}

	return AYYI_NEW(AGlActor,
		.name = "GlTracks",
		.init = arr_gl_tracks_init,
		.paint = arr_gl_tracks_paint,
#ifdef USE_FBO
		.invalidate = arr_gl_tracks_invalidate,
#endif
		.region = (AGliRegion){0, 0, 128000, 800}
	);
}


