/**
* +----------------------------------------------------------------------+
* | This file is part of the Ayyi project. http://www.ayyi.org           |
* | copyright (C) 2004-2017 Tim Orford <tim@orford.org>                  |
* +----------------------------------------------------------------------+
* | This program is free software; you can redistribute it and/or modify |
* | it under the terms of the GNU General Public License version 3       |
* | as published by the Free Software Foundation.                        |
* +----------------------------------------------------------------------+
*
*/
#include "global.h"
#include <gdk/gdkkeysyms.h>

#include "widgets/svgtogglebutton.h"
#include "windows.h"
#include "toolbox.h"
#include "toolbar.h"
#include "arrange.h"
#include "part_manager.h"
#include "arrange/part.h"
#include "support.h"
#include "song.h"
#include "icon.h"

static GtkWidget* toolbox_menu_init    (Arrange*);
static void       tool_on_click        (GtkWidget*, gpointer _tooltype);


static ToolbarButton toolboxi[TOOL_MAX] = {
	{"Normal",   "pointer",  0,                },
	{"Pencil",   "pencil",   CURSOR_PENCIL,    },
	{"Mute",     "cross",    CURSOR_CROSSHAIR, },
	{"Zoom",     "mag",      CURSOR_ZOOM_IN,   },
	{"Scissors", "scissors", CURSOR_SCISSORS,  },
	{"Scroll",   "hand",     CURSOR_HAND_OPEN, }
};


void
toolbox_free(Arrange* arrange)
{
	g_free(arrange->toolbox_buttons);
}


static GtkWidget*
toolbox_menu_init(Arrange* arrange)
{
	// Create the context menu for the arrange window root canvas.

	GtkWidget* menu = gtk_menu_new();

	int i;
	for(i=0;i<TOOL_MAX;i++){
		ToolbarButton* tool = &arrange->toolbox[i];
		GtkWidget* item = gtk_image_menu_item_new_with_label(tool->key);
		GtkWidget* image = svg_image_new(tool->icon);
		gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(item), image);
		gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
		g_object_set_data(G_OBJECT(item), "panel", arrange);
		g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(tool_on_click), GINT_TO_POINTER(i));
	}  	

	gtk_widget_show_all(menu);
	return menu;
}


void
toolbox_new(Arrange* arrange, GtkWidget* container)
{
	arrange->toolbox_buttons = g_malloc0(sizeof(GtkWidget*) * TOOL_MAX);

	arrange->toolbox = (ToolbarButton*)&toolboxi;
	arrange->rootmenu = toolbox_menu_init(arrange); // canvas context menu

	int i;
	for(i=0;i<TOOL_MAX;i++){
		ToolbarButton* item = &arrange->toolbox[i];
		GtkWidget* tb_toolbox = arrange->toolbox_buttons[i] = (GtkWidget*)svg_toggle_button_new(item->icon);
		g_object_set_data(G_OBJECT(tb_toolbox), "panel", arrange);
		gtk_box_pack_start(GTK_BOX(container), tb_toolbox, FALSE, FALSE, 0);

		g_signal_connect(G_OBJECT(tb_toolbox), "clicked", G_CALLBACK(tool_on_click), GINT_TO_POINTER(i));
		g_signal_connect(G_OBJECT(tb_toolbox), "enter-notify-event", G_CALLBACK(mouseover_enter), (gpointer)item->key);
		g_signal_connect(G_OBJECT(tb_toolbox), "leave-notify-event", G_CALLBACK(mouseover_leave), &arrange->panel);
	}
}


void
toolbox_set_canvas(Arrange* arrange, gpointer* callbacks)
{
	// Update the canvas-specific callbacks

	int i; for(i=1;i<TOOL_MAX;i++){ // <--- note 1 - should default tool be handled differently?
		arrange->toolbox[i].callback = callbacks[i];
	}
}


static void
tool_on_click(GtkWidget* widget, gpointer _tooltype)
{
	PF;
	static Arrange* a = NULL; //temp(!) workaround for double triggering. probably due to the gtk_toggle_button_set_active call?

	Arrange* arrange = g_object_get_data((GObject*)widget, "panel");
	if(a && a == arrange){ /*gwarn("aleady in handler! aborting...");*/ return; }
	a = arrange;
	g_return_if_fail(arrange_verify(arrange));

	toolbox_change_tool(arrange, GPOINTER_TO_INT(_tooltype));

	a = NULL;
}


void
toolbox_change_tool(Arrange* arrange, int tooltype)
{
	if(tooltype < TOOL_DEFAULT || tooltype >= TOOL_MAX){ gwarn("unknown tool type: %i", tooltype); return; }

	ToolbarButton* item     = &arrange->toolbox[tooltype];
	GtkSignalFunc  callback = (GtkSignalFunc)item->callback;
	arrange->tool = tooltype;
	dbg(2, "type=%i callback=%p", tooltype, item->callback);

	if(item && arrange->canvas->widget && GDK_IS_WINDOW(arrange->canvas->widget->window)) set_cursor(arrange->canvas->widget->window, item->cursor);

	// Show button state
	int i; for(i=0;i<TOOL_MAX;i++) svg_toggle_button_set_active ((SvgToggleButton*)arrange->toolbox_buttons[i], false);
	GtkWidget* active_button = arrange->toolbox_buttons[arrange->tool];
	svg_toggle_button_set_active ((SvgToggleButton*)active_button, true);

	call(arrange->canvas->on_tool_change, arrange, callback); 
}


void
toolbox_position(GtkMenu* menu, gint* x, gint* y, gboolean* push_in, gpointer user_data)
{
	// The menu is offset so that the mouse pointer is over the first entry.

	GdkEvent* event = user_data;

	gint ox, oy;
	gdk_window_get_origin(event->any.window, &ox, &oy);

	*x = ox + (gint)event->button.x -12;
	*y = oy + (gint)event->button.y -12;
}


