/**
* +----------------------------------------------------------------------+
* | This file is part of the Ayyi project. http://www.ayyi.org           |
* | copyright (C) 2004-2017 Tim Orford <tim@orford.org>                  |
* +----------------------------------------------------------------------+
* | This program is free software; you can redistribute it and/or modify |
* | it under the terms of the GNU General Public License version 3       |
* | as published by the Free Software Foundation.                        |
* +----------------------------------------------------------------------+
*
*/
#include "global.h"
#include <gdk/gdkkeysyms.h>
#ifdef HAVE_GTK_2_22
#include <gdk/gdkkeysyms-compat.h>
#endif
#include "agl/utils.h"
											#define USE_VIEW
											#ifdef USE_VIEW
#include "waveform/view.h"
											#else
											#include "waveform/view_plus.h"
											#endif
#include "shell.h"
#include "support.h"
#include "song.h"
#include "pool_model.h"
#include "../shortcuts.h"
#include "audio.h"

G_DEFINE_TYPE (AyyiAudioWin, ayyi_audio_win, AYYI_TYPE_PANEL)

static GObject* audio_win_construct  (GType, guint n_construct_properties, GObjectConstructParam*);
static void   audio_win_destroy      (GtkObject*);
static void   audio_win_on_ready     (AyyiPanel*);
static void   audio_win_on_allocate  (GtkWidget*, gpointer);
static void   audio_win_update       (AyyiAudioWin*);
static bool   audio_win_on_key_press (GtkWidget*, GdkEventKey*, gpointer);
static bool   audio_win_on_focus_in  (GtkWindow*, GdkEventFocus*, AyyiPanel*);
static gint   audio_drag_received    (GtkWidget*, GdkDragContext*, gint x, gint y, GtkSelectionData*, guint info, guint time, gpointer);


void
audio_win_init()
{
	g_type_class_unref (g_type_class_ref (AYYI_TYPE_AUDIO_WIN));
}


static void
ayyi_audio_win_class_init (AyyiAudioWinClass* klass)
{
	GObjectClass* g_object_class = G_OBJECT_CLASS (klass);
	GtkObjectClass* object_class = GTK_OBJECT_CLASS (klass);

    g_object_class->constructor = audio_win_construct;

	object_class->destroy  = audio_win_destroy;

	AyyiPanelClass* panel = AYYI_PANEL_CLASS(klass);
	panel->name           = "Audio";
	panel->has_link       = true;
	panel->has_follow     = true;
	panel->accel          = 0;
	panel->default_size.x = 480;
	panel->default_size.y = 200;
	panel->new            = audio_win_new;
	panel->on_ready       = audio_win_on_ready;

	void dummy_callback(AyyiPanel* panel){
		dbg(0, "...");
	}

	AMAccel list_key[] = {
    	{"Dummy",    {{(char)'j',    GDK_CONTROL_MASK}, {0, 0}}, dummy_callback, NULL, NULL},
		// needs to be trapped elsewhere
    	{"Zoom In",  {{GDK_KP_Equal, 0               }, {0, 0}}, dummy_callback, NULL, NULL},
	};
	panel->action_group = shortcuts_add_group(panel->name);
	panel->accel_group = gtk_accel_group_new();
	make_accels(panel->accel_group, panel->action_group, list_key, G_N_ELEMENTS(list_key), NULL);
}


static GObject*
audio_win_construct (GType type, guint n_construct_properties, GObjectConstructParam* construct_param)
{
	if(_debug_) printf("%s%s%s\n", constructor_colour, __func__, white);

	GObject* g_object = G_OBJECT_CLASS(ayyi_audio_win_parent_class)->constructor(type, n_construct_properties, construct_param);

	ayyi_panel_set_name((AyyiPanel*)g_object, AYYI_TYPE_AUDIO_WIN);

	return g_object;
}


static void
ayyi_audio_win_init(AyyiAudioWin* object)
{
}


GtkWidget*
audio_win_new(AyyiShell* shell, Size* config_size, GtkOrientation orientation)
{
	AyyiAudioWin* audio_win = AYYI_AUDIO_WIN (g_object_new (AYYI_TYPE_AUDIO_WIN, "behavior", GDL_DOCK_ITEM_BEH_NORMAL, "orientation", orientation, NULL));

	ayyi_panel_on_new((AyyiPanel*)audio_win, config_size, orientation);

	return (GtkWidget*)audio_win;
}


static void
audio_win_destroy(GtkObject* object)
{
	PF;
	AyyiAudioWin* audio = (AyyiAudioWin*)object;
	if(((AyyiPanel*)object)->destroyed) return;

	#define parts_handler_disconnect() if(g_signal_handlers_disconnect_matched (am_parts, G_SIGNAL_MATCH_DATA, 0, 0, NULL, NULL, audio) != 1) gwarn("handler disconnection");
	parts_handler_disconnect();
	#undef parts_handler_disconnect

	GTK_OBJECT_CLASS(ayyi_audio_win_parent_class)->destroy(object);

	((AyyiPanel*)object)->destroyed = true;
}


static void
audio_win_on_ready(AyyiPanel* panel)
{
	AyyiAudioWin* audio_win = (AyyiAudioWin*)panel;

	if(audio_win->waveform) return;

	GtkWidget* win = panel->shell->widget;
	g_signal_connect(G_OBJECT(win), "size-allocate", G_CALLBACK(audio_win_on_allocate), NULL);

#ifdef USE_VIEW
	//waveform_view_set_gl(agl_get_gl_context()); //TODO this only needs to be done once

	audio_win->waveform = waveform_view_new(NULL);
#else
	waveform_view_plus_set_gl(agl_get_gl_context()); // TODO this only needs to be done once
	audio_win->waveform = waveform_view_plus_new(NULL);
#endif
	gtk_widget_show((GtkWidget*)audio_win->waveform);
	panel_pack((GtkWidget*)audio_win->waveform, EXPAND_TRUE);

	void audio_on_part_selection_change(GObject* am_song, AyyiPanel* sender, gpointer audio){ audio_win_update((AyyiAudioWin*)audio); }
	g_signal_connect(am_parts, "selection-change", G_CALLBACK(audio_on_part_selection_change), audio_win);

	g_signal_connect((gpointer)audio_win->waveform, "key-press-event", G_CALLBACK(audio_win_on_key_press), panel);
	g_signal_connect(G_OBJECT(panel), "focus-in-event", G_CALLBACK(audio_win_on_focus_in), panel);

	gtk_drag_dest_set((GtkWidget*)audio_win->waveform, GTK_DEST_DEFAULT_ALL, app->dnd.file_drag_types, app->dnd.file_drag_types_count, (GdkDragAction)(GDK_ACTION_MOVE | GDK_ACTION_COPY));
	g_signal_connect(G_OBJECT(audio_win->waveform), "drag-data-received", G_CALLBACK(audio_drag_received), panel);

	audio_win_update(audio_win);
}


static void
audio_win_on_allocate(GtkWidget* win, gpointer user_data)
{
	// update object sizes following a change in window dimensions.
}


static void
audio_win_update(AyyiAudioWin* audio)
{
	PF;
	GList* p = am_parts_selection;
	AMPart* part = p
		? p->data
		: am_list_first(song->parts);
	if(part){
		PoolItem* pi = part->pool_item;
		if(pi){
			if(audio->waveform->waveform != (Waveform*)pi){
#ifdef USE_VIEW
				waveform_view_set_waveform(audio->waveform, (Waveform*)pi);
#else
				waveform_view_plus_set_waveform(audio->waveform, (Waveform*)pi);
				waveform_view_plus_set_colour(audio->waveform,
					part->fg_colour,
					palette_get_rgba(&song->palette, part->bg_colour, 0xff),
					0x777777ff,
					0xaaaaaaff
				);
#endif
			}
#ifdef USE_VIEW
			waveform_view_set_colour(audio->waveform, part->fg_colour, song->palette[part->bg_colour]);
#endif
		}
	}
}


static bool
audio_win_on_key_press(GtkWidget* widget, GdkEventKey* event, gpointer panel)
{
	AyyiAudioWin* audio = (AyyiAudioWin*)panel;
	WaveformView* waveform = audio->waveform;
	int handled = false;
	switch(event->keyval){
		case GDK_equal:
		case GDK_plus:
#ifdef USE_VIEW
			waveform_view_set_zoom(waveform, waveform->zoom * 1.5);
#else
			waveform_view_plus_set_zoom((WaveformViewPlus*)waveform, waveform->zoom * 1.5);
#endif
			break;
		case GDK_minus:
#ifdef USE_VIEW
			waveform_view_set_zoom(waveform, waveform->zoom / 1.5);
#else
			waveform_view_plus_set_zoom(waveform, waveform->zoom / 1.5);
#endif
			break;
		case GDK_Left:
		case GDK_KP_Left:
			waveform_view_set_start(waveform, waveform->start_frame - 8192 / waveform->zoom);
			handled = true;
			break;
		case GDK_Right:
		case GDK_KP_Right:
			waveform_view_set_start(waveform, waveform->start_frame + 8192 / waveform->zoom);
			handled = true;
			break;
	}
	return handled;
}


static gboolean
audio_win_on_focus_in(GtkWindow* window, GdkEventFocus* event, AyyiPanel* panel)
{
	AyyiAudioWin* audio_win = (AyyiAudioWin*)panel;
	gtk_widget_grab_focus((GtkWidget*)audio_win->waveform);
	return NOT_HANDLED;
}


static gint
audio_drag_received(GtkWidget* widget, GdkDragContext* drag_context, gint x, gint y, GtkSelectionData* data, guint info, guint time, gpointer user_data)
{
	// supported drag types:
	//  - colour     no
	//  - file       yes (from pool window)
	//  - part       no

	AyyiAudioWin* audio = user_data;

	if(data == NULL || data->length < 0){ gwarn ("no data!"); return -1; }

	dbg(1, "%s", data->data);

	switch(info){
		case AYYI_TARGET_URI_LIST:
			;GList* list = uri_list_to_glist((char*)data->data);
			GList* l = list;
			for(;l;l=l->next){
				char* u = l->data;
				gchar* method_string;
				vfs_get_method_string(u, &method_string);
				dbg (1, "method=%s", method_string);

				if(!strcmp(method_string, "pool")){
					PoolDrop data;
					pool_get_dropdata(u, &data);
					if(data.file){
						dbg(1, "item=%s", data.file->leafname);
#ifdef USE_VIEW
						waveform_view_set_waveform(audio->waveform, (Waveform*)data.file);
#else
						waveform_view_plus_set_waveform(audio->waveform, (Waveform*)data.file);
#endif
					}
				}
			}
			uri_list_free(list);
			break;
	}

	return false;
}
