/**
* +----------------------------------------------------------------------+
* | This file is part of the Ayyi project. http://www.ayyi.org           |
* | copyright (C) 2004-2017 Tim Orford <tim@orford.org>                  |
* +----------------------------------------------------------------------+
* | This program is free software; you can redistribute it and/or modify |
* | it under the terms of the GNU General Public License version 3       |
* | as published by the Free Software Foundation.                        |
* +----------------------------------------------------------------------+
*
*/

#include "global.h"
#include <sys/stat.h>
#include "gdl/gdl-dock-layout.h"
#include "model/am_utils.h"
#include "windows.h"
#include "shell.h"
#include "song.h"
#include "support.h"
#include "panels/log.h"

#include "settings.h"

/*

What do other applications do?
	- gimp: uses a gobject. All settings are properties of the gobject, and use GType etc.
	- gqview: simpler, uses global vars, and parser knows the type of each option.

*/
extern SMConfig* config;
extern GHashTable* canvas_types;


#ifdef HAVE_YAML_H
static ConfigWindow* config_get_window        (AyyiPanel*);
static void          config_load_windows_yaml (yaml_parser_t*, yaml_event_t*);
#else
static void          config_set_float         (char* group_name, const char* param, double value);
static void          config_set_int           (char* group_name, const char* param, int value);
static int           config_param_lookup      (WindowClass*, const char* param);
static void          config_load_windows      ();
#endif
static gboolean      create_config_dir        ();

#ifdef HAVE_YAML_H
unsigned char* str_tag = (unsigned char*)"tag:yaml.org,2002:str";
unsigned char* map_tag = (unsigned char*)"tag:yaml.org,2002:map";
yaml_emitter_t emitter;
#endif

#define plain_implicit_ TRUE

#ifdef XXX
#define end_map_(E) \
		if(!yaml_mapping_end_event_initialize(E)) goto error; \
		if(!yaml_emitter_emit(&emitter, E)) goto error;

#define map_open_(E, A) \
		if(!yaml_scalar_event_initialize(E, NULL, str_tag, (guchar*)A, -1, plain_implicit_, 0, YAML_PLAIN_SCALAR_STYLE)) goto error; \
		if(!yaml_emitter_emit(&emitter, E)) goto error; \
		if(!yaml_mapping_start_event_initialize(E, NULL, map_tag, 1, YAML_BLOCK_MAPPING_STYLE)) goto error; \
		if(!yaml_emitter_emit(&emitter, E)) goto error; \

		void gdl_dock_layout_foreach_object_save (GdlDockObject* object, gpointer user_data)
		{
			struct {
				yaml_event_t* event;
				GHashTable*   placeholders;
			} *info = user_data, info_child;

			guint n_props, i;
			GValue attr = {0,};

			g_return_if_fail (object != NULL && GDL_IS_DOCK_OBJECT (object));

			map_open_(info->event, gdl_dock_object_nick_from_type (G_TYPE_FROM_INSTANCE (object)));

			// get object exported attributes
			GParamSpec** props = g_object_class_list_properties (G_OBJECT_GET_CLASS (object), &n_props);
			g_value_init (&attr, GDL_TYPE_DOCK_PARAM);
			for (i = 0; i < n_props; i++) {
				GParamSpec* p = props [i];

				if (p->flags & GDL_DOCK_PARAM_EXPORT) {
					GValue v = {0,};

					// export this parameter
					// get the parameter value
					g_value_init (&v, p->value_type);
					g_object_get_property (G_OBJECT (object), p->name, &v);

					/*
					// only save the object "name" if it is set
					//   (i.e. don't save the empty string)
					if (strcmp (p->name, GDL_DOCK_NAME_PROPERTY) || g_value_get_string (&v)) {
						if (g_value_transform (&v, &attr))
							xmlSetProp (node, BAD_CAST p->name, BAD_CAST g_value_get_string (&attr));
					}
					*/

					// free the parameter value
					g_value_unset (&v);
				}
			}
			g_value_unset (&attr);
			g_free (props);

			end_map_(info->event);

			/*
			info_child = *info;
			info_child.where = node;

			// save placeholders for the object
			if (info->placeholders && !GDL_IS_DOCK_PLACEHOLDER (object)) {
				GList *lph = g_hash_table_lookup (info->placeholders, object);
				for (; lph; lph = lph->next)
					gdl_dock_layout_foreach_object_save (GDL_DOCK_OBJECT (lph->data), (gpointer) &info_child);
			}

			// recurse the object if appropiate
			if (gdl_dock_object_is_compound (object)) {
				gtk_container_foreach (GTK_CONTAINER (object),
                               (GtkCallback) gdl_dock_layout_foreach_object_save,
                               (gpointer) &info_child);
			}
			*/
			return;

		  error:
			switch (emitter.error){
				case YAML_MEMORY_ERROR:
					fprintf(stderr, "Memory error: Not enough memory for emitting\n");
					break;
				case YAML_WRITER_ERROR:
					fprintf(stderr, "Writer error: %s\n", emitter.problem);
					break;
				case YAML_EMITTER_ERROR:
					log_print(LOG_FAIL, "yaml emitter error: %s", emitter.problem);
					break;
				default:
					fprintf(stderr, "save: yaml internal error: %s\n", emitter.problem);
					break;
			}
			/*
			yaml_event_delete(&event);
			yaml_emitter_delete(&emitter);
			fclose(fp);
			*/
		}
#endif


#ifdef GDLDOCK_XML
static void
save_layout()
{
	GdlDockLayout* layout = gdl_dock_layout_new(app->master_dock);

	/*       -- it seems we cannot use gdl_dock_master_foreach to iterate over the dock
	GList* names = NULL;

	{
		// dock item names mess up re-loading of the layout so remove them temporarily
		void clear_name (GdlDockObject *object, gpointer user_data)
		{
			GList **l = (GList **) user_data;

			if(GDL_IS_DOCK_ITEM (object)){
				*l = g_list_append (*l, object->name);
				object->name = NULL;
			}
		}
		gdl_dock_master_foreach (GDL_DOCK_OBJECT_GET_MASTER(app.master_dock), (GFunc)clear_name, &names);
	}
	*/

	gdl_dock_layout_attach(layout, (GdlDockMaster*)GDL_DOCK_OBJECT(app->master_dock)->master);
	gdl_dock_layout_save_layout(layout, 0);
	char* fname = g_strdup_printf("%s/"USER_CONFIG_DIR"/layout.xml", g_get_home_dir());
	if(gdl_dock_layout_save_to_file(layout, fname)){
		dbg(1, "save done");
	}
	else gwarn("save error");
	g_free(fname);

	/*
	{
		void restore_name (GdlDockObject *object, gpointer user_data)
		{
			GList **l = (GList **) user_data;

			if(GDL_IS_DOCK_ITEM (object)){
				object->name = (*l)->data;
			}
		}
		gdl_dock_master_foreach (GDL_DOCK_OBJECT_GET_MASTER(app.master_dock), (GFunc)restore_name, &names);
		g_list_free(names);
	}
	*/
}
#endif


bool
config_save()
{
	if(app->state == APP_STATE_ABORT){ dbg(0, "not saving."); return FALSE; }
	PF;
	#define MAX_WIN_COUNT 9 //cannot save more than this number of windows per type.

	if(!create_config_dir()) return false;

	char value[256];
#ifdef HAVE_YAML_H
	if(!yaml_emitter_initialize(&emitter)){ gerr("failed to initialise yaml writer."); return FALSE; }

	FILE* fp = fopen(app->config_filename, "wb");
	if(!fp){
		log_print(LOG_FAIL, "cannot open config file for writing (%s).", app->config_filename);
		return FALSE;
	}

	yaml_emitter_set_output_file(&emitter, fp);
	yaml_emitter_set_canonical(&emitter, false); //needed?

	//stream start
	yaml_event_t event;
	yaml_stream_start_event_initialize(&event, YAML_UTF8_ENCODING);
	if(!yaml_emitter_emit(&emitter, &event)) goto error;

	//document start
	if (!yaml_document_start_event_initialize(&event, NULL, NULL, NULL, 0)) goto error;
	yaml_emitter_emit(&emitter, &event);

	int plain_implicit = 1;
	//int quoted_implicit = 0;

	if(!yaml_mapping_start_event_initialize(&event, NULL, (guchar*)"tag:yaml.org,2002:map", 1, YAML_BLOCK_MAPPING_STYLE)) goto error;
	if(!yaml_emitter_emit(&emitter, &event)) goto error;

	snprintf(value, 255, "0x%08x", config->part_outline_colour);
	if(!yaml_add_key_value_pair("part_outline_colour", value)) goto error;
	snprintf(value, 255, "0x%08x", config->part_outline_colour_selected);
	if(!yaml_add_key_value_pair("part_outline_colour_selected", value)) goto error;
	if(config->arr_background_image && !yaml_add_key_value_pair("background_image", config->arr_background_image)) goto error;
	if(!yaml_add_key_value_pair_int("snap_mode", song->snap_mode)) goto error;
	if(!yaml_add_key_value_pair("icon_theme", config->icon_theme)) goto error;

	//---------------------------------------------------------

	map_open("windows");

	if(_debug_) gdl_dock_print_recursive(GDL_DOCK_MASTER(GDL_DOCK_OBJECT(((AyyiShell*)shell_list->data)->dock)->master));
	struct _AyyiDockInfo* info[32];
	memset(info, 0, 32 * sizeof(struct _AyyiDockInfo*));
	gdl_dock_get_layout_info(GDL_DOCK_MASTER(GDL_DOCK_OBJECT(((AyyiShell*)shell_list->data)->dock)->master), (void*)info);

	AyyiShell*
	get_shell_from_dockobject(GdlDockObject* object)
	{
		GList* s = shell_list;
		for(;s;s=s->next){
			AyyiShell* shell = s->data;
			if(gtk_widget_get_toplevel(GTK_WIDGET(shell->widget)) == gtk_widget_get_toplevel(GTK_WIDGET(object))) return shell;
		}
		return NULL;
	}

	void show_k(struct _AyyiDockInfo* info)
	{
		int i; for(i=0;i<G_N_ELEMENTS(info->k);i++){
			printf("%i ", info->k[i]);
		}
		printf("\n");
	}

	char* dock_info_to_string(struct _AyyiDockInfo* nfo)
	{
		static char str[32] = "";
		int i; for(i=0;i<G_N_ELEMENTS(nfo->k);i++){
			sprintf(str+i, "%i", nfo->k[i]);
		}
		return str;
	}

	void dock_info_free()
	{
		//TODO test.
		PF;
		int idx = 0;
		while(info[idx]){
			//dbg(0, "  %i %s", info[idx]->depth, info[idx]->object->name);
			g_free(info[idx]);
			idx++;
		}
	}

	char group_name [64];
	gint width, height, x, y;
	GHashTable* window_count = windows__new_count_table();
	int i = 0;
	struct _AyyiDockInfo* nfo;
	for(;(nfo=info[i]);i++){
		dbg(1, "i=%i", i);
		AyyiShell* shell = get_shell_from_dockobject(nfo->object);
		if(!shell){ gerr("dockobject contains no shell"); continue; }
		GList* panels = shell__get_panels(shell);
		if(!panels){ gerr ("empty shell."); continue; }

		//currently the shell is named after its first child panel.
		AyyiPanel* first = (AyyiPanel*)panels->data;
		PanelType dtype = G_OBJECT_TYPE(first);
		int count = GPOINTER_TO_INT(g_hash_table_lookup(window_count, &dtype));
		dbg(1, "first=%p type=%i count=%i", first, dtype, count);
		config_make_group_name(AYYI_PANEL_GET_CLASS(first), count, group_name);
		dbg(1, "saving group: %s", group_name);

		map_open(group_name);

		gtk_window_get_size(GTK_WINDOW(shell->widget), &width, &height);
		if(!yaml_add_key_value_pair_int("width", width)) goto error;
		if(!yaml_add_key_value_pair_int("height", height)) goto error;

		gtk_window_get_position(GTK_WINDOW(shell->widget), &x, &y);
		if(!yaml_add_key_value_pair_int("x", x)) goto error;
		if(!yaml_add_key_value_pair_int("y", y)) goto error;

		GdkWindowState state = gdk_window_get_state(GTK_WIDGET(shell->widget)->window);
		gboolean maximised = state & GDK_WINDOW_STATE_MAXIMIZED;
		dbg(1, "maximised=%i", maximised);
		if(!yaml_add_key_value_pair_int("window_state", maximised)) goto error;
		if(!yaml_add_key_value_pair_bool("maximised", maximised)) goto error;

		//child panels.
		//-currently we use the same naming convention as for parent windows.
		if(panels){
			map_open("panels");
			char panel_name [64];

			for(i++;(nfo=info[i]);i++){
				GdlDockObject* object = nfo->object;
				if(G_OBJECT_TYPE(object) == GDL_TYPE_DOCK_PANED) continue;
				if(G_OBJECT_TYPE(object) == GDL_TYPE_DOCK) break; //top level! start a new window section
				dbg(1, "  %i: depth=%i %s", i, nfo->depth, dock_info_to_string(nfo));
				AyyiPanel* panel = AYYI_PANEL(object);
				g_return_val_if_fail(panel, false);
				snprintf(panel_name, 63, "%s", object->long_name ? object->long_name : "<Unnamed>");

				//window_count [panel->type] ++;
				PanelType dtype = G_OBJECT_TYPE(panel);
				int n = windows__count_table_add(window_count, dtype);
				if(n > MAX_WIN_COUNT) continue;
				//if(panel->type >= WINDOW_TYPE_MAX) continue;

				map_open(panel_name);

#ifndef GDLDOCK_XML
				if(!yaml_add_key_value_pair_int("width", GTK_WIDGET(object)->allocation.width)) goto error;
				if(!yaml_add_key_value_pair_int("height", GTK_WIDGET(object)->allocation.height)) goto error;
				if(!yaml_add_key_value_pair_int("x", GTK_WIDGET(object)->allocation.x)) goto error;
				if(!yaml_add_key_value_pair_int("y", GTK_WIDGET(object)->allocation.y)) goto error;
				int depth = ayyi_panel_get_depth(panel);
				if(!yaml_add_key_value_pair_int("depth", depth)) goto error;
				if(!yaml_add_key_value_pair("dock", dock_info_to_string(nfo))) goto error;

				GdlDockObject* parent = gdl_dock_object_get_parent_object(object);
				if(parent){
					if(!GDL_IS_DOCK_CLASS(G_OBJECT_GET_CLASS(parent))){
						//possibly a shell with only one panel.
						GtkOrientation orientation;
						g_object_get((GdlDockItem*)parent, "orientation", &orientation, NULL);
						if(!yaml_add_key_value_pair("orientation", get_orientation_string(orientation))) goto error;
					}
				} else gwarn("no parent");
#endif

				AyyiPanelClass* k = (AyyiPanelClass*)G_OBJECT_GET_CLASS(panel);
				if(k->get_config){
					void** vals = k->get_config(panel);
					ConfigParam** panel_config = k->config;
					int p; ConfigParam* param; for(p=0;(param=panel_config[p]);p++){
						if(vals[p]){
							switch(param->utype){
								case G_TYPE_FLOAT:
									dbg(1, "    %s=%.2f", param->name, *(float*)vals[p]);
									if(!yaml_add_key_value_pair_float(param->name, *(float*)vals[p])) goto error;
									break;
								case G_TYPE_INT:
									dbg(1, "    %s=%i", param->name, *(int*)vals[p]);
									if(!yaml_add_key_value_pair_int(param->name, *(int*)vals[p])) goto error;
									break;
								default:
									dbg(1, "    %s=%s", param->name, *(char**)vals[p]);
									if(!yaml_add_key_value_pair(param->name, *(char**)vals[p])) goto error;
									break;
							}
						}
					}
					g_free(vals);
				}

				end_map;
			}
			g_list_free(panels);

			end_map; //end panels
        }

		end_map; //end window

		dbg(2, "map ended");
	}
	end_map; //end windows
	dock_info_free();

	if(!yaml_mapping_end_event_initialize(&event)) goto error;
	if(!yaml_emitter_emit(&emitter, &event)) goto error;

	//---------------------------------------------------------

/*	sequence example:

	if(!yaml_sequence_start_event_initialize(&event, NULL, (unsigned char*)"tag:yaml.org,2002:seq", 1, YAML_BLOCK_SEQUENCE_STYLE)) goto error;
	if(!yaml_emitter_emit(&emitter, &event)) goto error;

	if(!yaml_scalar_event_initialize(&event, "hello", (unsigned char*)"tag:yaml.org,2002:str", (unsigned char*)"value", -1, plain_implicit, quoted_implicit, YAML_DOUBLE_QUOTED_SCALAR_STYLE)) goto error;
	if(!yaml_emitter_emit(&emitter, &event)) goto error;

	//yaml_scalar_event_initialize(&event, (unsigned char*)"anchor", (unsigned char*)"tag", (unsigned char*)"value", 5, plain_implicit, quoted_implicit, YAML_ANY_SCALAR_STYLE);
	if(!yaml_scalar_event_initialize(&event, NULL, (unsigned char*)"tag:yaml.org,2002:str", (unsigned char*)"value", -1, plain_implicit, quoted_implicit, YAML_PLAIN_SCALAR_STYLE)){ gwarn("scaler event int error."); goto error; }
	if(!yaml_emitter_emit(&emitter, &event)){ gwarn("scaler emit error."); goto error; }
	  ...
	  //Emit more events.
	  ...

	if(!yaml_sequence_end_event_initialize(&event)) goto error;
	if(!yaml_emitter_emit(&emitter, &event)) goto error;
*/

	end_document;
	yaml_emitter_delete(&emitter);
	fclose(fp);
	dbg(1, "yaml write finished ok.");

#ifdef GDLDOCK_XML
	save_layout();
#endif

	GString* sb = g_string_new("");
#ifdef HAVE_YAMLGLIB
	GError* error = NULL;
	#ifdef locallibyamlglib
	#include "libyaml-glib/src/libyaml-glib-1.0.h"
	GYAMLWriter* writer = g_yaml_writer_new(NULL);
	GError* error = NULL;
	//g_yaml_writer_stream_object (writer, G_OBJECT(song->controller), sb, &error);
	//g_yaml_writer_stream_object (writer, G_OBJECT(app.master_dock), sb, &error);
	//g_yaml_writer_stream_object (writer, G_OBJECT(song->tracks), sb, &error);

	#else
	#if 0
	#include "yaml_writer.h"

	printf("\n");
	dbg(0, "eventsize=%i", sizeof(yaml_event_t));
	printf("\n");

	YAMLWriter* writer = yaml_writer_new(NULL);
	yaml_writer_stream_object (writer, G_OBJECT(GDL_DOCK_OBJECT(app.master_dock)->master), sb, &error);
	//yaml_writer_stream_object (writer, G_OBJECT(song->controller), sb, &error);
	#endif
	#endif

	if(error){
		gwarn("%s", error->message);
	} else dbg(1, "no errors");
	if(_debug_){
		printf("\n");
		dbg(1, "yaml=%s", sb->str);
		printf("\n");
	}
#endif

	return TRUE;

  error:
	switch (emitter.error){
		case YAML_MEMORY_ERROR:
			fprintf(stderr, "Memory error: Not enough memory for emitting\n");
			break;
		case YAML_WRITER_ERROR:
			fprintf(stderr, "Writer error: %s\n", emitter.problem);
			break;
		case YAML_EMITTER_ERROR:
			log_print(LOG_FAIL, "yaml emitter error: %s", emitter.problem);
			break;
		default:
			fprintf(stderr, "Internal error\n");
			break;
	}
	yaml_event_delete(&event);
	yaml_emitter_delete(&emitter);
	fclose(fp);
	return FALSE;

#else
#if 0
	//global settings:
	gint x, y;
	snprintf(value, 256, "0x%08x", config->part_outline_colour);
	g_key_file_set_value(app.key_file, "Seismix", "part_outline_colour", value);
	snprintf(value, 256, "0x%08x", config->part_outline_colour_selected);
	g_key_file_set_value(app.key_file, "Seismix", "part_outline_colour_selected", value);
	snprintf(value, 256, "0x%08x", arrange->bg_colour);
	g_key_file_set_value(app.key_file, "Seismix", "arr_background_colour", value);
	snprintf(value, 256, "%i", song->snap_mode);
	g_key_file_set_value(app.key_file, "Seismix", "snap_mode", value);
	snprintf(value, 256, "%s", config->icon_theme);
	g_key_file_set_value(app.key_file, "Seismix", "icon_theme", value);

	//recent songs:
	int i;
	for(i=0;i<MAX_RECENT_SONGS;i++){
		char keyname[256];
		snprintf(keyname, 255, "recent_songs_%i", i+1);
		snprintf(value, 255, "%s", app.recent_songs[i]);
		g_key_file_set_value(app.key_file, "Seismix", keyname, value);
	}

	char group_name [64];
	//clear old window entries:
	for(i=1;i<WINDOW_TYPE_MAX;i++){
		int k;
		for(k=0;k<=MAX_WIN_COUNT;k++){
			config_make_group_name(i, k, group_name);
			g_key_file_remove_group(app.key_file, group_name, NULL);
			dbg (3, "deleting group: %s", group_name);
		}
	}

	dbg (2, "saving other windows...");
	//save other windows:
	int window_count[WINDOW_TYPE_MAX] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
	panel_foreach {
		if(window_count [panel->type] > MAX_WIN_COUNT) continue;
		if(panel->type >= WINDOW_TYPE_MAX) continue;

		config_make_group_name(panel->type, window_count[panel->type], group_name);
		dbg (2, "saving group: %s", group_name);
		window_count [panel->type] ++;

		if(!GDL_IS_DOCK_OBJECT(panel)){ gerr("bad dock_item. panel=%s", group_name); continue; }
		GdlDockObject* object = GDL_DOCK_OBJECT(panel);

		//gtk_window_get_size(GTK_WINDOW(sm_window->shell->widget), &width, &height);
		snprintf(value, 256, "%i", GTK_WIDGET(object)->allocation.width);
		g_key_file_set_value(app.key_file, group_name, "width", value);
		snprintf(value, 256, "%i", GTK_WIDGET(object)->allocation.height);
		g_key_file_set_value(app.key_file, group_name, "height", value);

		gtk_window_get_position(GTK_WINDOW(panel->shell->widget), &x, &y);
		snprintf(value, 256, "%i", x + GTK_WIDGET(panel)->allocation.x);
		g_key_file_set_value(app.key_file, group_name, "x", value);
		snprintf(value, 256, "%i", y + GTK_WIDGET(panel)->allocation.y);
		g_key_file_set_value(app.key_file, group_name, "y", value);

		switch(panel->type){
			case TYPE_ARRANGE:
				{
					Arrange* arr = (Arrange*)panel;
					snprintf(value, 255, "%i", arr->canvas->type);
					g_key_file_set_value(app.key_file, group_name, "canvas_type", value);
					config_set_float(group_name, "horizontal_zoom",   arr->hzoom);
					config_set_float(group_name, "vertical_zoom",     arr->vzoom);
					config_set_int  (group_name, "trackctl_width",    arr->ctl_viewport->allocation.width);
					config_set_int  (group_name, "show_overview",     arr->view_options[SHOW_SONG_OVERVIEW].value);
					config_set_int  (group_name, "show_part_contents",arr->view_options[SHOW_PART_CONTENTS].value);
					config_set_int  (group_name, "show_songmap",      arr->view_options[SHOW_SONG_MAP].value);
				}
				break;
			case TYPE_MIXER:
				snprintf(value, 255, "%.2f", mixer_get_zoom((MixerWin*)panel));
				g_key_file_set_value(app.key_file, group_name, "strip_width", value);
				break;
		}
	} end_panel_foreach

	//serialise the data:
	GError* error = NULL;
	gsize length;
	gchar* string = g_key_file_to_data(app.key_file, &length, &error);
	if(error){
		gerr ("error saving config file: %s", error->message);
		g_error_free(error);
	}

	FILE* fp;
	if(!(fp = fopen(app.config_filename, "w"))){
		errprintf("cannot open config file for writing (%s).\n", app.config_filename);
		return FALSE;
	}else{
		if(fprintf(fp, "%s", string) < 0){
			errprintf("error writing data to config file (%s).\n", app.config_filename);
		}
		fclose(fp);
		return FALSE;
	}
	log_print(0, "config saved");
	g_free(string);
#endif //0
#endif //HAVE_YAML

	return true;
}


void
config_load0()
{
	add_pixmap_directory(PACKAGE_DATA_DIR "/" PACKAGE "/pixmaps");

	static const char* dirs[] = {IMAGES_DIR, "../pixmaps"};
	const char* image_dir = find_path(dirs);
	dbg (2, "image_dir=%s", image_dir);

	gchar* cwd = g_get_current_dir();

	config->arr_background_image = NULL;
	config->svgdir = g_strdup_printf("%s/svg", image_dir);
	config->accels_file = g_strdup_printf("%s/"USER_CONFIG_DIR"/seismix.keys", g_get_home_dir());
	config->default_send_count = 2;
#ifdef USE_LV2
	config->plugindir = g_strdup_printf(PACKAGE_LIB_DIR);
#endif

	g_free(cwd);
}


gboolean
config_load()
{
	/* Difficult to know when to run this. On the one hand, its nice to be able to apply
	   settings while we parse the file, but on the other, we cant create create objects
	   without config information.
	*/

	//set defaults:
	config->part_outline_colour          = 0x00000000; //default is to have no border.
	config->part_outline_colour_selected = 0xffffffff; //default is white when selected.

#ifdef HAVE_YAML_H
	//see tmp/yaml-0.0.1/tests/example-deconstructor.c
	//
	snprintf(app->config_filename, 255, "%s/"USER_CONFIG_DIR"/seismix.yml", g_get_home_dir());

	enum {
		CONFIG_SECTION_NONE,
		CONFIG_SECTION_WINDOWS,
	};
	int section = 0;

	yaml_parser_t parser; yaml_parser_initialize(&parser);

	FILE* file_input = fopen(app->config_filename, "rb");
	if(!file_input){
		log_print(LOG_FAIL, "unable to load config file: %s", app->config_filename);
		return FALSE;
	}
	yaml_parser_set_input_file(&parser, file_input);

	//read the event sequence.
	int safety = 0;
	int depth = 0;
	char key[64] = "";
	gboolean end = FALSE;
	yaml_event_t event;
	do {

		if (!yaml_parser_parse(&parser, &event)) goto error; // Get the next event.

		switch (event.type) {
			case YAML_STREAM_START_EVENT:
				dbg(2, "YAML_STREAM_START_EVENT");
				break;
			case YAML_STREAM_END_EVENT:
				end = TRUE;
				dbg(2, "YAML_STREAM_END_EVENT");
				break;
			case YAML_DOCUMENT_START_EVENT:
				dbg(2, "YAML_DOCUMENT_START_EVENT");
				break;
			case YAML_DOCUMENT_END_EVENT:
				end = TRUE;
				dbg(2, "YAML_DOCUMENT_END_EVENT");
				break;
			case YAML_ALIAS_EVENT:
				dbg(0, "YAML_ALIAS_EVENT");
				break;
			case YAML_SCALAR_EVENT:
				if(event.data.scalar.anchor) dbg(0, "YAML_SCALAR_EVENT: %s => %s", event.data.scalar.anchor, event.data.scalar.tag);
				//dbg(0, "YAML_SCALAR_EVENT: value=%s %i plain=%i style=%i", event.data.scalar.value, event.data.scalar.length, event.data.scalar.plain_implicit, event.data.scalar.style);

				if(!key[0]){ //1st half of a pair. We dont yet know what to do with it
					strncpy(key, (char*)event.data.scalar.value, 63);
				}else{
					//2nd half of a pair
					dbg(2, "      %s=%s", key, event.data.scalar.value);

					if(!strcmp(key, "part_outline_colour")){
						long long colour = strtoll((char*)event.data.scalar.value, NULL, 16);
						config->part_outline_colour = colour;
					}
					else if(!strcmp(key, "part_outline_colour_selected")){
						long long colour = strtoll((char*)event.data.scalar.value, NULL, 16);
						config->part_outline_colour_selected = colour;
					}
					else if(!strcmp(key, "snap_mode")){
						guint snap_mode = atoi((char*)event.data.scalar.value);
						if(snap_mode < MAX_SNAP) config->snap_mode = snap_mode;
					}
					else if(!strcmp(key, "background_image")){
						config->arr_background_image = g_strdup((char*)event.data.scalar.value);
					}
					else if(!strcmp(key, "icon_theme")){
						config->icon_theme = g_strdup((char*)event.data.scalar.value);
					}

					key[0] = '\0';
				}
				break;
			case YAML_SEQUENCE_START_EVENT:
				dbg(2, "YAML_SEQUENCE_START_EVENT");
				break;
			case YAML_SEQUENCE_END_EVENT:
				dbg(2, "YAML_SEQUENCE_END_EVENT");
				break;
			case YAML_MAPPING_START_EVENT:
				depth++;
				if(key[0]){
					if(!section){
						if(!strcmp(key, "windows")){
							//if(section) gwarn("parse error: 'windows' expected to be top level section.");
							dbg(2, "found Windows section!");
							section = CONFIG_SECTION_WINDOWS;
							config_load_windows_yaml(&parser, &event);

							//check the loaded windows:
							/*
							int w;for(w=0;w<config->windows->len;w++){
								ConfigWindow* window = &g_array_index(config->windows, ConfigWindow, w);
								dbg(0, "%i: type=%s", w, g_type_name(window->type));
								GList* l = window->params; for(;l;l=l->next){
									ConfigParamValue* v = l->data;
									dbg(0, "  param: %s", v->param->name);
								}
							}
							*/
						}
						else gwarn("unexpected section: %s", key);
					}
					else if(section == CONFIG_SECTION_WINDOWS){
						dbg(0, "new window!: %s", key);
						//window = &g_array_index(config->windows, ConfigWindow, window_num);
					}
					else dbg(0, "new section: %s", key);
					key[0] = '\0';
				}
				else dbg(2, "YAML_MAPPING_START_EVENT");
				break;
			case YAML_MAPPING_END_EVENT:
				dbg(2, "YAML_MAPPING_END_EVENT");
				if(--depth < 0) gwarn("too many YAML_MAPPING_END_EVENT's.");
				break;
			case YAML_NO_EVENT:
				dbg(0, "YAML_NO_EVENT");
				break;
		}

		//the application is responsible for destroying the event object.
		yaml_event_delete(&event);

	} while(!end /*event.type != YAML_STREAM_END_EVENT*/ && safety++ < 50);

	yaml_parser_delete(&parser);

	config->mixer_use_rotaries_for_auxs = TRUE;
	return TRUE;

  error:
	yaml_parser_delete(&parser);

	return FALSE;
#else
	//use glib
	snprintf(app->config_filename, 255, "%s/"USER_CONFIG_DIR"/seismix", g_get_home_dir());

	GError* error = NULL;
	app.key_file = g_key_file_new();
	if(g_key_file_load_from_file(app.key_file, app->config_filename, G_KEY_FILE_KEEP_COMMENTS | G_KEY_FILE_KEEP_TRANSLATIONS, &error)){
		gchar* groupname = g_key_file_get_start_group(app.key_file);
		if(!strcmp(groupname, "Seismix")){

			//FIXME 64 is not long enough for directories
			#define STR____LEN 64
			struct {char* name; char* loc; } str_keys[] = {
				{"recent_songs_1", app.recent_songs[0]},
				{"recent_songs_2", app.recent_songs[1]},
				{"recent_songs_3", app.recent_songs[2]},
				{"recent_songs_4", app.recent_songs[3]},
				{"icon_theme",     config->icon_theme},
			};

			//no float vals currently
			struct {char* name; float* loc;} fkeys[] = {
			};

			struct {char* name; int* loc;} ikeys[] = {
				{"snap_mode",          (int*)&config->snap_mode},
			};

			int k;
			gchar* keyname;
			for(k=0;k<G_N_ELEMENTS(str_keys);k++){
				if((keyname = g_key_file_get_string(app.key_file, groupname, str_keys[k].name, &error))){
					snprintf(str_keys[k].loc, STR____LEN-1, "%s", keyname);
					g_free(keyname);
					print_gerror(&error);
				}else str_keys[k].loc = '\0';
			}

			//get float vals:
			gchar* keyval;
			for(k=0;k<G_N_ELEMENTS(fkeys);k++){
				if((keyval = g_key_file_get_string(app.key_file, groupname, fkeys[k].name, &error))){
					*fkeys[k].loc = atof(keyval);
					g_free(keyval);
					print_gerror(&error);
				}else *fkeys[k].loc = 1.0;
			}
			//config->horizontal_zoom = CLAMP(config->horizontal_zoom, ARR_MIN_HZOOM, ARR_MAX_HZOOM);
			//config->vertical_zoom   = CLAMP(config->vertical_zoom,   ARR_MIN_VZOOM, ARR_MAX_VZOOM);

			keyval = g_key_file_get_string(app.key_file, groupname, "part_outline_colour", NULL);
			long long colour;
			if(keyval) colour = strtoll(keyval, NULL, 16); else colour = 0x00000000; //default is to have no border.
			config->part_outline_colour = colour;

			keyval = g_key_file_get_string(app.key_file, groupname, "part_outline_colour_selected", NULL);
			if(keyval) colour = strtoll(keyval, NULL, 16); else colour = 0xffffffff; //default is white when selected.
			config->part_outline_colour_selected = colour;

			keyval = g_key_file_get_string(app.key_file, groupname, "arr_background_colour", NULL);
			if(keyval) colour = strtoll(keyval, NULL, 16); else colour = 0x000000ff; //would be better to use style colour. -- try set alpha to 00 !!
			config->arr_background_colour = colour;

			//get int vals:
			for(k=0;k<G_N_ELEMENTS(ikeys);k++){
				if((keyval = g_key_file_get_string(app.key_file, groupname, ikeys[k].name, &error))){
					*ikeys[k].loc = atoi(keyval);
					g_free(keyval);
					P_GERR;
				}else *ikeys[k].loc = 0;
			}

		}
		else { gwarn ("cannot find Seismix key group."); return FALSE; }
		g_free(groupname);

		config->mixer_use_rotaries_for_auxs = TRUE;

		config_load_windows();

	}else{
		if (error->code == 1){
			log_print(LOG_FAIL, "config file is empty");
		} else if (error->code == 4){
			log_print(0, "no config file found.");
		} else {
			log_print(LOG_FAIL, "unable to load config file: (%i) %s", error->code, error->message);
		}
		g_error_free(error);
		error = NULL;
		config_new();
		return FALSE;
	}
#endif

	log_print(LOG_OK, "config loaded");
	return TRUE;
}


static ConfigParam*
window_class__config_lookup(AyyiPanelClass* class, const char* param)
{
	dbg(2, "%s", param);
	int i = 0;
	ConfigParam* p = NULL;
	while((p = class->config[i++])){
		if(!strcmp(param, p->name)) return p;
	}
	gwarn("not found: %s", param);
	return NULL;
}


int
config_get_window_setting_int(AyyiPanel* panel, const char* param)
{
	//return the value of a single config setting.
	//-used for settings that cannot be stored at window creation time.  XX no longer needed - width & height?
	// note: it is marginally more efficient to apply all settings at once in window initialiser.

	AyyiPanelClass* klass = AYYI_PANEL_GET_CLASS(panel);
#if HAVE_YAML_H
	if(config->windows){
		ConfigWindow* window_settings = config_get_window(panel);
		g_return_val_if_fail(window_settings, false);

		GList* l = window_settings->params;
		for(;l;l=l->next){
			ConfigParamValue* pval = l->data;
			ConfigParam* i = pval->param;
			int v = pval->val.i;
			if(!strcmp(i->name, param)){
				dbg(2, "%s=%i", i->name, v);
				return CLAMP(v, i->min.i, i->max.i);
			}
		}
	}

	//not found, so return the default value.

	//AyyiPanelClass* panel_class = AYYI_PANEL_GET_CLASS(panel);
	//if(!strcmp(param, "width")){
	//	return panel_class->default_size.x;
	//}
	ConfigParam* setting = window_class__config_lookup(klass, param);
	if(setting) dbg(2, "%s: using default: %i", param, setting->dfault);
	int val = setting ? setting->dfault.i : 0;
#else
	GError* error = NULL;
	static char group_name[64];
	config_make_group_name(panel->type, class->instances-1, group_name);
	int val = g_key_file_get_integer(app.key_file, group_name, param, &error);
	if(error){
		g_error_free(error); error = NULL;
		//use default:
		int i = config_param_lookup(class, param);
		val = class->config[i]->i.dfault;
	}
	dbg (2, "'%s' %s=%i", group_name, param, val);
#endif
	return val;
}


double
config_get_window_setting_float(AyyiPanel* panel, const char* param)
{
	AyyiPanelClass* klass = AYYI_PANEL_GET_CLASS(panel);

#if HAVE_YAML_H
	if(config->windows){
		ConfigWindow* window_settings = config_get_window(panel);
		if(window_settings){
			GList* l = window_settings->params;
			for(;l;l=l->next){
				ConfigParamValue* pval = l->data;
				ConfigParam* f = pval->param;
				float v = pval->val.f;
				if(!strcmp(f->name, param)){
					dbg(2, "%s=%.2f", f->name, v);
					return v;
				}
			}
		}
	}

	// not found, so return the default value.
	ConfigParam* setting = window_class__config_lookup(klass, param);
	if(!setting) gwarn("%s: cannot get default. setting not found", param);
	double val = setting ? setting->dfault.f : 0;

#else
	GError* error = NULL;
	static char group_name[64];
	config_make_group_name(window->type, class->instances-1, group_name);
	double val = g_key_file_get_double(app.key_file, group_name, param, &error);
	if(error){
		g_error_free(error); error = NULL;
		//use default:
		int i = config_param_lookup(class, param);
		val = class->config[i]->f.dfault;
	}
	dbg (2, "'%s' %s=%i", group_name, param, val);
#endif
	return val;
}


const char*
config_get_window_setting_char(AyyiPanel* panel, const char* param)
{
	if(config->windows){
		ConfigWindow* window_settings = config_get_window(panel);
		g_return_val_if_fail(window_settings, false);

		GList* l = window_settings->params;
		for(;l;l=l->next){
			ConfigParamValue* pval = l->data;
			ConfigParam* f = pval->param;
			char* v = pval->val.c;
			if(!strcmp(f->name, param)){
				dbg(2, "%s=%s", f->name, v);
				return v;
			}
		}
	}
	//not found, so return the default value.
	ConfigParam* setting = window_class__config_lookup(AYYI_PANEL_GET_CLASS(panel), param);
	if(!setting) gwarn("%s: cannot get default. setting not found", param);
	return setting ? setting->dfault.c : NULL;
}


#ifdef HAVE_YAML_H
static void yaml_window_set_width      (ConfigWindow* window, GValue* width ){ window->width  = g_value_get_int(width);  }
static void yaml_window_set_height     (ConfigWindow* window, GValue* height){ window->height = g_value_get_int(height); }
static void yaml_window_set_x          (ConfigWindow* window, GValue* x     ){ window->x      = g_value_get_int(x);      }
static void yaml_window_set_y          (ConfigWindow* window, GValue* y     ){ window->y      = g_value_get_int(y);      }
static void yaml_window_set_depth      (ConfigWindow* window, GValue* depth ){ window->depth  = g_value_get_int(depth);  }
static void yaml_window_set_orientation(ConfigWindow* window, GValue* o     ){ window->orientation = !strcmp(g_value_get_string(o), "vertical") ? GTK_ORIENTATION_VERTICAL : GTK_ORIENTATION_HORIZONTAL; }
static void yaml_window_set_docktree   (ConfigWindow* window, GValue* d     ){
	const gchar* s = g_value_get_string(d);
	int i=0; for(;i<16;i++){
		gchar* c = g_strndup(s+i, 1);
		window->docktree[i] = atoi(c);
		g_free(c);
	}
}
static void yaml_window_set_maximised(ConfigWindow* window, GValue* o     ){ window->maximised = g_value_get_boolean(o); dbg(2, "maximised=%i", window->maximised); }

struct {char* name; int type; void (*set)(ConfigWindow*, GValue*); } yaml_windows[] = {
	{"width",       G_TYPE_INT,    yaml_window_set_width      },
	{"height",      G_TYPE_INT,    yaml_window_set_height     },
	{"x",           G_TYPE_INT,    yaml_window_set_x          },
	{"y",           G_TYPE_INT,    yaml_window_set_y          },
	{"depth",       G_TYPE_INT,    yaml_window_set_depth      },
	{"orientation", G_TYPE_STRING, yaml_window_set_orientation},
	{"dock",        G_TYPE_STRING, yaml_window_set_docktree   },
	{"maximised",   G_TYPE_BOOLEAN,yaml_window_set_maximised  },
};


static void
config_load_windows_yaml(yaml_parser_t* parser, yaml_event_t* event)
{
	//at this point in the parsing, we have just found a "windows" map. First event should be a YAML_SCALAR_EVENT...

	ConfigWindow* window = NULL;
	ConfigWindow* panel  = NULL;
	int num_windows = 0;
	config->windows = g_array_sized_new(ZERO_TERMINATED, TRUE, sizeof(ConfigWindow), 4);

	ConfigParam** params = NULL;

	char key[64] = "";
	int depth = 0;
	while(yaml_parser_parse(parser, event)){ //get the next event.
		switch (event->type) {
			case YAML_SCALAR_EVENT:
				if(event->data.scalar.anchor) dbg(0, "YAML_SCALAR_EVENT: %s => %s", event->data.scalar.anchor, event->data.scalar.tag);
				dbg(2, "YAML_SCALAR_EVENT: value='%s' %i plain=%i style=%i", event->data.scalar.value, event->data.scalar.length, event->data.scalar.plain_implicit, event->data.scalar.style);

				if(!key[0]){
					//1st half of a pair. We dont yet know what to do with it
					strncpy(key, (char*)event->data.scalar.value, 63);
				}else{
					//2nd half of a pair
					dbg(2, "      %s=%s", key, event->data.scalar.value);
					if(window){
						ConfigWindow* thing = (depth == 1) ? window : panel;
						gboolean found = FALSE;
						GValue gval = {0,};
						//check generic config-params:
						int i; for(i=0;i<G_N_ELEMENTS(yaml_windows);i++){
							if(!strcmp(yaml_windows[i].name, key)){
								found = TRUE;
								switch(yaml_windows[i].type){
									case G_TYPE_INT:
										g_value_init(&gval, G_TYPE_INT);
										g_value_set_int(&gval, atoi((char*)event->data.scalar.value));
										break;
									case G_TYPE_STRING:
										g_value_init(&gval, G_TYPE_STRING);
										g_value_set_string(&gval, (char*)event->data.scalar.value);
										break;
									case G_TYPE_BOOLEAN:
										;gboolean set = !strcmp("true", (char*)event->data.scalar.value);
										g_value_init(&gval, G_TYPE_BOOLEAN);
										g_value_set_boolean(&gval, set);
										break;
								}
								yaml_windows[i].set(thing, &gval);
								break;
							}
						}
						if(!found && params){
							//check window-specific config-params:
							i = 0;
							ConfigParam* p = NULL;
							gboolean found2 = false;
							while((p = params[i++])){
								if(!strcmp(key, p->name)){
									dbg(2, "  %s --> %s", p->name, event->data.scalar.value);
									ConfigParamValue* setting = g_new0(ConfigParamValue, 1);
									setting->param = p;
									switch(p->utype){
										case G_TYPE_STRING:
											setting->val.c = (char*)event->data.scalar.value;
											dbg(2, "char type: %s=%s", p->name, setting->val.c);
											break;
										case G_TYPE_INT:
											setting->val.i = atoi((char*)event->data.scalar.value);
											dbg(2, "int type: %s=%i", p->name, setting->val.i);
											break;
										default:
											setting->val.f = atof((char*)event->data.scalar.value);
											break;
									}
									window->params = g_list_append(window->params, setting);
									found2 = true;
									break;
								}
								if(i > 100) break;
							}
							if(!found2 && strcmp(key, "window_state") && _debug_) gwarn("unhandled: %p: %s", window, key);
						}
					}
					key[0] = '\0';
				}
				break;
			case YAML_MAPPING_START_EVENT:
				depth++;
				if(key[0]){
					dbg(2, "key=%s", key);
					switch(depth){
						case 1:
							g_array_set_size(config->windows, ++num_windows);
							dbg(2, "YAML_MAPPING_START_EVENT: new toplevel window: '%s' arraysize=%i type=", key, config->windows->len);
							window = &g_array_index(config->windows, ConfigWindow, num_windows - 1);
window->maximised = false;
							window->type = windows__panel_type_from_str(key);
							if(window->type){
								AyyiPanelClass* klass = g_type_class_peek(window->type);
								params = klass->config;
							}
							else gwarn ("bad config file. unknown window type: %s", key);
							break;
						case 2:
							if(strcmp(key, "panels")) gwarn("expected 'panels'. no other valid Shell subsections.");
							break;
						default:
							dbg(2, "Panels subsection found: '%s' depth=%i window=%p.", key, depth, window);
							if(window){
								if(panel) gerr("panel should be NULL before adding new panel.");
								window->child_panels = g_list_append(window->child_panels, panel = g_new0(ConfigWindow, 1));
								//dbg(0, "*** n_panels=%i", g_list_length(window->child_panels));
								PanelType dpt = windows__panel_type_from_str(key);
								if((panel->type = dpt)){
									AyyiPanelClass* klass = g_type_class_peek(window->type);
									params = klass->config;
								}
								else gwarn ("bad config file. unknown window type. %s", key);
							}
							else gwarn ("no window.");
							break;
					}
					key[0] = '\0';
				}
				else gwarn("mapping event has no name. depth=%i", depth);
				break;
			case YAML_MAPPING_END_EVENT:
				if(depth < 3) window = NULL;
				panel = NULL;

				dbg(2, "YAML_MAPPING_END_EVENT");
				depth--;
				if(depth < 0){ dbg(2, "done. n windows found: %i", num_windows); return; }
				break;
			default:
				gwarn("unexpected parser event type: %i", event->type);
				break;
		}
	}
}
#endif //HAVE_YAML_H


#ifndef HAVE_YAML_H
static void
config_load_windows()
{
	//load info on open windows into config->windows.

	GError* error = NULL;
	unsigned int group_count, g;
	gchar** groups = g_key_file_get_groups(app.key_file, &group_count);
	int num_windows = group_count - 1;
	config->windows = g_array_sized_new(ZERO_TERMINATED, TRUE, sizeof (ConfigWindow), num_windows);
	g_array_set_size(config->windows, num_windows);

	ConfigWindow* window;
	int i=0;
	for(g=0;g<num_windows;g++){
		window               = &g_array_index(config->windows, ConfigWindow, i);
		window->type         = win_atoi(groups[g+1]);
		window->config_group = groups[g+1];
		if(window->type){
			window->width  = g_key_file_get_integer(app.key_file, groups[g+1], "width", &error); //no need to check for errors on every key.
			window->height = g_key_file_get_integer(app.key_file, groups[g+1], "height", NULL);
			window->x      = g_key_file_get_integer(app.key_file, groups[g+1], "x", NULL);
			window->y      = g_key_file_get_integer(app.key_file, groups[g+1], "y", NULL);
			window->depth  = g_key_file_get_integer(app.key_file, groups[g+1], "depth", NULL);
			dbg(3, "%i: type=%i width=%i x=%i depth=%i\n", g, window->type, window->width, window->x, window->depth);
			i++;

			if(error){ dbg(2, "%s", error->message); g_error_free(error); error = NULL; }
		}
		else gwarn ("bad config file. unknown window type. %s", groups[g+1]);
	}
	g_array_set_size(config->windows, i);
#ifdef CLOSE_CONFIG_FILE
	//its convenient to leave it open so we can access settings directly.
	g_strfreev(groups);
#endif
}
#endif


void
config_load_window_instance(AyyiPanel* panel)
{
	// get config info specific to an individual instance

#ifdef HAVE_YAML_H
	AyyiPanelClass* klass = g_type_class_peek(G_OBJECT_TYPE(panel));

	ConfigParam** params = (ConfigParam**)klass->config;
	if(params){
		int i; for(i=0;params[i];i++){
			ConfigParam* param = params[i];

			switch(param->utype){
				case G_TYPE_INT:
					;int ival = config_get_window_setting_int(panel, param->name);
					dbg(2, "int value found: %s %i", param->name, ival);
					//if(ival < param->min.i || ival > param->max.i) gwarn("out of range: %s: %i", param->name, ival);
					if(param->set.i) param->set.i(panel, ival);
					break;
				case G_TYPE_STRING:
					;const char* cval = config_get_window_setting_char(panel, param->name);
					dbg(2, "char value found: %s %s", param->name, cval);
					if(param->set.c) param->set.c(panel, cval);
					break;
				default:
					;float fval = config_get_window_setting_float(panel, param->name); // unneccesarily slow
					dbg(2, "value found: %s %f", param->name, fval);
					if(param->set.f) param->set.f(panel, fval);
					break;
			}
		}
	}
#endif
}


#ifndef HAVE_YAML_H
static int
config_param_lookup(WindowClass* window_class, const char* param)
{
	config_param** config = window_class->config;

	int i; for(i=0;i<100;i++){
		config_param* p = config[i];
		if(!p) break;
		if(!strcmp(p->i.name, param)) return i;
	}
	gerr ("not found!");
	return 0;
}
#endif


gboolean
config_log_window_is_open()
{
	//return true if the config file reports at least one open log window.

	ConfigWindow* window;
	if(!config->windows) return false;

	int w; for(w=0;w<config->windows->len;w++){
		window = &g_array_index(config->windows, ConfigWindow, w);
		if(window->type == AYYI_TYPE_LOG_WIN) return true;
	}

	return false;
}


#if HAVE_YAML_H
static ConfigWindow*
config_get_window(AyyiPanel* panel)
{
	//config_window doesnt have instance_num prop. do we need to add it?

	g_return_val_if_fail(config->windows, NULL);

	GType d = G_OBJECT_TYPE(panel);
	int w;
	for(w=0;w<config->windows->len;w++){
		ConfigWindow* window = &g_array_index(config->windows, ConfigWindow, w);
		if(window->type == d) return window;
		if(window->child_panels){
			GList* l = window->child_panels;
			for(;l;l=l->next){
				ConfigWindow* child = l->data;
				if(child->type == d) return child;
			}
		}
	}

	if(_debug_) gwarn("no config found for window->type=%s n_windows=%i", AYYI_PANEL_CLASS(g_type_class_peek(d))->name, config->windows->len);
	return NULL;
}
#endif


void
config_make_group_name(AyyiPanelClass* class, int count, char* group_name)
{
	g_return_if_fail(class);

	snprintf(group_name, 63, "%s %i", class->name, count + 1);
}


int
config_get_n_windows(PanelType type)
{
	// Return the number of windows in the config for the specified type
	// or if not specified, the total number of windows.

	// TODO this is only really useful if there is only one shell.

	int n = 0;

	int w;
	for(w=0;w<config->windows->len;w++){
		ConfigWindow* window = &g_array_index(config->windows, ConfigWindow, w);
		if(window->type == type || !type) n++;
	}
	return n;
}


#ifndef HAVE_YAML_H
static void
config_set_int(char* group_name, const char* param, int value)
{
	static char str[256];
	snprintf(str, 255, "%i", value);
	g_key_file_set_value(app.key_file, group_name, param, str);
}


static void
config_set_float(char* group_name, const char* param, double value)
{
	static char str[256];
	snprintf(str, 255, "%.1f", value);
	g_key_file_set_value(app.key_file, group_name, param, str);
}
#endif


void
config_new()
{
	//g_key_file_has_group(GKeyFile *key_file, const gchar *group_name);

	GError* error = NULL;
	char data[256 * 256];
	sprintf(data, "#this is a Seismix application config file.\n"
		"[Seismix]\n"
		);

	if(!g_key_file_load_from_data(app->key_file, data, strlen(data), G_KEY_FILE_KEEP_COMMENTS | G_KEY_FILE_KEEP_TRANSLATIONS, &error)){
		gerr ("error creating new key_file from data. %s", error->message);
		g_error_free(error);
		error = NULL;
		return;
	}

	log_print(0, "new config file created.");
}


static gboolean
create_config_dir()
{
	gboolean ret = false;
	gchar* path = g_build_filename(g_get_home_dir(), USER_CONFIG_DIR, NULL);
	if(file_exists(path)) ret = true;
	else {
		if(!g_mkdir_with_parents(path, S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP)) ret = true;
		else gwarn("cannot create config dir: %s", path);
	}
	if(!ret) gwarn("cannot access config dir: %s", path);
	g_free(path);
	return ret;
}


#if 0 // TODO a GType so that we can save colours like 0x777777ff as numbers instead of strings.
GType
colour32_get_type (void)
{
	static GType spec_type = 0;

	if(!spec_type){
		static const GTypeInfo type_info = {
			sizeof (GParamSpecClass),
			NULL, NULL,
			(GClassInitFunc)gimp_param_unit_class_init,
			NULL, NULL,
			sizeof(GimpParamSpecUnit),
			0, NULL, NULL
		};

		spec_type = g_type_register_static (G_TYPE_PARAM_INT, "Colour32", &type_info, 0);
	}

	return spec_type;
}
#endif
