/**
* +----------------------------------------------------------------------+
* | This file is part of the Ayyi project. http://www.ayyi.org           |
* | copyright (C) 2004-2017 Tim Orford <tim@orford.org>                  |
* +----------------------------------------------------------------------+
* | This program is free software; you can redistribute it and/or modify |
* | it under the terms of the GNU General Public License version 3       |
* | as published by the Free Software Foundation.                        |
* +----------------------------------------------------------------------+
*
*/

#include "global.h"
#include <gdk/gdkkeysyms.h>

#include <model/am_time.h>
#include <model/am_utils.h>
#include "windows.h"
#include "support.h"
#include "settings.h"
#include "menu.h"
#include "statusbar.h"
#include "icon.h"
#include "panels/panel.h"
#include "shell.h"

extern SMConfig* config;
extern void song_save ();
extern GList* windowlist; // list of AyyiPanel*, one for each open window panel. DEPRECATED in favour of shell_list->panels

static AyyiShell* shell__lookup_by_window    (GtkWidget*);
static void       shell__close               (AyyiShell*);
static void       shell__on_dock_change      (GdlDockMaster*, gpointer);
static gint       shell__on_close_request    (GtkWidget*, GdkEvent*, AyyiShell*);
static void       panel__on_reparented       (GdlDockItem*, GdlDockObject*, GdlDockPlacement*, GValue*, gpointer data);
static void       shell__on_realise          (GtkWidget*, AyyiShell*);
static void       shell__on_progress         (GObject*, double, AyyiShell*);
static void       shell__remove_empty        ();
static void       shell__update_from_dock    (AyyiShell*, GdlDock*);
static AyyiShell* get_shell_from_dock_item   (GdlDockItem*);
static void       on_style_change            (GtkWidget*);


static AyyiShell*
_shell__new()
{
	AyyiShell* shell = g_new0(AyyiShell, 1);

	shell_list = g_list_append(shell_list, shell);

	shell->widget = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	g_signal_connect(G_OBJECT(shell->widget), "delete-event", G_CALLBACK(shell__on_close_request), shell);
	g_signal_connect(G_OBJECT(shell->widget), "realize",      G_CALLBACK(shell__on_realise), shell);

	gtk_widget_show(shell->vbox = gtk_vbox_new(NON_HOMOGENOUS, 0));
	gtk_container_add(GTK_CONTAINER(shell->widget), GTK_WIDGET(shell->vbox));

	gtk_window_add_accel_group(GTK_WINDOW(shell->widget), app->global_accel_group); // set global keyboard shortcuts

	gtk_window_set_icon((GtkWindow*)shell->widget, get_icon("arrange", 16));

	void shell__on_style_set (GtkWidget* widget, GtkStyle* previous_style, gpointer user_data)
	{
		if(!app->styles_are_set){
			get_style_font();
			on_style_change(widget);
		}
	}
	g_signal_connect(G_OBJECT(shell->widget), "style-set", G_CALLBACK(shell__on_style_set), shell);

	am_song__connect("progress", G_CALLBACK(shell__on_progress), shell);

	return shell;
}


/*
 *  This is an internal fn. To add a new shell to a running setup, use shell__new_floating instead
 */
AyyiShell*
shell__new()
{
	PF;

	AyyiShell* shell = _shell__new();
	dbg(2, "adding shell to shell_list: dock=%s", shell->dock ? GDL_DOCK_OBJECT(shell->dock)->name : "<unnamed>");

	char win_id[32];
	static int i = 0;
	sprintf(win_id, "Shell %i", i++); // this is an empty shell. Name is set later when first child is added.

	void shell__on_dock_item_added(GdlDockMaster* master, gpointer item, gpointer _shell)
	{
		g_return_if_fail(item);

		dbg(1, "-> item added! %s", ((GdlDockObject*)item)->name);

		AyyiShell* shell = NULL;

		if(GDL_IS_DOCK(item)){
			shell = shell__lookup_by_dock((GdlDock*)item);
			GtkWidget* top = gtk_widget_get_toplevel((GtkWidget*)item);
			if(GTK_IS_WINDOW(top)){
				if(!shell->panels) shell__update_from_dock (shell, (GdlDock*)item);
			}else{
				if(!shell){
					shell = shell__new_from_dock((GdlDock*)item);
					shell__update_from_dock (shell, (GdlDock*)item);
				}
			}
		}else if(GDL_IS_DOCK_PANED(item)){
			// ignore these - wait for the top level dock

		}else{
			GType parent_type = G_TYPE_FROM_CLASS(g_type_class_peek_parent(G_OBJECT_GET_CLASS(item)));
			if(parent_type == AYYI_TYPE_PANEL){
				AyyiPanel* panel = item;

				GtkWidget* top = gtk_widget_get_toplevel((GtkWidget*)item);
				if(GDL_IS_DOCK(top)){
					shell = shell__lookup_by_dock((GdlDock*)top);
					if(!shell){
						shell = shell__new_from_dock((GdlDock*)top);
						gtk_widget_show_all(shell->widget);

						GList* panels = gtk_container_get_children((GtkContainer*)top);
						if(panels){
							AyyiPanel* panel = panels->data;
							if(((GdlDockObject*)panel)->long_name)
								gtk_window_set_title(GTK_WINDOW(shell->widget), ((GdlDockObject*)panel)->long_name);
							g_list_free(panels);
						}
					}
				}
				else if(GTK_IS_WINDOW(top))
					shell = shell__lookup_by_window(top);
				else if(GDL_IS_DOCK_PANED(top))
					return;
				else
					dbg(0, "unexpected item type");

				shell__add_panel(shell, panel);
			}
		}
	}

	if(!app->master_dock){
		GtkWidget* dock = gdl_dock_new();
		app->master_dock = shell->dock = GDL_DOCK(dock);
		g_signal_connect(GDL_DOCK_MASTER(GDL_DOCK_OBJECT(app->master_dock)->master), "layout-changed", G_CALLBACK(shell__on_dock_change), app->master_dock);
		g_signal_connect(GDL_DOCK_MASTER(GDL_DOCK_OBJECT(app->master_dock)->master), "dock-item-added", G_CALLBACK(shell__on_dock_item_added), shell);
	}
	else shell->dock = GDL_DOCK(gdl_dock_new_from(app->master_dock, FALSE));

	char dock_name[32]; snprintf(dock_name, 31, "%s Shell", win_id);
	GDL_DOCK_OBJECT(shell->dock)->long_name = g_strdup(dock_name);

	//what do these actually do?
	char place_holder_name[32];
	snprintf(place_holder_name, 31, "%s ph1", win_id);
	gdl_dock_placeholder_new (place_holder_name, GDL_DOCK_OBJECT(shell->dock), GDL_DOCK_TOP, FALSE);
	snprintf(place_holder_name, 31, "%s ph2", win_id);
	gdl_dock_placeholder_new (place_holder_name, GDL_DOCK_OBJECT(shell->dock), GDL_DOCK_BOTTOM, FALSE);
	snprintf(place_holder_name, 31, "%s ph3", win_id);
	gdl_dock_placeholder_new (place_holder_name, GDL_DOCK_OBJECT(shell->dock), GDL_DOCK_LEFT, FALSE);
	snprintf(place_holder_name, 31, "%s ph4", win_id);
	gdl_dock_placeholder_new (place_holder_name, GDL_DOCK_OBJECT(shell->dock), GDL_DOCK_RIGHT, FALSE);

#if 0
	void shell_on_allocate (GtkWidget* widget, GdkRectangle* allocation, gpointer user_data)
	{
		PF2;
	}
	g_signal_connect(G_OBJECT(shell->widget), "size-allocate", G_CALLBACK(shell_on_allocate), shell);
#endif

	gtk_box_pack_start(GTK_BOX(shell->vbox), (GtkWidget*)shell->dock, EXPAND_TRUE, FILL_TRUE, 0);

	GtkWidget* window = shell->widget;
	if(g_list_length(shell_list) == 1) gtk_window_set_title(GTK_WINDOW (window), "Ayyi");

	return shell;
}


AyyiShell*
shell__new_floating (AyyiPanelClass* class)
{
	void shell__add_panel_to_dock(AyyiShell* shell, AyyiPanel* panel, GdlDockPlacement placement, GdlDockObject* parent)
	{
		// called from: shell__add_panel_().

		if(parent) g_return_if_fail(GDL_IS_DOCK_OBJECT(parent));
		dbg(2, "name=%s parent=%s", GDL_DOCK_OBJECT(panel)->name, parent ? GDL_DOCK_OBJECT(parent)->name : "<empty>");
		if(parent && GDL_IS_DOCK_ITEM(parent)){
			dbg(3, "path1............");

			// can happen if docking of previous panels failed.
			g_return_if_fail(gdl_dock_object_get_parent_object(parent));

			GDL_DOCK_OBJECT_CLASS(AYYI_PANEL_GET_CLASS(panel))->dock (parent, GDL_DOCK_OBJECT(panel), placement, NULL);
		}
		else if(parent && (g_list_length(shell->panels) > 1)){
			dbg(2, "path2............");
			gdl_dock_item_dock2(parent, GDL_DOCK_ITEM(panel), placement);
		}
		else{
			dbg(2, "path3...........");
			gdl_dock_add_item (GDL_DOCK(shell->dock), GDL_DOCK_ITEM(panel), placement);
		}

		//the signal we created for this seems to be not needed.
		//g_signal_connect (GDL_DOCK_OBJECT(dock_item), "item_reparented", G_CALLBACK (panel__on_reparented), NULL);

		//g_signal_connect(dock_item, "dock_drag_begin", G_CALLBACK(gdl_dock_master_drag_begin), master);
		//g_signal_connect(shell->dock_item, "dock_drag_end", G_CALLBACK(gdl_dock_master_drag_end), master);
		//g_signal_connect(dock_item, "dock_drag_begin", G_CALLBACK(window__on_dock_drag_begin), NULL);
	}

	AyyiShell* shell = shell__new();
	AyyiPanel* panel = class->new
		? (AyyiPanel*)class->new(NULL, NULL, GTK_ORIENTATION_HORIZONTAL)
		: g_object_new(G_OBJECT_CLASS_TYPE(class), NULL);

	shell__add_panel(shell, panel);

	GtkOrientation orientation = GTK_ORIENTATION_HORIZONTAL;
	GdlDockPlacement placement = (orientation == GTK_ORIENTATION_VERTICAL) ? GDL_DOCK_BOTTOM : GDL_DOCK_RIGHT;
	GdlDockObject* parent = NULL;
	shell__add_panel_to_dock(shell, panel, placement, parent);

	ayyi_panel_on_new(panel, NULL, GTK_ORIENTATION_HORIZONTAL); // set size
									// TODO this is now also called in an idle
	call(AYYI_PANEL_GET_CLASS(panel)->on_ready, panel);

	if(_debug_) print_shelllist_items();

	return shell;
}


AyyiShell*
shell__new_from_dock(GdlDock* dock)
{
	PF;
	AyyiShell* shell = _shell__new();
	shell->dock = dock;

	if(!((GtkWidget*)dock)->parent) gtk_box_pack_start(GTK_BOX(shell->vbox), (GtkWidget*)dock, EXPAND_TRUE, FILL_TRUE, 0);

	return shell;
}


void
shell__free(AyyiShell* shell)
{
	PF;
	g_return_if_fail(shell);
	g_return_if_fail(g_list_find(shell_list, shell));

	shell_menu_free(shell);

	g_signal_handlers_disconnect_matched (song, G_SIGNAL_MATCH_FUNC | G_SIGNAL_MATCH_DATA, 0, 0, NULL, shell__on_progress, shell);

	shell_list = g_list_remove(shell_list, shell);

	g_list_free0(shell->panels);

	g_free(shell);
}


AyyiShell*
shell__lookup_by_dock (GdlDock* dock)
{
	GList* l = shell_list;
	for(;l;l=l->next){
		if(((AyyiShell*)l->data)->dock == dock) return l->data;
	}
	return NULL;
}


static AyyiShell*
shell__lookup_by_window (GtkWidget* window)
{
	GList* l = shell_list;
	for(;l;l=l->next){
		if(((AyyiShell*)l->data)->widget == window) return l->data;
	}
	return NULL;
}


void
shell__add_panel (AyyiShell* shell, AyyiPanel* panel)
{
	g_return_if_fail(shell);

	dbg(1, "adding: %s to: %s", ((GdlDockObject*)panel)->long_name, ((GdlDockObject*)shell->dock)->long_name);

	if(panel->shell) gwarn("panel already has a shell");

	panel->shell = shell;
	g_return_if_fail(!g_list_find(shell->panels, panel));
	shell->panels = g_list_append(shell->panels, panel);

	// the windowlist is not modified during redocking so the panel may already be present
	if(!g_list_find(windowlist, panel)){
		windowlist = g_list_append(windowlist, panel); // dont use prepend as it messes up instance counting.
	}

	AyyiPanelClass* panel_class = AYYI_PANEL_CLASS(G_OBJECT_GET_CLASS(panel));
	if(panel_class->has_statusbar && !shell->statusbar.widget[0]){
		GtkWidget* hbox_sb = statusbar__new(shell);
		gtk_box_pack_end((GtkBox*)shell->vbox, hbox_sb, EXPAND_FALSE, FILL_FALSE, 0);
	}

	config_load_window_instance(panel);

	// add menu if panel type requests it, or is the first window
	if(!shell->window_menu && (g_list_length(shell_list) == 1 || panel_class->has_menubar)){
		GtkWidget* menu = shell_menu_new(shell);
		if(menu){
			gtk_box_pack_start(GTK_BOX(shell->vbox), menu, EXPAND_FALSE, FILL_FALSE, 0);
			gtk_box_reorder_child(GTK_BOX(shell->vbox), menu, 0);
		}
	}

	menu_update_windows(); 

	// TODO each of these signals must be disconnected preferably in shell__remove_panel

	g_signal_connect (GDL_DOCK_OBJECT(panel), "dock", G_CALLBACK (panel__on_reparented), panel);

	void panel__on_reparented2(GtkWidget* widget, GtkWidget* old_parent, gpointer user_data)
	{
		// during a move, a widget is first removed, then re-added

		GdlDock* find_parent_dock(GtkWidget* widget)
		{
			while((widget = gtk_widget_get_parent(widget))){
				if(GDL_IS_DOCK(widget)) return (GdlDock*)widget;
			}
			return NULL;
		}

		bool idle_add_to_shell(gpointer _widget)
		{
			GtkWidget* widget = _widget;
			g_return_val_if_fail(AYYI_IS_PANEL(widget), G_SOURCE_REMOVE);

			AyyiShell* shell = lookup_shell_by_widget((GtkWindow*)gtk_widget_get_toplevel(widget->parent));
			if(!shell){
				GdlDockObject* dock_parent = gdl_dock_object_get_parent_object((GdlDockObject*)widget);
				if(GDL_IS_DOCK(dock_parent)){
					// it looks like a new floating window has been created that does not yet have a shell
					shell = shell__new_from_dock(GDL_DOCK(dock_parent));
				}
			}
			if(shell){
				if(!g_list_find(shell->panels, widget)){
					dbg(2, "has parent: shell=%s", ((GdlDockObject*)shell->dock)->long_name);
					shell__add_panel(shell, (AyyiPanel*)widget);
				}
			}
			else dbg(0, "*** shell not found. cannot add. toplevel=%s parent_dock=%p", G_OBJECT_CLASS_NAME(G_OBJECT_GET_CLASS(gtk_widget_get_toplevel(widget->parent))), find_parent_dock(widget));

			return G_SOURCE_REMOVE;
		}

		if(widget->parent){
			// parent is likely to be a GtkPaned or some other container widget
			dbg(2, "has parent: widget=%s parent=%s (%s)", ((GdlDockObject*)widget)->long_name, GDL_IS_DOCK_OBJECT(widget->parent) ? ((GdlDockObject*)widget->parent)->long_name : "", G_OBJECT_CLASS_NAME(G_OBJECT_GET_CLASS(widget->parent)));

			// sometimes has a root window at this point, sometimes not.
			GtkWidget* toplevel = gtk_widget_get_toplevel(widget->parent);
			if(GTK_IS_WINDOW(toplevel)){
				idle_add_to_shell(widget);
			}else{
				g_idle_add(idle_add_to_shell, widget);
			}

		}else{
			dbg(2, "widget does not have parent: %s", ((GdlDockObject*)widget)->long_name);
			AyyiShell* old_shell = lookup_shell_by_widget((GtkWindow*)gtk_widget_get_toplevel(old_parent));
			if(old_shell){
				old_shell->panels = g_list_remove(old_shell->panels, widget);
				((AyyiPanel*)widget)->shell = NULL;
			}
		}
	}
	g_signal_connect (panel, "parent-set", G_CALLBACK (panel__on_reparented2), panel);

	void shell_panel_finalize_notify(gpointer _shell, GObject* was)
	{
		PF;
		if(was == (GObject*)app->active_panel){
			app->active_panel = NULL;
			((AyyiShell*)_shell)->focussed_panel = NULL;
		}
	}
	g_object_weak_ref((GObject*)panel, shell_panel_finalize_notify, shell);
}


void
shell__destroy_panel(AyyiShell* shell, AyyiPanel* panel)
{
	PF;
	g_return_if_fail(panel);

	GtkWidget* widget = GTK_WIDGET(shell->widget);

	shell__remove_panel(shell, panel);

	dbg(1, "destroying panel widget... %s", G_OBJECT_CLASS_NAME(G_OBJECT_GET_CLASS(panel)));
	gtk_widget_destroy(GTK_WIDGET(panel));

	if(!shell->panels){
		shell__free(shell);
		gtk_widget_destroy(widget);
	}
}


void
shell__remove_panel(AyyiShell* shell, AyyiPanel* panel)
{
	dbg(1, "shell=%s panel=%s", GDL_DOCK_OBJECT(shell->dock)->long_name, GDL_DOCK_OBJECT(panel)->long_name);
	g_return_if_fail(panel);

	windowlist = g_list_remove(windowlist, panel);

	// unnamed objects are not in the hashtable so this does very little if anything
	gdl_dock_master_remove(GDL_DOCK_MASTER(GDL_DOCK_OBJECT(panel)->master), GDL_DOCK_OBJECT(panel));

	shell->panels = g_list_remove(shell->panels, panel);

	menu_update_windows();
}


int
shell__get_n_panels(AyyiShell* shell)
{
	return g_list_length(shell->panels);
}


static void
shell__on_dock_change(GdlDockMaster* master, gpointer data)
{
	// Gets called following a "layout-changed" signal. This doesnt tell us what has changed.
	// The potential advantage of this signal is that it occurs _after_ an operation has _completed_.
	// Change can be new dock item added, or re-arrange

	windows__print_dock();

	shell__remove_empty(); // this runs in an idle but from here it could run directly.
}


static void
shell__remove_empty()
{
	bool check()
	{
		GList* l = shell_list;
		for(;l;l=l->next){
			AyyiShell* shell = l->data;
			dbg(2, "shell=%s%s%s n_items=%i", bold, GDL_DOCK_OBJECT(shell->dock)->name, white, g_list_length(shell->panels));
			if(!shell->panels){
				dbg(2, "shell '%s' is empty.", GDL_DOCK_OBJECT(shell->dock)->long_name);
				shell__close(shell);
				break;
			}
		}
		return G_SOURCE_REMOVE;
	}

	g_idle_add(check, NULL);
}


/*
 *  Updates shell/panel properties following a re-dock
 *
 *  @param requestor - the item that has been reparented.
 */
static void
panel__on_reparented(GdlDockItem* item, GdlDockObject* requestor, GdlDockPlacement* position, GValue* other_data, gpointer _panel)
{
	AyyiPanel* dest_panel = (AyyiPanel*)_panel;
	AyyiPanel* src_panel = (AyyiPanel*)requestor;

	AyyiShell* dest_shell = dest_panel->shell = get_shell_from_dock_item(item);
	dbg(0, "item='%s' requestor='%s' dest_shell='%s'", GDL_DOCK_OBJECT(item)->long_name, requestor->long_name, GDL_DOCK_OBJECT(dest_shell->dock)->long_name);

	//TODO change to use shell__remove_dock_item - take care when modifying list in loop.

	//update all dock_item lists:
	AyyiShell* src_shell = src_panel ? src_panel->shell : NULL;
	dbg(2, "src_panel='%s'", src_panel ? GDL_DOCK_OBJECT(src_panel)->long_name: "");
	if(src_shell) shell__remove_panel(src_shell, AYYI_PANEL(requestor));

	shell__remove_empty();

	//TODO try not to get here if the item is already in the list.
	if(_debug_ > -1 && g_list_find(dest_shell->panels, requestor)) gwarn("already in list!");
	if(!g_list_find(dest_shell->panels, requestor)){
		dbg(2, "adding %s%s%s %s", bold, requestor->name, white, G_OBJECT_CLASS_NAME(G_OBJECT_GET_CLASS(requestor)));
		dest_shell->panels = g_list_append(dest_shell->panels, requestor);
	}

	if(AYYI_IS_PANEL(item) && !g_list_find(dest_shell->panels, item)){
		dest_shell->panels = g_list_append(dest_shell->panels, item);
	}

	if(src_panel){
		src_panel->shell = dest_panel->shell;
	}
	else gwarn("src shell not found.");
}


/*
 *  Callback for button press on toplevel window.
 *  Connected to toplevel window close buttons.
 */
static gint
shell__on_close_request(GtkWidget* widget, GdkEvent* event, AyyiShell* shell)
{
	PF;

#warning TODO check song is saved properly when closing last shell window
	if(g_list_length(shell_list) == 1){
		song_save();
	}

	GList* panels = shell__get_panels(shell);

#if 0
	GList* l = shell->panels;
	for(;l;l=l->next){
		GdlDockObject* item = l->data;
		gdl_dock_master_remove(GDL_DOCK_MASTER(item->master), item);
	}
#endif

	dbg(1, "shell has %i panels", g_list_length(panels));
	if(panels){
		GList* l = panels;
		for(;l;l=l->next){
			AyyiPanel* panel = l->data;
			shell__destroy_panel(shell, panel);
		}
		g_list_free(panels);
	}

	PF_DONE;

	if(_debug_ > 1) print_shelllist_items();

	if(!shell_list) app_quit();

	return NOT_HANDLED;
}


/*
 *  Close a window and all its panels
 */
static void
shell__close(AyyiShell* shell)
{
	PF;

	if(!g_list_find(shell_list, shell)){
		dbg(0, "shell not found. already removed?");
		return;
	}

	GList* l = shell->panels;
	if(l){
		for(;l;l=l->next){
			AyyiPanel* panel = l->data;
			shell__destroy_panel(panel->shell, panel);
		}
	}else{
		dbg(1, "destroying window %p...", shell->widget);
		gtk_widget_destroy(shell->widget);
		shell->widget = NULL;
		shell__free(shell);
	}
}


static void
shell__on_realise(GtkWidget* widget, AyyiShell* shell)
{
	static bool first = true;

	if(first){
		first = false;

		int get_layout_height(GtkWidget* widget)
		{
			PangoLayout* layout = gtk_widget_create_pango_layout(widget, "bygA1");
			char font_string[64];
			get_font_string(font_string, 0);

			PangoFontDescription* font_desc = pango_font_description_from_string(font_string);
			pango_layout_set_font_description(layout, font_desc);

			PangoRectangle ink_rect;
			PangoRectangle logical_rect;
			pango_layout_get_pixel_extents(layout, &ink_rect, &logical_rect);
			dbg(2, "line_height=%i %i", ink_rect.height, logical_rect.height);
			return MAX(ink_rect.height, logical_rect.height);
		}

		app->line_height = get_layout_height(widget);
		get_style_font();
		g_signal_emit_by_name (app, "styles-available", NULL);
	}
}


static void
shell__on_progress(GObject* _song, double val, AyyiShell* shell)
{
	// FIXME using progressbar can cause windows to become wider (eg pool window)

	g_return_if_fail(shell);
	if(shell->statusbar.progress && shell->widget->allocation.width > 480){
		if(!GTK_WIDGET_VISIBLE(shell->statusbar.progress)) gtk_widget_show(shell->statusbar.progress);
		gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(shell->statusbar.progress), val);
	}
}


gboolean
shell__has_panel_type(AyyiShell* shell, PanelType type)
{
	// Return TRUE if the shell contains a visible panel of the given type.

	int
	get_page(GdlDockNotebook* container, GdlDockItem* child)
	{
		//return the position of the child within a notebook.

		gboolean found = false;
		int n = 0;
		GList* children = gtk_container_get_children(GTK_CONTAINER(container));
		if(children){
			GList* i = children;
			for(;i;i=i->next){
				GtkWidget* w = i->data;
				if(w == (GtkWidget*)child) {found = true; break; };
				n++;
			}
			g_list_free(children);
		}
		return found ? n : -1;
	}

	panel_foreach {
		if(G_OBJECT_TYPE(panel) == type){
			GdlDockItem* item = (GdlDockItem*)panel;
			GdlDockObject* parent = gdl_dock_object_get_parent_object(GDL_DOCK_OBJECT(item));
			if(GDL_IS_DOCK_NOTEBOOK(parent)){
				int visible_tab;
				g_object_get(parent, "page", &visible_tab, NULL);
				dbg(2, "is notebook! page=%i panel=%i", visible_tab, get_page(GDL_DOCK_NOTEBOOK(parent), item));
				if(get_page(GDL_DOCK_NOTEBOOK(parent), item) != visible_tab) dbg(0, "not visible!");
				if(get_page(GDL_DOCK_NOTEBOOK(parent), item) != visible_tab) return false;
			}
			return true;
		}
	} end_panel_foreach;

	return false;
}


/*
 *  Find a panel instance of the specified type. If there is more than one, the focused panel takes precedence.
 */
AyyiPanel*
shell__find_panel (AyyiShell* shell, PanelType dt)
{
	g_return_val_if_fail(shell, NULL);

	AyyiPanel* primary = NULL;

	GList* p = shell->panels;
	for(;p;p=p->next){
		if(G_OBJECT_TYPE(p->data) == dt){
			AyyiPanel* panel = p->data;
			if(panel == app->active_panel){
				primary = panel; // active panel takes precedence
				break;
			}else{
				if(!primary) primary = panel;
			}
		}
	}

	return primary;
}


/*
 *  Return a newly allocated list of GdlDockItem's.
 */
GList*
shell__get_dock_items(AyyiShell* shell)
{
	// -perhaps gdl-dock has a better way of doing this?

	// note: after a drop, the dock-items are contained inside other dock-items. Is that supposed to happen?

	GList* ret = NULL;
	GList* l = gdl_dock_get_named_items(shell->dock); //this gets dock items for ALL docks, but only if they are named.
	GList* i = l;
	for(;i;i=i->next){
		GdlDockItem* di = i->data;
		int depth = 0;
		//gdl_dock_item_print(GDL_DOCK_ITEM(di));
		GdlDockObject* parent = gdl_dock_object_get_parent_object(GDL_DOCK_OBJECT(di));
		depth++;
		if(GDL_IS_DOCK_ITEM(parent)){
			parent = gdl_dock_object_get_parent_object(parent);
			depth++;
			if(!GDL_IS_DOCK(parent)){
				if(GDL_IS_DOCK_ITEM(parent)){

					parent = gdl_dock_object_get_parent_object(parent);
					depth++;
					/*
					if(!GDL_IS_DOCK(parent)){
					gwarn("(%s) parent3 is not Dock! parent=%p parent->name=%s", GDL_DOCK_OBJECT(di)->name, parent, parent->name);
					}
					*/
					if(GDL_IS_DOCK_ITEM(parent)){
						parent = gdl_dock_object_get_parent_object(parent);
						depth++;
						if(!GDL_IS_DOCK(parent)){
							gwarn("(%s) parent4 is not Dock! parent=%p parent->name=%s", GDL_DOCK_OBJECT(di)->name, parent, parent->name);
						}
						else dbg(2, "item is at depth: %i", depth);
						//TODO we quickly get to this level of nestedness - need to add more recursion.
					}
				}
			}
		}
		else if(GDL_IS_DOCK_MASTER(parent)){
			dbg(0, "parent is dock-master.");
			continue;
		}
		else if(GDL_IS_DOCK_PLACEHOLDER(parent)){
			dbg(0, "parent is placeholder.");
			continue;
		}
		else if(GDL_IS_DOCK(parent)){
		}
		else if(!parent){
			gwarn("parent null. scheduled for deletion? (item=%s)", GDL_DOCK_OBJECT(di)->name);
			continue;
		}
		else {
			gwarn("unknown parent type. item=%s parent=%s", GDL_DOCK_OBJECT(di)->name, parent ? parent->name : "");
			continue;
		}

		dbg(2, "parent='%s'", parent ? parent->name : "");
		if(parent == GDL_DOCK_OBJECT(shell->dock)) ret = g_list_append(ret, di);
	}
	g_list_free(l);
	dbg(2, "n_items=%i", g_list_length(ret));
	return ret;
}


/*
 *  Return a newly allocated list of AyyiPanel*.
 */
GList*
shell__get_panels(AyyiShell* shell)
{
	return g_list_copy(shell->panels);
}


void
shell__menu_popup(AyyiShell* shell, GdkEvent* event)
{
	if(!shell->context_menu) shell->context_menu = shell_context_menu_new(shell); // context menu is accessed from the top of the track ctl area.
	gtk_menu_popup(GTK_MENU(shell->context_menu), NULL, NULL, NULL, event, event->button.button, (guint32)(((GdkEventButton*)event)->time));
}


void
shell__statusbar_print(AyyiShell* shell, int n, const char* fmt, ...)
{
	if(n < STATUSBAR_MAIN || n >= STATUSBAR_MAX) n = STATUSBAR_MAIN;

	if(!shell){
		if(!(shell = ((AyyiPanel*)app->latest_arrange)->shell)) return;
	}

	va_list ap;
	char text[256];
	if (fmt == NULL)
		*text = '\0';
	else {
		va_start(ap, fmt);
		vsprintf(text, fmt, ap);
		va_end(ap);
	}

	GtkWidget* statusbar = shell->statusbar.widget[n];

	if(!statusbar) return;

	gint cid = gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar), "Song");
	gtk_statusbar_push(GTK_STATUSBAR(statusbar), cid, text);

	//if(shell->statusbar_icon) gtk_widget_hide(shell->statusbar_icon);
}


void
shell__statusbar_print_if_panel(AyyiShell* shell, int n, PanelType type, const char* fmt, ...)
{
	//only print to the statusbar if the shell contains a panel of Type, and it is visible.

	va_list ap;
	char text[256];
	if (fmt == NULL)
		*text = '\0';
	else {
		va_start(ap, fmt);
		vsnprintf(text, 255, fmt, ap);
		va_end(ap);
	}

	if(shell__has_panel_type(shell, type)){
		shell__statusbar_print(shell, n, text);
	}
}


#ifdef UNUSED
void
shell__statusbar_print_bbs(AyyiShell* shell, int n, GPos* pos)
{
	g_return_if_fail(pos);
	char bbs[64];
	pos2bbs(pos, bbs);
	shell__statusbar_print(shell, n, bbs);
}
#endif


/*
 *  Return newly allocated list of AyyiPanel*
 */
static GList*
shell__get_dock_panels (GdlDock* dock)
{
	typedef struct {
		GList* panels;
	} C;
	C c = {0,};

	void each_(GdlDockObject* object, gpointer _c)
	{
		C* c = _c;
		//windows__print_dock_object(object, "  ");
		GType parent_type = G_TYPE_FROM_CLASS(g_type_class_peek_parent(G_OBJECT_GET_CLASS(object)));
		if(parent_type == AYYI_TYPE_PANEL){
			c->panels = g_list_append(c->panels, object);
		}

		if(gdl_dock_object_is_compound (object)){
			gtk_container_foreach (GTK_CONTAINER (object), (GtkCallback)each_, c);
		}
	}
	each_((GdlDockObject*)dock, &c);

	return c.panels;
}


static void
shell__update_from_dock (AyyiShell* shell, GdlDock* dock)
{
	g_list_free0(shell->panels);

	GList* panels = shell__get_dock_panels(dock);
	GList* l = panels;
	for(;l;l=l->next){
		shell__add_panel(shell, l->data);
	}
	g_list_free(panels);
}


static AyyiShell*
get_shell_from_dock_item(GdlDockItem* item)
{
	// This is for use when panel->shell is _not_ set.

	GtkWidget* top_level_window = gtk_widget_get_toplevel(GTK_WIDGET(item));
	GList* l = shell_list;
	for(;l;l=l->next){
		AyyiShell* shell = l->data;
		if(shell->widget == top_level_window) return shell;
	}
	gwarn("shell not found.");
	return NULL;
}


void
print_shelllist_items()
{
	if(!shell_list){ dbg(0, "shelllist empty"); return; }

	dbg(0, "");
	GList* l = shell_list;
	for(;l;l=l->next){
		AyyiShell* s = l->data;
		GList* i = s->panels;
		printf("  shell %s%s%s:\n", bold, GDL_DOCK_OBJECT(s->dock)->long_name, white);
		if(!i) printf("    <no items>\n");
		for(;i;i=i->next){
			GdlDockItem* di = i->data;

#if 0
			GValue name = {0,};
			g_value_init(&name, G_TYPE_STRING);
			g_object_get_property(G_OBJECT(di), "long-name", &name);
#endif

			printf("    %s     %s\n", ((GdlDockObject*)di)->long_name, get_orientation_string(di->orientation));
		}
	}
}


static void
on_style_change(GtkWidget* widget)
{
	gboolean
	now_set_colours(gpointer _widget)
	{
		GtkWidget* widget = GTK_WIDGET(_widget);
		if(widget->style){
			GtkStyle* style = gtk_style_copy(gtk_widget_get_style(widget));
			am_palette_set_style(song->palette, style);
			g_object_unref(style);
			app->styles_are_set = true;
		}
		else gwarn("widget still not realised!");
		return false;
	}

	if(widget->style){
		now_set_colours(widget);
	}else{
		// this path never used.
		g_idle_add((GSourceFunc)now_set_colours, widget);
	}
}

