/*
  svgbutton.c - multi-state button with svg images.

  This file is part of the Ayyi Project. http://www.ayyi.org
  copyright (C) 2007-2016 Tim Orford <tim@orford.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 3
  as published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include "global.h"
#include <sys/time.h>
#include <gtk/gtk.h>
#include "gtk/gtksignal.h"
#include "gtk/gtktogglebutton.h"
#include "svgbutton.h"
#include "icon.h" //needed for icon cache.

extern SMConfig* config;

static void svgbutton_class_init    (SvgButtonClass*);
static void svgbutton_init          (SvgButton*);
static void svgbutton_finalize      (GObject*);
static void svgbutton_clicked       (GtkButton*);
static void svgbutton_set_state     (SvgButton*, struct _state_item*);

static GtkButtonClass* parent_class = NULL;


guint
svgbutton_get_type ()
{
	static guint svgb_type = 0;

	if(!svgb_type){
		GtkTypeInfo svgb_info = {
			"SvgButton",
			sizeof (SvgButton),
			sizeof (SvgButtonClass),
			(GtkClassInitFunc) svgbutton_class_init,
			(GtkObjectInitFunc) svgbutton_init,
			NULL,
		};
		svgb_type = gtk_type_unique(gtk_button_get_type(), &svgb_info);
	}

	return svgb_type;
}


static void
svgbutton_class_init(SvgButtonClass *class)
{
	GObjectClass* gobject_class = G_OBJECT_CLASS(class);
	GtkButtonClass* button_class = GTK_BUTTON_CLASS(class);

	parent_class = g_type_class_peek_parent(class);

	button_class->clicked = svgbutton_clicked;

	gobject_class->finalize = svgbutton_finalize;

	class->svgbutton = NULL;
}


static void
svgbutton_init(SvgButton* button)
{
	button->size = 16;
}


static void
svgbutton_finalize(GObject *object)
{
	SvgButton* button = SVGBUTTON(object);

	g_return_if_fail(IS_SVGBUTTON(object));

#ifdef HAVE_GTK_2_22
	g_list_free_full(button->state_list, g_free);
#else
	g_list_free(button->state_list);
#endif
	button->state_list = NULL;

	G_OBJECT_CLASS (parent_class)->finalize (object);
}


GtkWidget*
svgbutton_new()
{
	return GTK_WIDGET(gtk_type_new(svgbutton_get_type()));
}


static GdkPixbuf*
svgbutton_new_state_pixbuf(SvgButton* button, SvgButtonState* state)
{
	GdkPixbuf* pixbuf = get_icon(state->svg_path, button->size);
	if(!pixbuf) gwarn("svg load failed '%s'.\n", state->svg_path);

	return pixbuf;
}


char*
svgbutton_get_state(SvgButton* button)
{
	return button->current_state->name;
}


void
svgbutton_set_state_n(SvgButton* button, int n)
{
	SvgButtonState* state = g_list_nth_data(button->state_list, n);
	if(state) svgbutton_set_state(button, state);
	else gwarn("bad state requested: state=%i", n);
}


static void
svgbutton_set_state(SvgButton* button, SvgButtonState* state)
{
	//TODO it might be better to just swap the pixbuf rather than swapping the GtkImage

	if(!state) return;
	GtkContainer* container = GTK_CONTAINER(button);

	button->mouseover_text = state->name;

	g_return_if_fail(state->image);
	//add the image to the button:
	GList* children;
	if((children = gtk_container_get_children(container))) gtk_container_remove(container, children->data);
	gtk_container_add(container, state->image);

	button->current_state = state;
}


void
svgbutton_add_state(SvgButton* button, const char* state_name, const char* svg_path)
{
	SvgButtonState* state = g_malloc0(sizeof(*state));
	g_strlcpy(state->name,     state_name, 256);
	g_strlcpy(state->svg_path, svg_path,   256);
	state->pixbuf = svgbutton_new_state_pixbuf(button, state);

	// put the pixbuf into an image
	state->image = gtk_image_new_from_pixbuf(state->pixbuf);
	gtk_widget_show(state->image);
	g_object_ref(state->image); // prevent the image from being deleted by gtk.

	button->state_list = g_list_append(button->state_list, state);

	// if the button doesnt yet have a state, set it
	if(!gtk_container_get_children(GTK_CONTAINER(button))) svgbutton_set_state(button, state);
}


void
svgbutton_set_size(SvgButton* button, int size)
{
	if(size < 4 || size > 800){ gwarn("bad size (%i).\n", size); return; }

	if(size == button->size) return;

	button->size = size;

	GList* l = button->state_list;
	for(;l;l=l->next){
		SvgButtonState* state = l->data;
		if(state->pixbuf){
			g_object_unref(state->pixbuf);
			state->pixbuf = NULL;

			state->pixbuf = svgbutton_new_state_pixbuf(button, state);
			gtk_image_set_from_pixbuf((GtkImage*)state->image, state->pixbuf);
		}
	}
}


static void
svgbutton_clicked(GtkButton *button)
{
	SvgButton* svgbutton = SVGBUTTON(button);

	//switch to the next state
	GList* list = g_list_find(svgbutton->state_list, svgbutton->current_state);
	GList* next = list->next ? list->next : svgbutton->state_list;
	svgbutton_set_state(svgbutton, next->data);
}


