/**
* +----------------------------------------------------------------------+
* | This file is part of the Ayyi project. http://www.ayyi.org           |
* | copyright (C) 2004-2017 Tim Orford <tim@orford.org>                  |
* +----------------------------------------------------------------------+
* | This program is free software; you can redistribute it and/or modify |
* | it under the terms of the GNU General Public License version 3       |
* | as published by the Free Software Foundation.                        |
* +----------------------------------------------------------------------+
*
*/
#include "global.h"
#include "model/am_time.h"
#ifdef HAVE_YAML_H
#include <yaml.h>
#include "yaml_utils.h"

extern unsigned char* str_tag;
extern unsigned char* map_tag;
extern yaml_emitter_t emitter;


gboolean
yaml_add_key_value_pair(const char* key, const char* value)
{
	yaml_event_t event;

	if(!value) value = "";

	if(!yaml_scalar_event_initialize(&event, NULL, str_tag, (guchar*)key, -1, 1, 0, YAML_PLAIN_SCALAR_STYLE)) return FALSE;
	if(!yaml_emitter_emit(&emitter, &event)) return FALSE;
	if(!yaml_scalar_event_initialize(&event, NULL, str_tag, (guchar*)value, -1, 0, 1, YAML_PLAIN_SCALAR_STYLE/*YAML_SINGLE_QUOTED_SCALAR_STYLE*/)) return FALSE;
	if(!yaml_emitter_emit(&emitter, &event)) return FALSE;
	return TRUE;
}


gboolean
yaml_add_key_value_pair_int(const char* key, int ival)
{
	yaml_event_t event;
	char value[256];
	snprintf(value, 255, "%i", ival);

	if(!yaml_scalar_event_initialize(&event, NULL, str_tag, (guchar*)key, -1, 1, 0, YAML_PLAIN_SCALAR_STYLE)) return FALSE;
	if(!yaml_emitter_emit(&emitter, &event)) return FALSE;
	if(!yaml_scalar_event_initialize(&event, NULL, str_tag, (guchar*)value, -1, 1, 0, YAML_ANY_SCALAR_STYLE)) return FALSE;
	if(!yaml_emitter_emit(&emitter, &event)) return FALSE;
	return TRUE;
}


gboolean
yaml_add_key_value_pair_float(const char* key, float fval)
{
	yaml_event_t event;
	char value[256];
	snprintf(value, 255, "%.2f", fval);

	if(!yaml_scalar_event_initialize(&event, NULL, str_tag, (guchar*)key, -1, 1, 0, YAML_PLAIN_SCALAR_STYLE)) return FALSE;
	if(!yaml_emitter_emit(&emitter, &event)) return FALSE;
	if(!yaml_scalar_event_initialize(&event, NULL, str_tag, (guchar*)value, -1, 0, 1, YAML_PLAIN_SCALAR_STYLE)) return FALSE;
	if(!yaml_emitter_emit(&emitter, &event)) return FALSE;
	return TRUE;
}


gboolean
yaml_add_key_value_pair_bool(const char* key, gboolean val)
{
	yaml_event_t event;
	char value[256];
	snprintf(value, 255, "%s", val ? "true" : "false");

	if(!yaml_scalar_event_initialize(&event, NULL, str_tag, (guchar*)key, -1, 1, 0, YAML_PLAIN_SCALAR_STYLE)) return FALSE;
	if(!yaml_emitter_emit(&emitter, &event)) return FALSE;
	if(!yaml_scalar_event_initialize(&event, NULL, (guchar*)YAML_BOOL_TAG, (guchar*)value, -1, 1, 0, YAML_ANY_SCALAR_STYLE)) return FALSE;
	if(!yaml_emitter_emit(&emitter, &event)) return FALSE;
	return TRUE;
}


gboolean
yaml_load(const char* filename, YamlHandler handlers[])
{
	yaml_parser_t parser; yaml_parser_initialize(&parser);

	FILE* file_input = fopen(filename, "rb");
	if(!file_input){
		log_print(LOG_FAIL, "unable to load config file: %s", filename);
		return FALSE;
	}
	yaml_parser_set_input_file(&parser, file_input);

	int section = 0;

	//read the event sequence.
	int safety = 0;
	int depth = 0;
	char key[64] = "";
	gboolean end = FALSE;
	yaml_event_t event;
	do {
		if (!yaml_parser_parse(&parser, &event)) goto error; // Get the next event.

		switch (event.type) {
			case YAML_STREAM_START_EVENT:
				dbg(2, "YAML_STREAM_START_EVENT");
				break;
			case YAML_STREAM_END_EVENT:
				end = TRUE;
				dbg(2, "YAML_STREAM_END_EVENT");
				break;
			case YAML_DOCUMENT_START_EVENT:
				dbg(2, "YAML_DOCUMENT_START_EVENT");
				break;
			case YAML_DOCUMENT_END_EVENT:
				end = TRUE;
				dbg(2, "YAML_DOCUMENT_END_EVENT");
				break;
			case YAML_ALIAS_EVENT:
				dbg(0, "YAML_ALIAS_EVENT");
				break;
			case YAML_SCALAR_EVENT:
				if(event.data.scalar.anchor) dbg(0, "YAML_SCALAR_EVENT: %s => %s", event.data.scalar.anchor, event.data.scalar.tag);
				//dbg(0, "YAML_SCALAR_EVENT: value=%s %i plain=%i style=%i", event.data.scalar.value, event.data.scalar.length, event.data.scalar.plain_implicit, event.data.scalar.style);

				if(!key[0]){ //1st half of a pair. We dont yet know what to do with it
					strncpy(key, (char*)event.data.scalar.value, 63);
				}else{
					//2nd half of a pair
					dbg(2, "      %s=%s", key, event.data.scalar.value);
					key[0] = '\0';
				}
				break;
			case YAML_SEQUENCE_START_EVENT:
				dbg(2, "YAML_SEQUENCE_START_EVENT");
				break;
			case YAML_SEQUENCE_END_EVENT:
				dbg(2, "YAML_SEQUENCE_END_EVENT");
				break;
			case YAML_MAPPING_START_EVENT:
				depth++;
				if(key[0]){
					if(!section){
						dbg(2, "found section! %s", key);
						int i = 0;
						YamlHandler* h;
						while((h = &handlers[i]) && h->key ){
							//dbg(0, "h=%p key=%s", h, h->key);
							if(!strcmp(h->key, key)){
								//dbg(0, ":-)");
								h->callback(&parser, key, h->data);
								break;
							}
							i++;
						}
					}
					else dbg(2, "new section: %s", key);
					key[0] = '\0';
				}
				else dbg(2, "YAML_MAPPING_START_EVENT");
				//dbg(0, "YAML_MAPPING_START_EVENT: anchor=%s tag=%s", event.data.mapping_start.anchor, event.data.mapping_start.tag);
				break;
			case YAML_MAPPING_END_EVENT:
				dbg(2, "YAML_MAPPING_END_EVENT");
				if(--depth < 0) gwarn("too many YAML_MAPPING_END_EVENT's.");
				break;
			case YAML_NO_EVENT:
				dbg(0, "YAML_NO_EVENT");
				break;
		}

		//the application is responsible for destroying the event object.
		yaml_event_delete(&event);

	} while(!end && safety++ < 1024);

	yaml_parser_delete(&parser);

	return TRUE;

  error:
	yaml_parser_delete(&parser);

	return FALSE;
}


gboolean
yaml_load_section(yaml_parser_t* parser, YamlHandler handlers[], YamlValueHandler vhandlers[])
{
	dbg(2, "%p %p", handlers, vhandlers);
	int depth = 0;
	char key[64] = "";
	int safety = 0;
	yaml_event_t event;
	do {
		if (!yaml_parser_parse(parser, &event)) goto error; // Get the next event.

		switch (event.type) {
			case YAML_STREAM_START_EVENT:
				dbg(2, "YAML_STREAM_START_EVENT");
				break;
			case YAML_STREAM_END_EVENT:
				gwarn("unexpected YAML_STREAM_END_EVENT");
				break;
			case YAML_DOCUMENT_START_EVENT:
				dbg(2, "YAML_DOCUMENT_START_EVENT");
				break;
			case YAML_DOCUMENT_END_EVENT:
				gwarn("enexpected YAML_DOCUMENT_END_EVENT");
				break;
			case YAML_ALIAS_EVENT:
				dbg(0, "YAML_ALIAS_EVENT");
				break;
			case YAML_SCALAR_EVENT:
				if(event.data.scalar.anchor) dbg(0, "YAML_SCALAR_EVENT: %s => %s", event.data.scalar.anchor, event.data.scalar.tag);
				//dbg(0, "YAML_SCALAR_EVENT: value=%s %i plain=%i style=%i", event.data.scalar.value, event.data.scalar.length, event.data.scalar.plain_implicit, event.data.scalar.style);

				if(!key[0]){ //1st half of a pair. We dont yet know what to do with it
					strncpy(key, (char*)event.data.scalar.value, 63);
					//dbg(0, "setting key: %s", key);
				}else{
					//2nd half of a pair
					dbg(2, "      %s=%s", key, event.data.scalar.value);
					int i = 0;
					YamlValueHandler* h;
					dbg(2, "        vh=%p %s", &vhandlers[0], &vhandlers[0] ? (&vhandlers[0])->key : "<none>");
					while((h = &vhandlers[i]) && h->key ){
						dbg(2, "             key=%s", h->key);
						if(!strcmp(h->key, key)){
							//dbg(3, "            key match");
							h->callback(&event, h->value);
							key[0] = '\0';
							break;
						}
						i++;
					}

					key[0] = '\0';
				}
				break;
			case YAML_SEQUENCE_START_EVENT:
				dbg(2, "YAML_SEQUENCE_START_EVENT");
				break;
			case YAML_SEQUENCE_END_EVENT:
				dbg(2, "YAML_SEQUENCE_END_EVENT");
				break;
			case YAML_MAPPING_START_EVENT:
				depth++;
				if(key[0]){
					dbg(2, "found sub-section: %s", key);
					int i = 0;
					YamlHandler* h;
					while((h = &handlers[i]) && h->key ){
						dbg(2, "             key=%s", h->key);
						if(!strlen(h->key) || !strcmp(h->key, key)){ //empty string is wildcard
							h->callback(parser, key, h->data);
							key[0] = '\0';
							depth--;
							break;
						}
						i++;
					}
					key[0] = '\0';
				}
				else dbg(2, "YAML_MAPPING_START_EVENT");
				//dbg(0, "YAML_MAPPING_START_EVENT: anchor=%s tag=%s", event.data.mapping_start.anchor, event.data.mapping_start.tag);
				break;
			case YAML_MAPPING_END_EVENT:
				dbg(2, "YAML_MAPPING_END_EVENT");
				if(--depth < -1) gwarn("too many YAML_MAPPING_END_EVENT's.");
				if(depth <= 0) goto done;
				break;
			case YAML_NO_EVENT:
				dbg(0, "YAML_NO_EVENT");
				break;
		}

		//the application is responsible for destroying the event object.
		yaml_event_delete(&event);

	} while(/*!end && */safety++ < 1024);
	dbg(0, "end - NEVER GET HERE?");

	return TRUE;

  done:
	yaml_event_delete(&event);
	return TRUE;

  error:
	//yaml_parser_delete(parser);
	return FALSE;
}


gboolean
yaml_load_section2(yaml_parser_t* parser, YamlHandler** handlers, YamlValueHandler** vhandlers)
{
	//this is exactly the same as above but args are pointers rather than arrays.

	dbg(2, "%p %p", handlers, vhandlers);
	int depth = 0;
	char key[64] = "";
	int safety = 0;
	yaml_event_t event;
	do {
		if (!yaml_parser_parse(parser, &event)) goto error; // Get the next event.

		switch (event.type) {
			case YAML_STREAM_START_EVENT:
				dbg(2, "YAML_STREAM_START_EVENT");
				break;
			case YAML_STREAM_END_EVENT:
				gwarn("unexpected YAML_STREAM_END_EVENT");
				break;
			case YAML_DOCUMENT_START_EVENT:
				dbg(2, "YAML_DOCUMENT_START_EVENT");
				break;
			case YAML_DOCUMENT_END_EVENT:
				gwarn("enexpected YAML_DOCUMENT_END_EVENT");
				break;
			case YAML_ALIAS_EVENT:
				dbg(0, "YAML_ALIAS_EVENT");
				break;
			case YAML_SCALAR_EVENT:
				if(event.data.scalar.anchor) dbg(0, "YAML_SCALAR_EVENT: %s => %s", event.data.scalar.anchor, event.data.scalar.tag);
				//dbg(0, "YAML_SCALAR_EVENT: value=%s %i plain=%i style=%i", event.data.scalar.value, event.data.scalar.length, event.data.scalar.plain_implicit, event.data.scalar.style);

				if(!key[0]){ //1st half of a pair. We dont yet know what to do with it
					strncpy(key, (char*)event.data.scalar.value, 63);
					//dbg(0, "setting key: %s", key);
				}else{
					//2nd half of a pair
					dbg(2, "      %s=%s", key, event.data.scalar.value);
					int i = 0;
					YamlValueHandler* h;
					//dbg(0, "        vh=%p %s", vhandlers[0], vhandlers[0] ? (vhandlers[0])->key : "<none>");
					while((h = vhandlers[i]) && h->key){
						dbg(2, "             key=%s", h->key);
						if(!strcmp(h->key, key)){
							//dbg(3, "            key match");
							h->callback(&event, h->value);
							key[0] = '\0';
							break;
						}
						i++;
					}

					key[0] = '\0';
				}
				break;
			case YAML_SEQUENCE_START_EVENT:
				dbg(2, "YAML_SEQUENCE_START_EVENT");
				break;
			case YAML_SEQUENCE_END_EVENT:
				dbg(2, "YAML_SEQUENCE_END_EVENT");
				break;
			case YAML_MAPPING_START_EVENT:
				depth++;
				if(key[0]){
					dbg(2, "found sub-section: %s", key);
					int i = 0;
					YamlHandler* h;
					while((h = handlers[i]) && h->key ){
						dbg(2, "             key=%s", h->key);
						if(!strlen(h->key) || !strcmp(h->key, key)){ //empty string is wildcard
							h->callback(parser, key, h->data);
							key[0] = '\0';
							depth--;
							break;
						}
						i++;
					}
					key[0] = '\0';
				}
				else dbg(2, "YAML_MAPPING_START_EVENT");
				//dbg(0, "YAML_MAPPING_START_EVENT: anchor=%s tag=%s", event.data.mapping_start.anchor, event.data.mapping_start.tag);
				break;
			case YAML_MAPPING_END_EVENT:
				dbg(2, "YAML_MAPPING_END_EVENT");
				if(--depth < -1) gwarn("too many YAML_MAPPING_END_EVENT's.");
				if(depth <= 0) goto done;
				break;
			case YAML_NO_EVENT:
				dbg(0, "YAML_NO_EVENT");
				break;
		}

		//the application is responsible for destroying the event object.
		yaml_event_delete(&event);

	} while(/*!end && */safety++ < 1024);
	dbg(0, "end - NEVER GET HERE?");

	return TRUE;

  done:
	yaml_event_delete(&event);
	return TRUE;

  error:
	//yaml_parser_delete(parser);
	return FALSE;
}


void yaml_set_string(yaml_event_t* event, gpointer data)
{
	g_return_if_fail(event->type == YAML_SCALAR_EVENT);
	*((char**)data) = g_strdup((char*)event->data.scalar.value);
}

void
yaml_set_int(yaml_event_t* event, gpointer data)
{
	g_return_if_fail(event->type == YAML_SCALAR_EVENT);
	*((int*)data) = atoi((char*)event->data.scalar.value);
}

void
yaml_set_uint64(yaml_event_t* event, gpointer data)
{
	g_return_if_fail(event->type == YAML_SCALAR_EVENT);
	*((uint64_t*)data) = strtoull((char*)(event->data.scalar.value), NULL, 10);
}


void
yaml_set_position(yaml_event_t* event, gpointer data)
{
	//sets a GPos from a value in samples.

	g_return_if_fail(event->type == YAML_SCALAR_EVENT);

	int samples = atoi((char*)(event->data.scalar.value));

	samples2pos(samples, data);
}


#endif
