/**
* +----------------------------------------------------------------------+
* | This file is part of the Ayyi project. http://www.ayyi.org           |
* | copyright (C) 2004-2017 Tim Orford <tim@orford.org>                  |
* +----------------------------------------------------------------------+
* | This program is free software; you can redistribute it and/or modify |
* | it under the terms of the GNU General Public License version 3       |
* | as published by the Free Software Foundation.                        |
* +----------------------------------------------------------------------+
*
*/

/*
 *  Unit Tests for AyyiGtk model libary
 *  -----------------------------------
 *
 *  For unit tests that dont require connecting to any Ayyi services.
 *  All tests should run synchronously without callbacks.
 *
 */
#define __main_c__
#include "config.h"
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <getopt.h>
#include <math.h>
#include <time.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <inttypes.h>
#include <glib.h>
#include <ayyi/ayyi_typedefs.h>
#include "model/ayyi_model.h"

static void app_init  ();
static bool next_test ();

typedef void (TestFn)();

TestFn test_ayyi_list, test_time_conversions,
	test_mu_samples_conversion,
	test_mu2pos,
	test_mu_gpos_conversion,
	test_pos_mu_conversion,
	test_pos_sample_conversion,
	test_gpos_to_samples,
	test_samples_to_gpos,
	test_pos_to_gpos,
	test_gpos_divide,
	test_q_to_prev_next,
	test_pos_add,
	test_pos_subtract,
	test_bbss,
	test_bbss2pos,
	test_fader_conversion;

struct _ayyi_client ayyi;

#define STOP FALSE;
#define OK 0
#define FAIL 1

#define assert(A, B, ...) \
	{bool __ok_ = (boolp)A; \
	{if(!__ok_) gerr(B, ##__VA_ARGS__); } \
	{if(!__ok_) { passed = false; return; }}}

gboolean passed = false;
int      current_test = -1;
gpointer tests[] = {
	test_time_conversions,
	test_ayyi_list,
	test_mu_samples_conversion,
	test_mu2pos,
	test_mu_gpos_conversion,
	test_pos_mu_conversion,
	test_pos_sample_conversion,
	test_gpos_to_samples,
	test_samples_to_gpos,
	test_pos_to_gpos,
	test_gpos_divide,
	test_pos_add,
	test_pos_subtract,
	test_bbss,
	test_bbss2pos,
	test_q_to_prev_next,
	test_fader_conversion,
};

#define MAX_RECENT_SONGS 4
typedef struct _canvas_op       canvas_op;

typedef struct
{
  int            state;

  gboolean       dbus;

  char           config_filename[256];
  GKeyFile*      key_file;             //config file data.

  char           recent_songs[MAX_RECENT_SONGS][256];

} App;

App app;


int
main (int argc, char *argv[])
{
	g_log_set_handler (NULL, G_LOG_LEVEL_WARNING | G_LOG_LEVEL_CRITICAL | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION, log_handler, NULL);
	g_log_set_handler ("Gtk", G_LOG_LEVEL_CRITICAL | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION, log_handler, NULL);
	g_log_set_handler ("Gdk", G_LOG_LEVEL_CRITICAL | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION, log_handler, NULL);
	g_log_set_handler ("GLib", G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION, log_handler, NULL);
	g_log_set_handler ("GLib-GObject", G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION, log_handler, NULL);
	g_log_set_handler ("Gdl", G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION, log_handler, NULL);
	g_log_set_handler ("Ayyi", G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION, log_handler, NULL);

	am__init(NULL);
	app_init();

	while(next_test());

	return EXIT_SUCCESS;
}


static void
app_init()
{
	ayyi.got_shm       = 0;     //before this is set, we must have _all_ required shm pointers.

	int i; for(i=0;i<MAX_RECENT_SONGS;i++) *app.recent_songs[i] = 0;
	app.dbus           = FALSE;

#if 0
	((AyyiSongService*)ayyi.service)->song->bpm = 120;
	((AyyiSongService*)ayyi.service)->song->sample_rate = 44100;
#endif
	if(!((AyyiSongService*)ayyi.service)->song){
		((AyyiSongService*)ayyi.service)->song = g_new0(struct _shm_song, 1);
		((AyyiSongService*)ayyi.service)->song->bpm = 120;
		((AyyiSongService*)ayyi.service)->song->sample_rate = 44100;
	}

	ayyi.log.to_stdout = true;
}


static gboolean
next_test()
{
	PF;
	current_test++;
	if(current_test < G_N_ELEMENTS(tests)){
		gboolean (*t)() = tests[current_test];
		passed = false;
		t();
		log_print(!passed ? LOG_FAIL : LOG_OK, "done");
		return true;
	}
	else printf("finished all.\n");
	return STOP;
}


gboolean
test1()
{
	PF;
	/*
	uint64_t mu = px2mu_nz(PX_PER_BEAT);
	if(mu != CORE_SUBS_PER_BEAT * CORE_MU_PER_SUB){ gerr("%Lu", mu); return FAIL; }
	*/
	return OK;
}

#define START_TEST \
	passed = true; \
	static int step = 0; /*TODO unit tests shouldnt have steps?*/ \
	if(!step) log_print(0, "running %s ...", __func__);

#define PASS_TEST \
	( \
	passed = true \
	)

#define FAIL_TEST \
	{ passed = false; return; }

void
test_time_conversions()
{
	START_TEST;

	PASS_TEST;
}


void
test_mu2pos()
{
	g_assert(sizeof(AyyiSongPos) == 8);

	typedef struct {
		AyyiMu      mu;
		AyyiSongPos result;
	} Case;

	Case cases[] = {
		{0,                                        {0,}},
		{1,                                        {0,0,1}},
		{AYYI_MU_PER_SUB - 1,                      {0,0,11024}},
		{AYYI_MU_PER_SUB,                          {0,1,0}},
		{AYYI_MU_PER_SUB + 1,                      {0,1,1}},
		{AYYI_MU_PER_SUB * AYYI_SUBS_PER_BEAT - 1, {0,3839,11024}},
		{AYYI_MU_PER_SUB * AYYI_SUBS_PER_BEAT,     {1,0,0}},
		{AYYI_MU_PER_SUB * AYYI_SUBS_PER_BEAT + 1, {1,0,1}},
		{-1,                                       {0,0,-1}},
	};

	START_TEST;
	bool pass = true;
	int i; for(i=0;i<G_N_ELEMENTS(cases);i++){
		Case* c = &cases[i];
		printf("  %"PRIi64"\n", c->mu);
		AyyiSongPos r; ayyi_mu2pos(c->mu, &r);
		g_assert(ayyi_pos_is_valid(&r));

		char bbst2[32]; ayyi_pos2bbst(&r, bbst2);
		char bbst0[32]; ayyi_pos2bbst(&c->result, bbst0);
		if(ayyi_pos_cmp(&r, &c->result)){
			gerr("%s != %s (%i)", bbst2, bbst0, c->result.beat);
			pass = false;
		}
	}
	if(pass) PASS_TEST; else FAIL_TEST;
}


void
test_mu_samples_conversion()
{
	START_TEST;

	uint32_t s[] = {0, 22050, 22051, 22052, 22053, 44100, 11254665};

	int i; for(i=0;i<G_N_ELEMENTS(s);i++){
		long long mu = ayyi_samples2mu(s[i]);
		long long b = ayyi_mu2samples(mu);
		long long mpb = AYYI_MU_PER_BEAT;
		dbg(0, "%8i mu=%8Li s=%8Li (%Li)", s[i], mu, b, mpb);

		assert((b == s[i]), "conversion");
	}
}


void
test_ayyi_list()
{
	//AyyiList's live only in shared memory, so can't be tested here.
	START_TEST;
	return;

	AyyiList* ayyi_list_prepend(AyyiList* l, int i)
	{
		AyyiList* new = g_new0(AyyiList, 1);
		new->id = i;
		return new;
	}
	int id = 99;
	AyyiList* l = ayyi_list_prepend(NULL, id);
	if(!l->id == id) FAIL_TEST
	if(!(ayyi_list__length(l) == 1)) FAIL_TEST;
}


void
test_q_to_prev_next()
{
	START_TEST;
	QMap q;
	strncpy(q.name, "exponential", 63);
	q.size = 5;
	q.pos[0] = (GPos){0,0,0};
	q.pos[1] = (GPos){0,0,AYYI_TICKS_PER_SUBBEAT/4};
	q.pos[2] = (GPos){0,1,0};
	q.pos[3] = (GPos){1,0,0};
	q.pos[4] = (GPos){4,0,0};
	q.pos[5] = (GPos){8,0,0};

	GPos cases[] = {{-4, 0, 0}, {0, 0, 0}, {0, 0, 1}, {0, 0, 96}, {0, 0, 97}, {1, 1, 0}, {7, 0, 0}, {9, 0, 0}, {16, 0, 0}};

	int i; for(i=0;i<G_N_ELEMENTS(cases);i++){
		GPos prev = cases[i];
		GPos next = cases[i];
		GPos orig = cases[i];
		q_to_prev(&prev, &q);
		q_to_next(&next, &q);

		char bbst1[64]; pos2bbst(&orig, bbst1);
		char bbst2[64]; pos2bbst(&prev, bbst2);
		char bbst3[64]; pos2bbst(&next, bbst3);
		dbg(1, "orig=%s prev=%s next=%s", bbst1, bbst2, bbst3);
	}
	PASS_TEST;
}


void
test_mu_gpos_conversion()
{
	START_TEST;

	#define CORRECTION_PER_BEAT 768

	uint64_t mu_beat = ((uint64_t)AYYI_TICKS_PER_BEAT) * ((uint64_t)AYYI_MU_PER_TICK) + CORRECTION_PER_BEAT;
	if(mu_beat != AYYI_MU_PER_BEAT){
		dbg(0, "            !! %Lu != %Lu", mu_beat, ((uint64_t)AYYI_MU_PER_BEAT));
		FAIL_TEST;
	}

	typedef struct {
		GPos     d;
		uint64_t mu;
	} F;
	F d[] = {{{0,0,0}}, {{0,0,1}}, {{0,1,0}}, {{0,1,1}}, {{0,1,2}}, {{0,1,3}}, {{0,3,383}}, {{1,0,0},AYYI_MU_PER_BEAT}, {{3,3,383}}, {{4,0,1}}, {{64,0,0},64*AYYI_MU_PER_BEAT}, {{1024,0,1}}};
	int i; for(i=0;i<G_N_ELEMENTS(d);i++){
		F* f = &d[i];
		uint64_t mu = pos2mu(&f->d);
		mu += f->d.beat * CORRECTION_PER_BEAT;
		if(f->mu && mu != f->mu) dbg(0, "            !! %Lu != %Lu", mu, f->mu);
		GPos out; mu2pos(mu, &out);

		char bbst1[32]; pos2bbst(&f->d, bbst1);
		char bbst2[32]; pos2bbst(&out, bbst2);
		gpos_is_valid(&out);
		if(pos_cmp(&f->d, &out)){
			log_print(LOG_FAIL, "test %i", i);
			passed = false;
		}
		dbg(1, "%s --> %s mu=%Li", bbst1, bbst2, mu);
	}
	if (passed) PASS_TEST; else FAIL_TEST;
}


void
test_pos_mu_conversion()
{
	START_TEST;
	AyyiSongPos d[] = {{0,0,0}, {0,0,1}, {0,1,0}, {1,0,0}};
	uint64_t result[] = {0, 1, AYYI_MU_PER_SUB, AYYI_MU_PER_BEAT}; 
	int i; for(i=0;i<G_N_ELEMENTS(d);i++){
		AyyiSongPos* pos = &d[i];
		if(ayyi_pos2mu(pos) != result[i]) FAIL_TEST;
	}
	PASS_TEST;
}


void
test_pos_sample_conversion()
{
	START_TEST;
	static int64_t accuracy;

	bool _ayyi_pos_cmp(const AyyiSongPos* a, const AyyiSongPos* b)
	{
		//return TRUE if the two positions are not within sample accuracy.

		int64_t mu1 = ayyi_pos2mu((AyyiSongPos*)a);
		int64_t mu2 = ayyi_pos2mu((AyyiSongPos*)b);
		int64_t diff = ABS(mu1 - mu2);
		//dbg(0, "error=%Li", diff);
		return diff > accuracy;
	}

	float tempo[] = {60.0, 120.0, 240.0};
	int t; for(t=0;t<G_N_ELEMENTS(tempo);t++){
		((AyyiSongService*)ayyi.service)->song->bpm = tempo[t];
		accuracy = ayyi_samples2mu(1);
		dbg(1, "tempo=%.1f accuracy=%Lu", tempo[t], accuracy);

		AyyiSongPos d[] = {{0,0,0}, {0,0,1}, {0,0,accuracy}, {0, 0, AYYI_MU_PER_SUB/2}, {0,0,AYYI_MU_PER_SUB-1}, {0,1,0}, {0,1,1}, {0,1,2}, {0,1,3}, {0,AYYI_SUBS_PER_BEAT/2,0}, {0,AYYI_SUBS_PER_BEAT-1,AYYI_MU_PER_SUB-1}, {1,0,0}, {3,3,383}, {4,0,1}};
		int i; for(i=0;i<G_N_ELEMENTS(d);i++){
			uint32_t samples = ayyi_pos2samples(&d[i]);
			AyyiSongPos out; ayyi_samples2pos(samples, &out);
			ayyi_pos_is_valid(&out);
			char bbst1[32]; ayyi_pos2bbst(&d[i], bbst1);
			char bbst2[32]; ayyi_pos2bbst(&out, bbst2);
			dbg(1, "%s --> %s samples=%u", bbst1, bbst2, samples);
			if(_ayyi_pos_cmp(&d[i], &out)) log_print(LOG_FAIL, "test %i", i);
			if(_ayyi_pos_cmp(&d[i], &out)) gwarn("test %i", i);

			//check that the displayed time is the same:
			GPos g_in, g_out;
			if(!songpos_ayyi2gui(&g_in, &d[i])) return;
			if(!songpos_ayyi2gui(&g_out, &out)) return;
			pos2bbst0(&g_in, bbst1);
			pos2bbst0(&g_out, bbst2);
			if(strcmp(bbst1, bbst2)){
				gwarn("%s --> %s", bbst1, bbst2);
			}
		}
	}
	PASS_TEST;
}


void
test_samples_to_gpos()
{
	//GPos is lower resolution that samples, so translating samples -> gpos -> samples is not possible with accuracy.

	START_TEST;
	gboolean pass = true;
	static int64_t accuracy;
	float tempo[] = {60.0, 120.0, 240.0};
	uint32_t cases[] = {0, 1, 10, 22050, 44099, 44100, 44101};

	int t; for(t=0;t<G_N_ELEMENTS(tempo);t++){
		((AyyiSongService*)ayyi.service)->song->bpm = tempo[t];
		accuracy = ayyi_samples2mu(1);

		int i; for(i=0;i<G_N_ELEMENTS(cases);i++){
			uint32_t in = cases[i];
			GPos pos;
			samples2pos(in, &pos);
			uint32_t out = pos2samples(&pos);

			char bbst[64]; pos2bbst(&pos, bbst);

			int diff = ABS(out - in);
			if(diff > accuracy / 2){
				gerr("in=%u out=%u %s err=%i", in, out, bbst, diff);
				pass = false;
			}
			else log_print(LOG_OK, "%5u %s", in, bbst);
		}
	}
	if(pass) PASS_TEST; else FAIL_TEST;
}


void
test_pos_to_gpos()
{
	//TODO this gives useful output but does not check for errors.
	START_TEST;
	gboolean pass = true;
	static int64_t accuracy;
	accuracy = ayyi_samples2mu(1);
	AyyiSongPos cases[] = {{0,0,0}, {0,0,1}, {0,0,accuracy}, {0, 0, AYYI_MU_PER_SUB/2}, {0,0,AYYI_MU_PER_SUB-1}, {0,1,0}, {0,1,1}, {0,1,2}, {0,1,3}, {0,AYYI_SUBS_PER_BEAT/2,0}, {0,AYYI_SUBS_PER_BEAT-1,AYYI_MU_PER_SUB-1}, {1,0,0}, {3,3,383}, {4,0,1}, {127,0,0}, {128,0,0}, {511,0,0}};
	int i; for(i=0;i<G_N_ELEMENTS(cases);i++){
		AyyiSongPos* in = &cases[i];
		char bbst1[64]; ayyi_pos2bbst(in, bbst1);
		GPos pos;
		songpos_ayyi2gui(&pos, in);
		char bbst2[64]; pos2bbst0(&pos, bbst2);
		dbg(0, "                                  %s %s", bbst1, bbst2);
		//if(in->beat != pos.beat) gerr("!!! beat: %i!=%i", in->beat, pos.beat);
	}
	if(pass) PASS_TEST; else FAIL_TEST;
}


void
test_gpos_to_samples()
{
	START_TEST;
	if(!((AyyiSongService*)ayyi.service)->song){ passed = false; return; }
	gboolean pass = true;
	GPos cases[] = {/*{-4, 0, 0}, */{0, 0, 0}, {0, 0, 1}, {0, 0, 96}, {0, 0, 97}, {0, 3, 383}, {1, 1, 0}, {7, 0, 0}, {9, 0, 0}, {16, 0, 0}};
	int i; for(i=0;i<G_N_ELEMENTS(cases);i++){
		GPos in = cases[i];
		uint32_t samples = pos2samples(&in);
		GPos out; samples2pos(samples, &out);

		char bbst1[64]; pos2bbst(&in, bbst1);
		char bbst2[64]; pos2bbst(&out, bbst2);
		if(!pos_cmp(&in, &out)){
			log_print(0, "%s %5u", bbst1, samples);
		} else {
			gerr("%s != %s %5u", bbst1, bbst2, samples);
			pass = false;
		}
	}
	passed = pass;
}


void
test_gpos_divide()
{
	START_TEST;
	gboolean pass = true;
	int n = 2;
	GPos cases[]   = {{0, 0, 2}, {0, 2, 0}, {2, 0, 0}, {0, 1, 0},                          {1, 0,               0}};
	GPos results[] = {{0, 0, 1}, {0, 1, 0}, {1, 0, 0}, {0, 0, (AYYI_TICKS_PER_SUBBEAT/n)}, {0, SUBS_PER_BEAT/n, 0}};
	int i; for(i=0;i<G_N_ELEMENTS(cases);i++){
		GPos pos = cases[i];
		pos_divide(&pos, n);

		char bbst0[64]; pos2bbst(&cases[i], bbst0);
		char bbst1[64]; pos2bbst(&pos, bbst1);
		char bbst2[64]; pos2bbst(&results[i], bbst2);
		if(!pos_cmp(&pos, &results[i])){
			dbg(1, "%s --> %s", bbst0, bbst1);
		} else {
			gerr("%s != %s", bbst1, bbst2);
			pass = false;
		}
	}
	if (pass) PASS_TEST; else FAIL_TEST;
}


void
test_pos_add()
{
	START_TEST;
	bool pass = true;
	AyyiSongPos cases[][2] = {{{0,1,},{0,1,}}, {{1,},{1,}}, {{1,},{2,}}, {{0,AYYI_SUBS_PER_BEAT-2},{0,1}}, {{0,AYYI_SUBS_PER_BEAT-1},{0,1}}, {{0,AYYI_SUBS_PER_BEAT-1},{0,2}}, {{0,0,AYYI_MU_PER_SUB-1},{0,0,1}}};
	AyyiSongPos results[]  = {{0,2},           {2,},        {3,},        {0,AYYI_SUBS_PER_BEAT-1},         {1,},                             {1,1},                            {0,1}};
	int i; for(i=0;i<G_N_ELEMENTS(cases);i++){
		AyyiSongPos a = cases[i][0];
		ayyi_pos_add(&a, &cases[i][1]);
		char bbst0[32]; ayyi_pos2bbst(&cases[i][0], bbst0);
		char bbst1[32]; ayyi_pos2bbst(&cases[i][1], bbst1);
		char bbst2[32]; ayyi_pos2bbst(&a, bbst2);
		char bbst3[32]; ayyi_pos2bbst(&results[i], bbst3);
		dbg(0, "   %s + %s = %s", bbst0, bbst1, bbst2);
		if(ayyi_pos_cmp(&a, &results[i])){
			gerr("%s != %s", bbst2, bbst3);
			pass = false;
		}
	}
	if (pass) PASS_TEST; else FAIL_TEST;
}


void
test_pos_subtract()
{
	typedef struct {
		AyyiSongPos pos1;
		AyyiSongPos pos2;
		AyyiSongPos result;
	} Case;

	Case cases[] = {
		{{0,},    {0,},                {0,}},
		{{1,},    {0,},                {1,}},
		{{1,},    {1,},                {0,}},
		{{2,},    {1,},                {1,}},
		{{1,1},   {1,},                {0,1}},
		{{2,1},   {1,},                {1,1}},
		{{1,1,1}, {0,},                {1,1,1}},
		{{1,1,1}, {1,},                {0,1,1}},
		{{2,},    {0,1},               {1,AYYI_SUBS_PER_BEAT-1}},
		{{2,1},   {0,1},               {2,}},
		{{0},     {1},                 {-1,}},
		{{1},     {2},                 {-1,}},
		{{0},     {0,1},               {0,-1}},
	};

	START_TEST;
	bool pass = true;
	int i; for(i=0;i<G_N_ELEMENTS(cases);i++){
		Case* c = &cases[i];
		AyyiSongPos a = c->pos1;
		AyyiSongPos b = c->pos2;
		g_assert(ayyi_pos_is_valid(&a));
		g_assert(ayyi_pos_is_valid(&b));
		char bbst1[32]; ayyi_pos2bbst(&a, bbst1);
		char bbst2[32]; ayyi_pos2bbst(&b, bbst2);
		dbg(0, "%i.%i", a.beat, a.sub);
		dbg(0, "%s - %s", bbst1, bbst2);
		ayyi_pos_sub(&a, &b);
		g_assert(ayyi_pos_is_valid(&a));

		char bbst3[32]; ayyi_pos2bbst(&a, bbst3);
		char bbst0[32]; ayyi_pos2bbst(&c->result, bbst0);
		if(ayyi_pos_cmp(&a, &c->result)){
			gerr("%s != %s", bbst2, bbst0);
			pass = false;
		}
	}
	if (pass) PASS_TEST; else FAIL_TEST;
}


void
test_bbss()
{
	typedef struct {
		AyyiSongPos pos;
		char*       result;
	} Case;

	Case cases[] = {
		{{0,},                          "001:1:1:001"},
		{{0,0,1},                       "001:1:1:001"},
		{{0,1,0},                       "001:1:1:001"},
		{{0,4,0},                       "001:1:1:001"},
		{{0,4,AYYI_MU_PER_SUB-1},       "001:1:1:001"},
		{{0,5,0},                       "001:1:1:002"},
		{{0,6,0},                       "001:1:1:002"},
		{{0,10,0},                      "001:1:1:003"},
		{{0,1*AYYI_SUBS_PER_BEAT/4,},   "001:1:2:001"},
		{{0,2*AYYI_SUBS_PER_BEAT/4,},   "001:1:3:001"},
		{{0,4*AYYI_SUBS_PER_BEAT/4-1,}, "001:1:4:192"},
		{{1,},                          "001:2:1:001"},
		{{-1,0,0},                      "000:4:1:001"},
		{{-2,0,0},                      "000:3:1:001"},
		{{-3,0,0},                      "000:2:1:001"},
		{{-4,0,0},                      "000:1:1:001"},
		{{-5,0,0},                      "-01:4:1:001"},
		{{-6,0,0},                      "-01:3:1:001"},
		{{-7,0,0},                      "-01:2:1:001"},
		{{-8,0,0},                      "-01:1:1:001"},
		{{-9,0,0},                      "-02:4:1:001"},
		{{-12,0,0},                     "-02:1:1:001"},
		{{-16,0,0},                     "-03:1:1:001"},
		{{-32,0,0},                     "-07:1:1:001"},
		{{-128,0,0},                    "-31:1:1:001"},
		{{-1024,0,0},                   "-255:1:1:001"},
		{{-8192,0,0},                   "-2047:1:1:001"},
		{{-32768,0,0},                  "-8191:1:1:001"},
		{{-131072,0,0},                 "-32767:1:1:001"},
		{{-1048576,0,0},                "-262143:1:1:001"},
		{{-1,1,0},                      "000:4:1:001"},
		{{-1,6,0},                      "000:4:1:002"},
	};

	START_TEST;
	gboolean pass = true;
	int i; for(i=0;i<G_N_ELEMENTS(cases);i++){
		Case* c = &cases[i];
		AyyiSongPos a = c->pos;
		g_assert(ayyi_pos_is_valid(&a));
		char bbss[16]; ayyi_pos2bbss(&a, bbss);
		dbg(0, "%8i,%4i %s %s", a.beat, a.sub, bbss, c->result);
		if(strcmp(bbss, c->result)){
			gerr("%s != %s", bbss, c->result);
			pass = false;
		}
	}
	if (pass) PASS_TEST; else FAIL_TEST;
}


void
test_bbss2pos()
{
	typedef struct {
		char*       bbss;
		AyyiSongPos result;
	} Case;

	Case cases[] = {
		{"001:1:1:001", {0,}},
		{"001:2:1:001", {1,}},
		{"002:1:1:001", {4,}},
		{"001:1:1:002", {0,5}},
		{"001:1:1:003", {0,10}},
		{"001:1:2:001", {0,1*AYYI_SUBS_PER_BEAT/4}},
		{"001:1:3:001", {0,2*AYYI_SUBS_PER_BEAT/4}},
		{"001:1:4:001", {0,3*AYYI_SUBS_PER_BEAT/4}},
		{"001:1:4:192", {0,AYYI_SUBS_PER_BEAT - 5}},
		{"001:1:4:191", {0,AYYI_SUBS_PER_BEAT - 10}},
	};

	START_TEST;
	gboolean pass = true;
	int i; for(i=0;i<G_N_ELEMENTS(cases);i++){
		Case* c = &cases[i];

		AyyiSongPos out;
		bbss2pos(c->bbss, &out);
		g_return_if_fail(ayyi_pos_is_valid(&out));

		char bbss [32]; ayyi_pos2bbss(&out, bbss);
		char bbss0[32]; ayyi_pos2bbss(&c->result, bbss0);
		char bbst1[32]; ayyi_pos2bbst(&c->result, bbst1);
		char bbst2[32]; ayyi_pos2bbst(&out, bbst2);
		if(ayyi_pos_cmp(&out, &c->result)){
			gerr("%i: %s != %s orig=%s (%s != %s)", i, bbss, c->bbss, bbss0, bbst2, bbst1);
			pass = false;
		}
	}
	if (pass) PASS_TEST; else FAIL_TEST;
}


void
test_fader_conversion()
{
	START_TEST;
	double in[] = {1.0, 0.999, 0.5, 0.01};
	int i; for(i=0;i<G_N_ELEMENTS(in);i++){
		float db = am_gain2db(in[i]);
		float out = am_db2gain(db);
		dbg(0, "in=%.3f out=%.3f dB=%.3f", in[i], out, db);
		if(fabs(in[i] - out) > 0.1) gwarn("fail!");
	}
	PASS_TEST;
}


