/**
* +----------------------------------------------------------------------+
* | This file is part of the Ayyi project. http://www.ayyi.org           |
* | copyright (C) 2004-2017 Tim Orford <tim@orford.org>                  |
* +----------------------------------------------------------------------+
* | This program is free software; you can redistribute it and/or modify |
* | it under the terms of the GNU General Public License version 3       |
* | as published by the Free Software Foundation.                        |
* +----------------------------------------------------------------------+
*
*/
#include "config.h"
#undef USE_VST

#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include <stdint.h>
#include <glib.h>
#include <glib-object.h>

#include <ayyi/ayyi.h>
#include <ayyi/ayyi_server.h>
#include "ayyi_example_plugin_so.h"
#include "dbus.h"
#include "dbus-service.h"

#define P_OUT printf("%s(): %s--->%s\n", __func__, yellow, white)

extern AyyiServer* ayyi_server;

enum
{
  CHANGED_SIGNAL,
  LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = { 0 };

static DBusGConnection *connection = NULL;
static DBusGProxy      *proxy      = NULL;

G_DEFINE_TYPE (ExampleDbus, example_dbus, G_TYPE_OBJECT);


static void
example_dbus_class_init (ExampleDbusClass *klass)
{
	signals[CHANGED_SIGNAL] = g_signal_new ("changed_signal",
                  G_OBJECT_CLASS_TYPE (klass),
                  (GSignalFlags)(G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED),
                  0,
                  NULL, NULL,
                  g_cclosure_marshal_VOID__VOID,
                  G_TYPE_NONE, 0);        //check
}


static void
example_dbus_init (ExampleDbus *application)
{
}


gboolean
example_dbus_register_service (ExampleDbus *application)
{
	//for successful registration, check security settings in /etc/dbus/.
	//-server must already be running.

	GError *err = NULL;

	if (connection) {
		g_warning ("Service already registered.");
		return FALSE;
	}

	#define BUS_COUNT 2
	struct _busses {
		DBusBusType type;
		char        name[64];
	} busses[BUS_COUNT] = {
		{DBUS_BUS_SESSION, "session"},
		{DBUS_BUS_SYSTEM, "system"}
	};

	int i; for(i=0;i<BUS_COUNT;i++){
		if (!(connection = dbus_g_bus_get (busses[i].type, &err))){
			printf("dbus bus connect (%i: %s) failed", i, busses[i].name);
			if(err->code != 0) printf(": %i: %s", err->code, err->message);
			printf("\n");
			g_error_free (err);
			err = NULL;
		}
		else break;
	}
	if(!connection){ errprintf("dbus service registration %s\n", fail); return FALSE; }
	application->connection = connection;
	dbg(2, "bus=%s %s %s %s", busses[i].name, DBUS_SERVICE_DBUS, DBUS_PATH_DBUS, DBUS_INTERFACE_DBUS);

	proxy = dbus_g_proxy_new_for_name (connection, DBUS_SERVICE_DBUS, DBUS_PATH_DBUS, DBUS_INTERFACE_DBUS);
	if(!proxy) gerr("!!");

	guint request_name_result;

	if (!org_freedesktop_DBus_request_name (proxy, APPLICATION_SERVICE_NAME, DBUS_NAME_FLAG_DO_NOT_QUEUE, &request_name_result, &err)) {
		errprintf("dbus service registration failed. %s\n", err->message);
		g_clear_error (&err);
		return FALSE;
	}
	dbg(2, "name registered: %s. result=%i", APPLICATION_SERVICE_NAME, request_name_result);

#ifdef NOT_TESTING
	g_object_unref (proxy);
#endif

	if (request_name_result == DBUS_REQUEST_NAME_REPLY_EXISTS) return FALSE;

	/*
		Install introspection information about the given object GType sufficient to allow methods on 
		the object to be invoked by name.

		The introspection information is normally generated by dbus-glib-tool, then this function is 
		called in the class_init() for the object class.

		Once introspection information has been installed, instances of the object registered with 
		dbus_g_connection_register_g_object() can have their methods invoked remotely.
	*/
	dbus_g_object_type_install_info (EXAMPLE_DBUS_TYPE_APPLICATION, &dbus_glib_example_dbus_object_info);

	dbus_g_connection_register_g_object (connection, DBUS_APP_PATH, G_OBJECT (application));

	return TRUE;
}


ExampleDbus*
example_dbus_get_instance (void)
{
	static ExampleDbus *instance;

	if (!instance) {
		instance = EXAMPLE_DBUS_APPLICATION (g_object_new (EXAMPLE_DBUS_TYPE_APPLICATION, NULL));
	}

	return instance;
}


void
example_dbus_register_signals()
{
	// we are not interested in any signals.
}


gboolean
example_dbus_get_shm(ExampleDbus *s, const char *_name, guint *seg_id, GError **error)
{
	P_IN;
	GList* segs = ayyi_server->shm_segs;
	if (segs) {
		AyyiShmSeg* seg = segs->data;
		*seg_id = seg->_shmfd;
		dbg(0, "seg_id=%i", *seg_id);
		return TRUE;
	}
	return FALSE;
}


gboolean
example_dbus_emit_changed(ExampleDbus *obj, GError **error)
{
	//dbg("type=%i %s--->%s", object_type, yellow, white);
	g_signal_emit (obj, signals[CHANGED_SIGNAL], 0);
	return TRUE;
}

